/*****************************************
 Copyright  2001-2003  
 Sigma Designs, Inc. All Rights Reserved
 Proprietary and Confidential
 *****************************************/
/**
  @file   rmdbg.h
  @brief  
  
  USERS SUMMARY
  -------------
  The global switch ``-D_DEBUG=1'' is the official toggle between
  debug and release mode (this fits the default behavior of
  VisualC++).

  When compiling without _DEBUG, all debugging output disappears.

  When compiling with _DEBUG, the file rmdef_project/test/main.c shows
  how the debug facilities RMDBGLOG and RMDBGPRINT are supposed to be
  used.

  Toggle the `#if 1' to `#if 0' at top of file allows the user to
  quickly enable or disable all the debugging messages within a .c file.

  If a message is very important, the logic is to not use
  RMDBGLOG((MYMODULEDBG,...)) but rather: RMDBGLOG((ENABLE,...)) so that it
  always appears in debug mode.

  Note1: when using RMDBGLOG the message should be \n terminated.
  
  Note2: since RMDBGLOG outputs file name and line number, it is not suitable to
  output an array. This is the purpose of RMDBGPRINT (see rmdef_project/test/main.c)

  LIBRARY IMPLEMENTATION DETAILS
  ------------------------------
  The debug facility supposes some access to a `printf' like
  function, and some available output stream. The file name may also
  have UNICODE characters.

  Such features are not available at rmdef.h level. So, in the
  proposed implementation:

  - (default) the user is supposed to implement the two functions
  RMDBGLOG_implementation and and RMDBGPRINT_implementation according to
  the prototypes below. A sample implementation can be found in 
  rmdef_project/test/dbgimplementation.c

  - if the project is compiled with global switch ``DISCARDDBG=1'', these functions are
  replaced by empty macros (and thus completely discarded), even in debug mode.
  
  @author Emmanuel Michon
  @date   2002-11-05
*/

#ifndef __RMDBG_H__
#define __RMDBG_H__

RM_EXTERN_C_BLOCKSTART

#define ENABLE TRUE, __FILE__, __LINE__
#define DISABLE FALSE, __FILE__, __LINE__

void RMDBGLOG_implementation(RMbool active,const RMascii *filename,RMint32 line,const RMascii *text,...);
void RMDBGPRINT_implementation(RMbool active,const RMascii *filename,RMint32 line,const RMascii *text,...);

#ifdef _DEBUG
#define RMDBGLOG(x) RMDBGLOG_implementation x
#define RMDBGPRINT(x) RMDBGPRINT_implementation x
#else
#define RMDBGLOG(x) do {} while (0)
#define RMDBGPRINT(x) do {} while (0)
#endif // _DEBUG

///
/**
   Tests if the argument is not zero.
   If it's zero prints a message telling what expression failed.
*/
#define RMASSERT(x) 						\
do {								\
	if (!(x)) 						\
		RMDBGLOG((ENABLE, "ASSERT: %s fails\n", # x));	\
} while (0)

#ifdef DISCARDDBG

#if (RMCOMPILERID==RMCOMPILERID_VISUALC)
// variadic macros do not exist with Visual
#define RMDBGLOG_implementation(x) do {} while (0)
#define RMDBGPRINT_implementation(x) do {} while (0)
#else
#define RMDBGLOG_implementation(x...) do {} while (0)
#define RMDBGPRINT_implementation(x...) do {} while (0)
#endif

#endif // DISCARDDBG

RM_EXTERN_C_BLOCKEND

#endif // __RMDBG_H__
