/*****************************************
 Copyright  2004-2005
 Sigma Designs, Inc. All Rights Reserved
 Proprietary and Confidential
 *****************************************/
/**
  @file   rmfeatures.h
  @brief  

  Some CHIPID dependent stuff are properly described by XML.  
  However this is not enough when this impacts code

  @author Emmanuel Michon
  @date   2005-02-16
*/

#ifndef __RMFEATURES_H__
#define __RMFEATURES_H__

/*******************************************************************
************************************** decoding capabilities
*******************************************************************/

#ifndef EM86XX_CHIP
/* Well, duh, warnings are considered errors in most of the build */
/* #warning Ignoring contents of rmfeatures.h (EM86XX_CHIP not defined) */
#else /* EM86XX_CHIP */

#if (EM86XX_CHIP>=EM86XX_CHIPID_TANGO)
#define RMFEATURE_HAS_WMV9 1
#endif

#if (EM86XX_CHIP>=EM86XX_CHIPID_TANGO15)
#define RMFEATURE_HAS_H264 1
#endif

/* hack: maybe audio wmapro here? */

/*******************************************************************
************************************** sizes of the DSP PMEM and DMEM
*******************************************************************/

#define CPU_LOCALRAM_SIZE 0x2000

#if (EM86XX_CHIP==EM86XX_CHIPID_TANGO2) && (EM86XX_REVISION>=4)
#define MPEG_PM_SIZE 0x3000
#elif (EM86XX_CHIP==EM86XX_CHIPID_TANGO15) && (EM86XX_REVISION>='B')
#define MPEG_PM_SIZE 0x2800
#else
#define MPEG_PM_SIZE 0x2000
#endif
#define MPEG_DM_SIZE 0x1000

#define DEMUX_PM_SIZE 0x4000
#define DEMUX_DM_SIZE 0x2000

#if (EM86XX_CHIP>=EM86XX_CHIPID_TANGO2)
#define AUDIO_PM_SIZE 0x3000
#define AUDIO_DM_SIZE 0x8000
#else
#define AUDIO_PM_SIZE 0x2800
#define AUDIO_DM_SIZE 0x8000
#endif

#if (EM86XX_CHIP==EM86XX_CHIPID_MAMBO)
#define RMFEATURE_HAS_AUDIO_ENGINE_1 1
#endif

/*******************************************************************
************************************** usage of PLLs
*******************************************************************/

#if (EM86XX_CHIP>=EM86XX_CHIPID_TANGO2)
#define PLL_SYSTEM 3
#else
#define PLL_SYSTEM 0
#endif

#if (EM86XX_CHIP<EM86XX_CHIPID_TANGO)
#define PLL_AUDIO PLLGen_pll_3  /* use normal PLL on Mambo */
#else
#define PLL_AUDIO PLLGen_cd_0  /* use clean divider on Tango and up */
#endif 

#if (EM86XX_CHIP<EM86XX_CHIPID_TANGO2)
#define PLL_VIDEO_PRI PLLGen_pll_1  /* use PLL 1 for primary route */
#define PLL_VIDEO_SEC PLLGen_pll_2  /* use PLL 2 for secondary route */
#else
#define PLL_VIDEO_PRI PLLGen_cd_8  /* use CD 8 for primary route */
#define PLL_VIDEO_SEC PLLGen_cd_9  /* use CD 9 for secondary route */
#endif 

/*******************************************************************
************************************** within display block
*******************************************************************/

/* Full or Light Display Block? */
#if ( \
	(EM86XX_CHIP==EM86XX_CHIPID_TANGOLIGHT) || \
	(EM86XX_CHIP==EM86XX_CHIPID_TANGO15) || \
	(EM86XX_CHIP==EM86XX_CHIPID_TANGO2) \
)
#define RMFEATURE_HAS_LIGHT_DISPLAY 1
#endif

/* Specific display features */
#define RMFEATURE_HAS_VIDEO_IN 1
#define RMFEATURE_HAS_DIGITAL_OUT 1
#define RMFEATURE_HAS_MAIN_ANALOG_OUT 1
#define RMFEATURE_HAS_GFX_SCALER 1
#define RMFEATURE_HAS_MAIN_MIXER 1

#ifndef RMFEATURE_HAS_LIGHT_DISPLAY
#define RMFEATURE_HAS_GRAPHIC_IN 1
#define RMFEATURE_HAS_COMPONENT_OUT 1
#define RMFEATURE_HAS_COMPOSITE_OUT 1
#define RMFEATURE_HAS_VCR_SCALER 1
#define RMFEATURE_HAS_CRT_SCALER 1
#define RMFEATURE_HAS_VCR_MIXER 1
#endif

#if (EM86XX_CHIP>=EM86XX_CHIPID_TANGO)
#define RMFEATURE_HAS_CAPTURE_COUNTER 1  /* chip provides line and pixel counters for input */
#define RMFEATURE_HAS_VBI_CAPTURE 1  /* chip has raw VBI capture capability */
#define RMFEATURE_HAS_GRAPHIC_IN 1
#endif

#if (EM86XX_CHIP>=EM86XX_CHIPID_TANGO15)
#define RMFEATURE_HAS_ANC_CAPTURE 1  /* chip can capture ANC-formatted (sliced) VBI data */
#endif

#if (EM86XX_CHIP>=EM86XX_CHIPID_TANGO15)
#define RMFEATURE_HAS_SPU_SCALER 1  /* chip has working SPU scaler */
#endif

#if (EM86XX_CHIP>=EM86XX_CHIPID_TANGO2)
#define RMFEATURE_HAS_GENERIC_SPU_SCALER /* chip has a 8bit graphic scaler for SPU */
#endif

#if ((EM86XX_CHIP==EM86XX_CHIPID_TANGOLIGHT) || (EM86XX_CHIP==EM86XX_CHIPID_TANGO2))
#define RMFEATURE_HAS_REDUCED_MULTISCALERBUFFER 1  /* size of multi-scaler input buffer is reduced */
#endif

#if (EM86XX_CHIP>=EM86XX_CHIPID_TANGO)
#define RMFEATURE_HAS_HWI2C 1  /* chip has hardware i2c interface (master and slave) */
#endif

/* hack tbd by Christian W. */
/* sizes of some buffers (IN BYTES) */
/* etc. */
/* use format RMFEATURE_HAS_XXX */

/*******************************************************************
************************************** cacheability issues
*******************************************************************/
#ifndef UNCACHED
/* Define all the cacheability macros only if they haven't
 * already be defined elsewhere. Case in point: when rmdef.h
 * is included in some kernel module code, then there are
 * already most likely definitions for these macros in the
 * kernel headers
 */
#if (EM86XX_CHIP>=EM86XX_CHIPID_TANGO2)

#ifdef __ASSEMBLY__
#define RMCPHYSADDR(p)            ((p) & 0x1fffffff)
#define CACHED(p)   (RMCPHYSADDR(p)|0x80000000)
#define UNCACHED(p) (RMCPHYSADDR(p)|0xa0000000)
#else /* !__ASSEMBLY__ */
#ifndef UNCACHED
#define RMCPHYSADDR(p)            ((unsigned int)(p) & 0x1fffffff)
#define CACHED(p)   (RMCPHYSADDR((unsigned int)(p))|0x80000000)
#define UNCACHED(p) (RMCPHYSADDR((unsigned int)(p))|0xa0000000)
#endif
#endif

#else  /* EM86XX_CHIP>=EM86XX_CHIPID_TANGO2 */

#ifdef __ASSEMBLY__
#define RMCPHYSADDR(p) ((p)&(~(1<<31)))
#define CACHED(p)   ((p)|(1<<31))
#define UNCACHED(p) RMCPHYSADDR(p)
#else /* !__ASSEMBLY__ */
#define RMCPHYSADDR(p) ((unsigned int)(p)&(~(1<<31)))
#define CACHED(p)   ((unsigned int)(p)|(1<<31))
#define UNCACHED(p) RMCPHYSADDR(p)
#endif

#endif /* EM86XX_CHIP>=EM86XX_CHIPID_TANGO2 */

#endif /* UNCACHED */

#endif /* EM86XX_CHIP */
#endif /* __RMFEATURES_H__ */
