/*****************************************
 Copyright  2001-2003  
 Sigma Designs, Inc. All Rights Reserved
 Proprietary and Confidential
 *****************************************/
/**
  @file   rmmacros.h
  @brief  

  @author Emmanuel Michon
  @date   2002-10-17
*/

#ifndef __RMMACROS_H__
#define __RMMACROS_H__

RM_EXTERN_C_BLOCKSTART

/**
   Standard min macro.
*/
#define RMmin(x,y) (((x) < (y)) ? (x) : (y))


/**
   Standard max macro.
*/
#define RMmax(x,y) (((x) < (y)) ? (y) : (x))

/**
   Standard abs macro.
 */
#define RMabs(x) (((x) > 0) ? (x) : (-(x)))

/**
   Standard sign macro.
 */
#define RMsign(x) (((x) >= 0) ? (1) : (-1))


/// 
/**
   @param l 64 bit unsigned integer.
   @return the lowest 32 bits from a 64 bits long unsigned integer.
*/
#define RMuint32LSBfromRMuint64(l)     ( (RMuint32)   (l)                      )

/// 
/**
   @param l 64 bit unsigned integer.
   @return the highest 32 bits from a 64 bits long unsigned integer.
*/
#define RMuint32MSBfromRMuint64(l)     ( (RMuint32) ( ((l)>>32) & 0xffffffff ) )

/// 
/**
   @param MSB high-order 32 bits.
   @param LSB low-order 32 bits.
   @return the 64 bits unsigned integer such that its 32 high-order bits are MSB and
           its 32 low-order bits are LSB.
*/
#define RMuint64from2RMuint32(MSB,LSB) ( (RMuint64) ( (((RMuint64)MSB)<<32) + ((RMuint64)LSB) )      )

/// 
/**
   @param l 32 bit unsigned integer.
   @param n 32 bit unsigned integer.
   @return 64 bits unsigned integer resulting from left shift of l.
*/
#define RMuint64fromRMuint32ShiftLeft(l,n)  ( ((RMuint64)l)<<n )

/// 
/**
   @param l 64 bit unsigned integer.
   @param n 32 bit unsigned integer.
   @return 32 bits unsigned integer resulting from right shift of l.
*/
#define RMuint32fromRMuint64ShiftRight(l,n)  ( (RMuint32) (l>>n) )


/**
   --------------- e.m. use this one! --------------------

   Insert the bits 0..bits-1 from data starting at bit number `shift' in *ptr, 
   replacing whatever was there.
   
   In the (rare) cases where bits or shift are variable arguments (not
   known at compile time), use RMinsShiftBitsVar.

   @param ptr: pointer to RMuint32 to be manipulated
   @param data: source data
   @param bits: number of bits to use from data
   @param shift: number of bits to shift data by to the left
   @return the new value of *ptr
*/
#define RMinsShiftBits(ptr, data, bits, shift) RMsetConsecutiveBits(ptr,shift,(shift)+(bits)-1,data)
#define RMinsShiftBitsVar(ptr, data, bits, shift) RMsetConsecutiveBitsVar(ptr,shift,(shift)+(bits)-1,data)

// e.m. deprecated --- use RMinsShiftBitsVar()
#define RMsetConsecutiveBitsVar(target,begin,end,value)		\
	*(target)=(						\
		   (						\
		    (*(target))					\
		    &						\
		    (~(						\
		       ( ( 1 << ((end)-(begin)+1) ) - 1 )	\
		       <<					\
		       (begin)					\
		      )						\
                    )						\
		   )						\
		   |						\
		   ( \
                    ( (value) & ( ( 1 << ((end)-(begin)+1) ) - 1 ) ) \
                    << \
                    (begin) \
		   ) \
                  )

// e.m. deprecated --- use RMinsShiftBits()
#if RMCOMPILERID==RMCOMPILERID_MIPSEL_GCC
#define RMsetConsecutiveBits(target,begin,end,value)			\
do __asm__ __volatile__(							\
		     "ins %0, %1, " #begin ", (" #end "-" #begin "+1)"	\
		     : "=r" (*target) /* may be same reg */ 	\
		     : "r" (value), "0" (*target)); while (0)
#else
#define RMsetConsecutiveBits RMsetConsecutiveBitsVar
#endif

// e.m. deprecated --- use RMinsShiftBits()
#define RMshiftBits(data, bits, shift) ( \
        (((RMuint32)(data)) & ((1 << (bits)) - 1)) << (shift) \
)

// e.m. deprecated --- use RMinsShiftBits()
#define RMinsShiftBool(ptr, data, shift) RMinsShiftBits(ptr,(data)?1:0,1,shift)

// e.m. deprecated --- use RMinsShiftBits()
#define RMshiftBool(data, shift) ( \
        (RMuint32)((data) ? (1 << (shift)) : 0) \
)

/**
   --------------- e.m. use this one! --------------------

   Bit manipulation operation.

   [RMunshiftBits(data, 3, 9) maps directly in mips to `ext $y, $x, 9, 3']

   In the (rare) cases where bits or shift are variable arguments (not
   known at compile time), use RMinsShiftBitsVar.

   @param data source data.
   @param bits: number of bits to use from data
   @param shift: number bits to shift data by to the right
   @return 'bits' bits from position 'shift' of 'data'.
   @remark for example, if 'data' is '00110100', RMunshiftBits(data, 3, 2) returns '101'
*/
#if RMCOMPILERID==RMCOMPILERID_MIPSEL_GCC
#define RMunshiftBits(data, bits, shift)			\
({								\
	RMuint32 dst;						\
								\
	__asm__ __volatile__(					\
			     "ext %0, %1, " #shift ", " #bits	\
			     : "=r" (dst) /* may be same reg */  \
			     : "r" (data));			\
								\
	dst;							\
})
#else
#define RMunshiftBits RMunshiftBitsVar
#endif

#define RMunshiftBitsVar(data, bits, shift) ( \
        (((RMuint32)(data)) >> (shift)) & ((1 << (bits)) - 1) \
)

// e.m. deprecated --- use RMunshiftBits()
#define RMunshiftBool(data, shift) RMunshiftBits(data, 1, shift)
// e.m. deprecated --- use RMunshiftBits()
#define RMgetConsecutiveBits(target,begin,end) RMunshiftBits(target, (end)-(begin)+1, begin)

/**
   @param buf 4 bytes-long buffer.
   @return a RMuint32 from 4 bytes of a bigendian buffer
*/
static inline RMuint32 RMbeBufToUint32(RMuint8 *buf)
{
	return (((RMuint32) buf[0] << 24) + \
		((RMuint32) buf[1] << 16) + \
		((RMuint32) buf[2] << 8) + \
		(RMuint32)  buf[3]);
}

/**
   @param buf 4 bytes-long buffer.
   @return a RMuint32 from 4 bytes of a littleendian buffer
*/
static inline RMuint32 RMleBufToUint32(RMuint8 *buf)
{
	return (((RMuint32) buf[3] << 24) + \
		((RMuint32) buf[2] << 16) + \
		((RMuint32) buf[1] << 8) + \
		(RMuint32)  buf[0]);
}

/**
   Transfroms a 32 bits integer into 4 bytes in big endian order

   @param val 
   @param buf   
*/
static inline void RMuint32ToBeBuf(RMuint32 val, RMuint8 *buf)
{
#if (defined WITH_32BIT_BUF_ALIGNMENT) && (RMPLATFORMID == RMPLATFORMID_JASPERMAMBO) 
        /* Standalone ARM platform : optimized swapping - as WMAPro lib uses this intensively now... */
        *(RMuint32 *) buf = ((val & 0xff000000) >> 24) | ((val & 0x00ff0000) >>  8) |
                            ((val & 0x0000ff00) <<  8) | ((val & 0x000000ff) << 24); 
#else
	/* default other platforms */
	buf[0] = (RMuint8)(val >> 24) & 0xff;
	buf[1] = (RMuint8)(val >> 16) & 0xff;
	buf[2] = (RMuint8)(val >>  8) & 0xff;
	buf[3] = (RMuint8)val & 0xff;
#endif /* WITH_32BIT_BUF_ALIGNMENT && RMPLATFORMID == */ 
}

/**
   Transfroms a 32 bits integer int o 4 bytes in little endian order

   @param val 
   @param buf   
*/
static inline void RMuint32ToLeBuf(RMuint32 val, RMuint8 *buf)
{
#if (defined WITH_32BIT_BUF_ALIGNMENT) && (RMPLATFORMID == RMPLATFORMID_JASPERMAMBO) 
	/* Standalone platform : optimized swapping - as WMAPro lib uses this intensively now... */
        *(RMuint32 *) buf = ((val & 0xff000000) >> 24) | ((val & 0x00ff0000) >>  8) |
                            ((val & 0x0000ff00) <<  8) | ((val & 0x000000ff) << 24); 
#else 
	/* default other platforms */
	buf[3] = (RMuint8)(val >> 24);
	buf[2] = (RMuint8)(val >> 16);
	buf[1] = (RMuint8)(val >>  8);
	buf[0] = (RMuint8)val;
#endif /* WITH_32BIT_BUF_ALIGNMENT && RMPLATFORMID == */ 
}

/**
   @param buf 3 bytes-long buffer.
   @return a RMuint24 from 3 bytes of a bigendian buffer
*/
static inline RMuint32 RMbeBufToUint24(RMuint8 *buf)
{
	return (((RMuint32) buf[0] << 16) + \
		((RMuint32) buf[1] << 8) + \
		(RMuint32) buf[2]);
}

/**
   @param buf 3 bytes-long buffer.
   @return a RMuint24 from 4 bytes of a littleendian buffer
*/
static inline RMuint32 RMleBufToUint24(RMuint8 *buf)
{
	return (((RMuint32) buf[2] << 16) + \
		((RMuint32) buf[1] << 8) + \
		(RMuint32) buf[0]);
}

/**
   Transform a 24 bits integer (stored in a RMuint32) into 3 bytes in big endian order
   
   @param val
   @param buf
*/
static inline void RMuint24ToBeBuf(RMuint32 val, RMuint8 *buf)
{
	buf[0] = (RMuint8)(val >>  16);
	buf[1] = (RMuint8)(val >>  8);
	buf[2] = (RMuint8)val;
}

/**
   Transfroms a 24 bits integer into 2 bytes in little endian order

   @param val 
   @param buf   
*/
static inline void RMuint24ToLeBuf(RMuint32 val, RMuint8 *buf)
{
	buf[2] = (RMuint8)(val >>  16);
	buf[1] = (RMuint8)(val >>  8);
	buf[0] = (RMuint8)val;
}

/**
   @param buf 2 bytes-long buffer.
   @return a RMuint16 from 2 bytes of a bigendian buffer
*/
static inline RMuint16 RMbeBufToUint16(RMuint8 *buf)
{
	return (((RMuint16) buf[0] << 8) + \
		(RMuint16)  buf[1]);
}

/**
   @param buf 2 bytes-long buffer.
   @return a RMuint16 from 2 bytes of a littleendian buffer
*/
static inline RMuint16 RMleBufToUint16(RMuint8 *buf)
{
 	return 	(((RMuint16) buf[1] << 8) + \
		(RMuint16)  buf[0]);
}

/**
   Transfroms a 16 bits integer into 2 bytes in big endian order

   @param val 
   @param buf   
*/
static inline void RMuint16ToBeBuf(RMuint16 val, RMuint8 *buf)
{
	buf[0] = (RMuint8)(val >>  8);
	buf[1] = (RMuint8)val;
}

/**
   Transfroms a 16 bits integer into 2 bytes in little endian order

   @param val 
   @param buf   
*/
static inline void RMuint16ToLeBuf(RMuint16 val, RMuint8 *buf)
{
	buf[1] = (RMuint8)(val >>  8);
	buf[0] = (RMuint8)val;
}

// sometimes RMMemset is remapped on a complex operation. Rather use this to zero memory at location:
#define clearmemory(location,size) 				\
do {								\
	RMuint32 i;							\
	for (i=0;i<(RMuint32)(size);i++) *((RMuint8 *)(location)+i)=0;	\
} while (0)

#define copymemory(dest, src, size) \
do { \
	RMuint32 i; \
	for (i=0;i<(RMuint32)(size);i++) *((RMuint8 *)(dest)+i)=*((RMuint8 *)(src)+i); \
} while (0)

#define RM64mult32div32(a,b,c) ( (RMuint64)(a) * (RMuint32)(b) / (RMuint32)(c) )

/// 
/**
   Returns the largest page aligned address <= x.

   @param x     
   @return -ReturnValue-
*/
#define RM_PAGE_SIZE 4096
#define RM_PAGE_ALIGN(x) (((RMuint32)(x))&~(RM_PAGE_SIZE-1))

/// 
/**
   Classic variant of the above, not to be mismatched

   Typically: p=RMMalloc(size+RM_PAGE_SIZE-1); Give out RM_NEXT_PAGE_ALIGN(p).

   @param x     
   @return -ReturnValue-
*/
#define RM_NEXT_PAGE_ALIGN(x) RM_PAGE_ALIGN((x)+RM_PAGE_SIZE-1)

RM_EXTERN_C_BLOCKEND

#endif // __RMMACROS_H__
