/*****************************************
 Copyright  2001-2003	
 Sigma Designs, Inc. All Rights Reserved
 Proprietary and Confidential
 *****************************************/

#define ALLOW_OS_CODE
#include "../include/rmlibcw.h"
#include "rmcore/include/rmcore.h"

#if 0
  #define THDBG DISABLE
#else
  #define THDBG ENABLE
#endif

// _NP below means Non Posix. This code uses GNU features (recursive mutex)
#if (RMCOMPILERID==RMCOMPILERID_GCC) 
#define _GNU_SOURCE
#elif (RMCOMPILERID==RMCOMPILERID_ARMELF_GCC)
#define __USE_GNU
#endif // RMCOMPILERID

#include <pthread.h>
#include <semaphore.h>

#include "robust.h"

#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <unistd.h>

/* cannot include string.h */
int strlen(const char *s);

typedef struct {
	RMfunc sr;
	void *arg;
	RMascii *name;
	const RMascii *userName;
	// we may add other fields as time goes by.
} registeredThread;

typedef struct {
	pthread_t tid;
} _RMthread;

static pthread_key_t key;
static RMbool rmthreads_initialized=FALSE;
static sem_t sonisalive;
static char main_name[20] = "Main thread";
pthread_once_t key_once = PTHREAD_ONCE_INIT;

static void registeredThread_destroy(void *p)
{
	registeredThread *rTh = (registeredThread*)p;
	
	if(rTh != (registeredThread *)NULL){
		if(rTh->name != (RMascii*)NULL){
			RMFreeAscii((RMascii*)rTh->name);
			rTh->name = (RMascii*)NULL;
		}
		RMFree(rTh);	
		rTh = (registeredThread *)NULL;
	}
}

static void key_alloc()
{
	pthread_key_create(&key, registeredThread_destroy);
}

static void *slave_init_code(void *p)
{
	registeredThread *rTh = (registeredThread*)p;

	// we assume here the max value of a pid is 65535: on 5 characters.
	rTh->name = RMMalloc(strlen(rTh->userName)+1+1+5+1);
	RMPrintAscii(rTh->name,"%s(%d)",rTh->userName, getpid());

	pthread_once(&key_once, key_alloc);

	if(robust_pthread_setspecific(key, rTh)) {
		perror("Error pthread_setspecific");
		RMPanic(RM_FATALCREATETHREAD);
	}

	RMDBGLOG((THDBG,"I am a new thread #%08lx (pid %d)\n",(RMuint32) pthread_self(), getpid()));

	robust_sem_post(&sonisalive);

	(* rTh->sr) (rTh->arg);

	rTh = (registeredThread*)robust_pthread_getspecific(key);
	
	return NULL;
}

RMthread RMCreateThread(const RMascii *name, RMfunc StartAddress, void * Parameter)
{
	pthread_t tid;
	registeredThread *rTh;

	if (!rmthreads_initialized) {
		robust_sem_init(&sonisalive,0,0);		
		rmthreads_initialized=TRUE;
	}

	rTh = RMMalloc(sizeof(registeredThread));
	rTh->sr = StartAddress;
	rTh->arg = Parameter;
	rTh->userName = name;
	rTh->name = (RMascii*)NULL;

	if (pthread_create(&tid, NULL, slave_init_code, rTh)) {
		perror("Error rmthread");
		RMPanic(RM_FATALCREATETHREAD);
	}

	robust_sem_wait(&sonisalive);

	RMDBGLOG((THDBG,"Father started new son #%08lx:\"%s\" at (%p,%p)\n",(RMuint32)tid, (char *)name, StartAddress, Parameter));

	return (RMthread)tid;
}

RMthread RMGetCurrentThread(void)
{
	return (RMthread) pthread_self();
}

const RMascii *RMGetCurrentThreadName(void)
{
	registeredThread *rTh;

	errno = 0;
	rTh = (registeredThread*)robust_pthread_getspecific(key);
	
	if( !rmthreads_initialized ||
	   (rTh == (registeredThread *)NULL)){
		return (const RMascii *)main_name;
	}

	return (const RMascii *) rTh->name;
}

void RMWaitForThreadToFinish(RMthread hThread)
{
	RMDBGLOG((THDBG,"Thread waiting for #%08lx to finish\n",hThread /*,(char *)reg->name*/));
	
	pthread_join((pthread_t)hThread,NULL);
}

void cancelallotherthreads(void);

void cancelallotherthreads(void)
{
}
