/*
 *--------------------------------------------------------------------
 *
 * COPYRIGHT (c)  2002-2004  INTEL CORPORATION.      ALL  RIGHTS 
 * RESERVED.  THIS  PROGRAM IS  FURNISHED  UNDER  LICENSE  AND MAY
 * ONLY BE USED OR  COPIED IN  ACCORDANCE WITH  THE  TERMS OF THAT 
 * LICENSE.  NO  PART  OF THIS  PROGRAM  OR  PUBLICATION  MAY  BE
 * REPRODUCED,  TRANSMITTED,  TRANSCRIBED,  STORED IN  A RETRIEVAL 
 * SYSTEM, OR  TRANSLATED INTO ANY  LANGUAGE OR COMPUTER  LANGUAGE 
 * IN ANY FORM OR BY  ANY MEANS, ELECTRONIC, MECHANICAL, MAGNETIC,
 * OPTICAL,  CHEMICAL, MANUAL, OR  OTHERWISE,  WITHOUT  THE  PRIOR 
 * WRITTEN PERMISSION OF INTEL  CORPORATION.  EXCEPT AS  EXPRESSLY
 * GRANTED IN A LICENSE AGREEMENT, THIS PROGRAM IS FURNISHED "AS-IS"
 * WITH NO WARRANTY WHATSOEVER, INCLUDING IMPLIED WARRANTIES OF
 * MERCHANTABILITY,  FITNESS FOR A  PARTICULAR PURPOSE,   AND NON-
 * INFRINGEMENT.    
 *           
 *--------------------------------------------------------------------
 *     File Name: CommandLineShell.c
 *     Author:    Intel Corporation
 *
 */

#include "CommandLineShell.h"
#include "ILibParsers.h"
#include "UPnPControlPointStructs.h"

#ifdef WIN32
#include "WinSemaphore.h"
#include <crtdbg.h>
#endif

#ifdef _POSIX
#include <semaphore.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>
#include "MmsCp.h"

/* #### Begin CARDEA code #### */
#define ALLOW_OS_CODE
#include "rmupnp/rmlibwmdrmndupnp/include/ms_upnp.h"
/* #### End CARDEA code #### */

#ifdef _DEBUG
#define DEBUGONLY(x) x
#else
#define DEBUGONLY(x)
#endif

#define DEFAULTDELIMITER '/'
#define MAX_STR_LEN 1024
#define TMP_BUFFER_SIZE 1024

#include "MyString.h"

#include "http.h"

/* Arguments to the browse function, needs to be global or malloced (callback) */
static struct MMSCP_BrowseArgs browseArgs;

// the hash lists of devices that have been discovered on the network.
static void *MS_KnownDevicesTable;

// pointers to the control point structs.  These are needed for making calls
// to the control point functions.
static void *MS_ControlPoint;


// BEGIN: Shell state variables

// the UDNs of the currently selected renderers.  They should be empty
// strings when no device is selected
static char *currMServerUDN;

// When browsing through the tree on a media server, this is the stack 
// symbolizing the current position in the tree.  The contents of the 
// stack are the object ID's of the nodes.  Note that the object id: "0" should
// always be at the bottom of the stack.
static void *currBrowseDirectoryStack;


// the most recent results from a browse request.  The results come back in a 
// callback instead of being returned to the calling function, so this 
// global variable is necessary in order to give the results to the calling 
// function.
struct MMSCP_ResultsList *mostRecentBrowseResults;

//the URI of current media server
char CurrMServerURI[MAX_STR_LEN];

// END: Shell state variables.

/* This buffer should only be used locally */
static char tmp_buffer[TMP_BUFFER_SIZE];

/* Pops items off the media server browsing directory stack until only the 
 * root directory is left on the stack.  Note that this will free the items
 * on the stack.
 */
static void popDirectoryStackToRoot(void)
{
	char *temp;
	while(strcmp("0", ILibPeekStack(&currBrowseDirectoryStack)))
	{
		temp = ILibPopStack(&currBrowseDirectoryStack);
		free(temp);
	}
}



/********************************************************************
*********************************************************************

Callbacks to be used with the MS control point invoke 
functions.  

*********************************************************************
********************************************************************/
/*static void MSCP_ResponseSink_ConnectionManager_GetCurrentConnectionInfo(struct UPnPService* Service,int ErrorCode,void *User,int RcsID,int AVTransportID,char* ProtocolInfo,char* PeerConnectionManager,int PeerConnectionID,char* Direction,char* Status)
{
	DEBUGONLY(printf("MSCP_ Invoke Response: ConnectionManager/GetCurrentConnectionInfo[ErrorCode:%d](%d,%d,%s,%s,%d,%s,%s)\r\n",ErrorCode,RcsID,AVTransportID,ProtocolInfo,PeerConnectionManager,PeerConnectionID,Direction,Status);)
}*/

/*static void MSCP_ResponseSink_ConnectionManager_GetProtocolInfo(struct UPnPService* Service,int ErrorCode,void *User,char* Source,char* Sink)
{
	DEBUGONLY(printf("MSCP_ Invoke Response: ConnectionManager/GetProtocolInfo[ErrorCode:%d](%s,%s)\r\n",ErrorCode,Source,Sink);)
}*/

/*static void MSCP_ResponseSink_ConnectionManager_GetCurrentConnectionIDs(struct UPnPService* Service,int ErrorCode,void *User,char* ConnectionIDs)
{
	DEBUGONLY(printf("MSCP_ Invoke Response: ConnectionManager/GetCurrentConnectionIDs[ErrorCode:%d](%s)\r\n",ErrorCode,ConnectionIDs);)
}*/

/*static void MSCP_ResponseSink_ContentDirectory_Browse(struct UPnPService* Service,int ErrorCode,void *User,char* Result,unsigned int NumberReturned,unsigned int TotalMatches,unsigned int UpdateID)
{
	DEBUGONLY(printf("MSCP_ Invoke Response: ContentDirectory/Browse[ErrorCode:%d](%s,%u,%u,%u)\r\n",ErrorCode,Result,NumberReturned,TotalMatches,UpdateID);)
}*/

/*static void MSCP_ResponseSink_ContentDirectory_GetSortCapabilities(struct UPnPService* Service,int ErrorCode,void *User,char* SortCaps)
{
	DEBUGONLY(printf("MSCP_ Invoke Response: ContentDirectory/GetSortCapabilities[ErrorCode:%d](%s)\r\n",ErrorCode,SortCaps);)
}*/

/*static void MSCP_ResponseSink_ContentDirectory_GetSystemUpdateID(struct UPnPService* Service,int ErrorCode,void *User,unsigned int Id)
{
	DEBUGONLY(printf("MSCP_ Invoke Response: ContentDirectory/GetSystemUpdateID[ErrorCode:%d](%u)\r\n",ErrorCode,Id);)
}*/

/*static void MSCP_ResponseSink_ContentDirectory_GetSearchCapabilities(struct UPnPService* Service,int ErrorCode,void *User,char* SearchCaps)
{
	DEBUGONLY(printf("MSCP_ Invoke Response: ContentDirectory/GetSearchCapabilities[ErrorCode:%d](%s)\r\n",ErrorCode,SearchCaps);)
}*/

/*static void MSCP_EventSink_ConnectionManager_SourceProtocolInfo(struct UPnPService* Service,char* SourceProtocolInfo)
{
	DEBUGONLY(printf("MSCP_ Event from %s/ConnectionManager/SourceProtocolInfo: %s\r\n",Service->Parent->FriendlyName,SourceProtocolInfo);)
}*/

/*static void MSCP_EventSink_ConnectionManager_SinkProtocolInfo(struct UPnPService* Service,char* SinkProtocolInfo)
{
	DEBUGONLY(printf("MSCP_ Event from %s/ConnectionManager/SinkProtocolInfo: %s\r\n",Service->Parent->FriendlyName,SinkProtocolInfo);)
}*/

/*static void MSCP_EventSink_ConnectionManager_CurrentConnectionIDs(struct UPnPService* Service,char* CurrentConnectionIDs)
{
	DEBUGONLY(printf("MSCP_ Event from %s/ConnectionManager/CurrentConnectionIDs: %s\r\n",Service->Parent->FriendlyName,CurrentConnectionIDs);)
}*/

/*static void MSCP_EventSink_ContentDirectory_SystemUpdateID(struct UPnPService* Service,unsigned int SystemUpdateID)
{
	DEBUGONLY(printf("MSCP_ Event from %s/ContentDirectory/SystemUpdateID: %u\r\n",Service->Parent->FriendlyName,SystemUpdateID);)
}*/


static void prompt(){
	printf("command> ");
	fflush(stdout);
}


static void MSCP_DeviceAddedOrRemoved(struct UPnPDevice *device, int added)
{
	ILibHashTree_Lock(MS_KnownDevicesTable);

	if(added)
	{
		ILibAddEntry(
			MS_KnownDevicesTable, 
			device->UDN, 
			(int)strlen(device->UDN), 
			device->FriendlyName);
	}
	else
	{
		ILibDeleteEntry(
			MS_KnownDevicesTable,
			device->UDN,
			(int)strlen(device->UDN));
	}

	ILibHashTree_UnLock(MS_KnownDevicesTable);
}



static void MSCP_BrowseResponded(void *serviceObj, 
		struct MMSCP_BrowseArgs *args, 
		int errorCode, 
		struct MMSCP_ResultsList *results)
{
	void (*callback)(void);
	
	DEBUGONLY(printf("A browse request came back...\n");)
	DEBUGONLY(printf("Error code: %d\n", errorCode);)
	if(0 == errorCode)
	{
		if(mostRecentBrowseResults != NULL)
		{
			printf("ERROR: browse results returned, but");
			printf(" mostRecentBrowseResults is not NULL\n");
			
			// prevent memory leaks leave the old results, and
			// destroy the new results because there might be 
			// other references to the old stuff.
			MMSCP_DestroyResultsList(results);
		}
		else
		{
			mostRecentBrowseResults = results;
			
		}

		callback = args->UserObject;
		if (callback){
			callback();
			/* Put the prompt back after we have finished the
			 * callback */
			prompt();
		}
	}
}



/***************************************************************************
****************************************************************************

End of callbacks

****************************************************************************
***************************************************************************/





//Identify the delimiter character in a path string
static void identifyDelimiter(char * delimiter, char * path)
{
	char * tmp = path;
	*delimiter = '/'; //default value
	if(tmp == NULL || strlen(tmp) == 0)
		return;

	while('\0' != *tmp && !('\\'== *tmp || '/' == *tmp))
		tmp++;

	if('\0' != *tmp)
		*delimiter = *tmp;
}

//do not show "0" for the root
static void ignoreRoot(char ** path)
{
	char * tmp = *path;
	if('0' == *tmp)
		(*path)++;		
}

static void ignoreParentPath(char ** path, char delimiter)
{
	char * tmp = *path;
	unsigned int pos = 0, i = 0;

	if(NULL == tmp) 
		return;
	while('\0' != *tmp){
		if(delimiter == *tmp)
			pos = i;
		i++;
		tmp++;
	}
	if(pos == strlen(*path) - 1)
		return;

	*path += pos+1;
}

//change delimiter of the path string that is 
//ended with '\0'
static void changeDelimiter(char * path, char dmt)
{
	char * tmp = path;
	if(NULL == tmp) 
		return;
	while(*tmp != '\0')
	{
		if(*tmp == '\\' || *tmp == '/')
			*tmp = dmt;
		tmp++;
	}
}

static struct UPnPDevice *getCurrentMServer()
{
	struct UPnPDevice *ret;
	
	CurrMServerURI[0] = '\0';
	if('\0' == currMServerUDN[0])
	{
		printf("No server selected, select one with setms\n");
		ret = NULL;
	}
	else
	{
		ret = MSCP_GetDeviceAtUDN(MS_ControlPoint, currMServerUDN);
		if(NULL == ret)
		{
			printf("The previously selected server has left the network\n");
			printf("Please select a new one with setms\n");
			currMServerUDN[0] = '\0';
			popDirectoryStackToRoot();
		}
		else
			Utf8ToAnsi(CurrMServerURI, ret->LocationURL, MAX_STR_LEN);
	}
	return ret;
}

/* Presents a list to the user, allows the user to choose
 * one, and then returns the ID of the device selected.  
 * Parameters:
 *		theILibHashTable: The ILibHashTree that contains the set of items from
 *                        which the user should choose.  The key should the the ID,
 *						  and the value should be a string describing the item. 
 *      returns a pointer to the buffer containing the freshly malloc'ed ID string        
 */
static char *chooseIDFromTable(void *theILibHashTable)
{
	char *key, *retVal, *tmppath;
	int keyLength;
	void *data;
	char *charData;
	void *enumerator;
	int i, numDevices, chosenIndex, attemptsLeft;

	ILibHashTree_Lock(theILibHashTable);
	enumerator = ILibHashTree_GetEnumerator(theILibHashTable);

	printf("Select one of the following:\n");
	numDevices = 0;
	do
	{
		key = NULL;
		keyLength = 0;
		data = NULL;
		charData = NULL;

		ILibHashTree_GetValue(enumerator, &key, &keyLength, &data);
		charData = (char*)data;

		if(0 != keyLength)
		{
			char * tmp = malloc(keyLength + 1);
			memcpy(tmp, key, keyLength);
			tmp[keyLength] = '\0';
			tmppath = tmp;
			ignoreRoot(&tmppath);
			changeDelimiter(tmppath, DEFAULTDELIMITER);
			printf("%d)\t%s (%s)\n", ++numDevices, charData, tmppath);
			free(tmp);
		}
	}
	while(!ILibHashTree_MoveNext(enumerator));
	ILibHashTree_DestroyEnumerator(enumerator);

	if(numDevices == 0)
	{
		printf("None available\n");
		retVal = MakeStrcpy(NULL);;
	}
	else
	{
		attemptsLeft = 3;
		while(attemptsLeft > 0)
		{
			fgets(tmp_buffer, TMP_BUFFER_SIZE, stdin);
			StrChomp(tmp_buffer);
			if(1 != sscanf(tmp_buffer, "%d", &chosenIndex))
			{
				printf("Please enter a number\n");
			}
			else if(chosenIndex < 0 || chosenIndex > numDevices)
			{
				printf("Please choose one of the above, or 0 for none\n");
				attemptsLeft--;
			}
			else
			{	
				attemptsLeft = 0;
			}
		}
		if(0 == chosenIndex)
		{
			retVal = MakeStrcpy(NULL);
		}
		else
		{
			enumerator = ILibHashTree_GetEnumerator(theILibHashTable);
			for(i=0; i<chosenIndex; i++) ILibHashTree_MoveNext(enumerator);
			key = NULL;
			keyLength = 0;
			data = NULL;
			charData = NULL;

			ILibHashTree_GetValue(enumerator, &key, &keyLength, &data);
			retVal = malloc(keyLength+1);
			memcpy(retVal, key, keyLength);
			retVal[keyLength] = '\0';
			ILibHashTree_DestroyEnumerator(enumerator);
		}		
	}
	ILibHashTree_UnLock(theILibHashTable);
	return retVal;
}


static void handleCmd_ls()
{
	char * tmp, delimiter;
	if(mostRecentBrowseResults != NULL)
	{
		unsigned int currObjectClass;
		void *current_node = ILibLinkedList_GetNode_Head(mostRecentBrowseResults->LinkedList);

		printf("There were %d results\n", mostRecentBrowseResults->TotalMatches);
		while(current_node != NULL)
		{
			struct MMSCP_MediaObject *currObject = ILibLinkedList_GetDataFromNode(current_node);

			currObjectClass = MMSCP_CLASS_MASK_OBJECTTYPE & currObject->MediaClass;
			tmp = currObject->ID;
			ignoreRoot(&tmp);
			identifyDelimiter(&delimiter, tmp);
			ignoreParentPath(&tmp, delimiter);

			if(MMSCP_CLASS_MASK_CONTAINER == currObjectClass)
			{					
				printf("Container: %s (%s)\n", currObject->Title, tmp);
			}
			else if(MMSCP_CLASS_MASK_ITEM == currObjectClass)
			{
				printf("Item:      %s (%s)\n", currObject->Title, tmp);
				/*if (currObject->Res)
				  {
				  if (currObject->Res->Uri)
				  printf("    %s\n", currObject->Res->Uri);
				  if (currObject->Res->ProtocolInfo)
				  printf("    proto : %s\n", currObject->Res->ProtocolInfo);
				  }*/
			}
			else
			{
				printf("Bad Class: %s (%s)\n", currObject->Title, tmp);
			}

			current_node = ILibLinkedList_GetNextNode(current_node);
		}

		MMSCP_DestroyResultsList(mostRecentBrowseResults);
		mostRecentBrowseResults = NULL;
	}
}




// returns 0 on success
static int DoBrowse(void (*callback)(void))
{
	int success = -1;
	struct UPnPService *service;
	struct UPnPDevice *dev = getCurrentMServer();

	if(NULL != dev)
	{

		/* #### Begin CARDEA code #### */
		if (MS_UPNP(dev)){
			if (!(ms_device_status(dev) & MS_ISAUTHORIZED)){
				fprintf(stderr,"Not authorized to browse %s\n", dev->FriendlyName);
				return -1;
			}
		}
		/* #### End CARDEA code #### */

		browseArgs.BrowseFlag = MMSCP_BrowseFlag_Children;
		browseArgs.Filter = "res";
		browseArgs.ObjectID = ILibPeekStack(&currBrowseDirectoryStack);
		browseArgs.RequestedCount = 256;
		browseArgs.SortCriteria = "";
		browseArgs.StartingIndex = 0;
		browseArgs.UserObject = callback;

		service = MSCP_GetService_ContentDirectory(dev);
		
		// send off the browse packet.  Note that this will
		// only block until the browse packet is queued up for
		// sending.  There is a call to waitForResponse in order
		// to block until the response comes back.
		MMSCP_Invoke_Browse(service, &browseArgs);
		
		// block until the list comes back so that the 
		// user doesn't get a prompt until the listing is
		// done. Applies if and only if ASYNC_UI is defined.
		//waitForResponse();

		success = 0;
	}
	return success;
}

static void handleCmd_getms()
{
	struct UPnPDevice *dev = getCurrentMServer();
	if(NULL != dev)
	{
		printf("Current media server: %s\n", dev->FriendlyName);
		MSCP_Release(dev);
	}
	else
	{
		// this output is taken care of by the getCurrentMServer call
	}
}

static void handleCmd_setms()
{
	// prompt the user to choose a new Media server from the 
	// list of known devices, and store the UDN to currMServerUDN
	free(currMServerUDN);
	currMServerUDN = chooseIDFromTable(MS_KnownDevicesTable);
	
	// make sure that if we were browsing on a previous server, we pop back 
	// down to the root
	popDirectoryStackToRoot();
}

/* #### Begin CARDEA code #### */
/**
 * Try to get a MS license for the current context of the current device
 * @param uri - URL to get
 * @param session_header - returned session header if MS DRM
 */
static void try_get_ms_license(char *uri, char **session_header)
{
	struct UPnPDevice *dev = getCurrentMServer();
	RMint32 status;

	if (uri == NULL)
		return;
	
	if (dev != NULL && MS_UPNP(dev)){
		struct timeval tv1, tv2;
		long int tv_sec, tv_usec;

		/* Try to get a license for this URL */
		fprintf(stderr,"MS server, try to get a license\n");

		gettimeofday(&tv1, NULL);
		status = get_ms_license(dev, uri, session_header);
		gettimeofday(&tv2, NULL);
		if (status){
			fprintf(stderr,"Cannot get MS license, or no license needed : %ld\n", -status);
			session_header = NULL;
			if (status == -DRM_MUST_REGISTER){
				/* We have to register again */
				RMint32 register_status;

				fprintf(stderr,"Try to register with %s\n", dev->FriendlyName);
				register_status = register_ms_client(dev);
				if (register_status < 0)
					fprintf(stderr,"Error registering at %s : %ld\n", dev->FriendlyName, register_status);
				else
					fprintf(stderr,"Registration at %s started.\n", dev->FriendlyName);
			}
		}

		tv_sec = tv2.tv_sec - tv1.tv_sec;
		tv_usec = tv2.tv_usec - tv1.tv_usec;
		if (tv_usec < 0) {
			tv_sec--;
			tv_usec = 1000000 + tv_usec;
		}
		printf("Time elapsed during license aquisition : %lu s %lu us\n", tv_sec, tv_usec);
	}
}
/* #### End CARDEA code #### */

static void handleCmd_play()
{
	char *newObjID;

	if(mostRecentBrowseResults != NULL)
	{
		struct ILibHashTree *items = ILibInitHashTree();
		struct ILibHashTree *urls = ILibInitHashTree();
		void *current_node = ILibLinkedList_GetNode_Head(mostRecentBrowseResults->LinkedList);

		while(current_node != NULL)
		{
			struct MMSCP_MediaObject *currObject = ILibLinkedList_GetDataFromNode(current_node);

			if(MMSCP_CLASS_MASK_ITEM & currObject->MediaClass)
			{
				ILibAddEntry(
						items, 
						currObject->ID, 
						(int)strlen(currObject->ID), 
						currObject->Title);
				if (currObject->Res && currObject->Res->Uri)
				{
					/*if (currObject->Res->ProtocolInfo)
					  printf("    proto : %s\n", currObject->Res->ProtocolInfo);*/
					ILibAddEntry(
							urls,
							currObject->ID,
							(int)strlen(currObject->ID),
							currObject->Res->Uri);
				}
			}
			current_node = ILibLinkedList_GetNextNode(current_node);
		}

		newObjID = chooseIDFromTable(items);

		if(newObjID[0] != '\0')
		{
			char *uri = NULL, *title = NULL;

			title = ILibGetEntry(items, newObjID, strlen(newObjID));
			uri = ILibGetEntry(urls, newObjID, strlen(newObjID));
			if (uri && title)
			{
				int status;
				char *session_header = NULL;

				printf("Play %s (%s) from %s\n", title, newObjID, uri);
				
				/* #### Begin CARDEA code #### */
				try_get_ms_license(uri, &session_header);
				/* #### End CARDEA code #### */

				status = start_playback(uri);
				if (status < 0){
					printf("Error : %d\n", status);
				} else {
					printf("OK\n");
				}
			}
			else
				printf("Cannot get %s, no URL\n", newObjID);
		}

		free(newObjID);

		MMSCP_DestroyResultsList(mostRecentBrowseResults);
		mostRecentBrowseResults = NULL;

		ILibDestroyHashTree(items);
		ILibDestroyHashTree(urls);
	}
}

static void handleCmd_get()
{
	char *newObjID;

	if(mostRecentBrowseResults != NULL)
	{
		struct ILibHashTree *items = ILibInitHashTree();
		struct ILibHashTree *urls = ILibInitHashTree();
		void *current_node = ILibLinkedList_GetNode_Head(mostRecentBrowseResults->LinkedList);

		while(current_node != NULL)
		{
			struct MMSCP_MediaObject *currObject = ILibLinkedList_GetDataFromNode(current_node);

			if(MMSCP_CLASS_MASK_ITEM & currObject->MediaClass)
			{
				ILibAddEntry(
						items, 
						currObject->ID, 
						(int)strlen(currObject->ID), 
						currObject->Title);
				if (currObject->Res && currObject->Res->Uri)
				{
					/*if (currObject->Res->ProtocolInfo)
					  printf("    proto : %s\n", currObject->Res->ProtocolInfo);*/
					ILibAddEntry(
							urls,
							currObject->ID,
							(int)strlen(currObject->ID),
							currObject->Res->Uri);
				}
			}
			current_node = ILibLinkedList_GetNextNode(current_node);
		}

		newObjID = chooseIDFromTable(items);

		if(newObjID[0] != '\0')
		{
			char *uri = NULL, *title = NULL;

			title = ILibGetEntry(items, newObjID, strlen(newObjID));
			uri = ILibGetEntry(urls, newObjID, strlen(newObjID));
			if (uri && title)
			{
				int status;
				char *session_header = NULL;
				printf("Get %s (%s) from %s\n", title, newObjID, uri);

				/* #### Begin CARDEA code #### */
				try_get_ms_license(uri, &session_header);
				/* #### End CARDEA code #### */

				status = http_get(uri, session_header);
				if (status < 0){
					printf("Error : %d\n", status);
				} else {
					printf("OK\n");
				}
			}
			else
				printf("Cannot get %s, no URL\n", newObjID);
		}

		free(newObjID);

		MMSCP_DestroyResultsList(mostRecentBrowseResults);
		mostRecentBrowseResults = NULL;

		ILibDestroyHashTree(items);
		ILibDestroyHashTree(urls);
	}
}

static void handleCmd_cd()
{
	char *newObjID;

	if(mostRecentBrowseResults != NULL)
	{
		struct ILibHashTree *containers = ILibInitHashTree();
		void *current_node = ILibLinkedList_GetNode_Head(mostRecentBrowseResults->LinkedList);

		while(current_node != NULL)
		{
			struct MMSCP_MediaObject *currObject = ILibLinkedList_GetDataFromNode(current_node);

			if(MMSCP_CLASS_MASK_CONTAINER & currObject->MediaClass)
			{
				ILibAddEntry(
						containers, 
						currObject->ID, 
						(int)strlen(currObject->ID), 
						currObject->Title);
			}
			current_node = ILibLinkedList_GetNextNode(current_node);
		}

		newObjID = chooseIDFromTable(containers);

		if(newObjID[0] != '\0')
		{
			ILibPushStack(&currBrowseDirectoryStack, newObjID);
		}
		else
		{
			free(newObjID);
		}

		MMSCP_DestroyResultsList(mostRecentBrowseResults);
		mostRecentBrowseResults = NULL;

		ILibDestroyHashTree(containers);
	}
}

static void handleCmd_cdup()
{
	// we don't want to pop the root off now....
	if(strcmp("0", ILibPeekStack(&currBrowseDirectoryStack)))
	{
		free(ILibPopStack(&currBrowseDirectoryStack));
	}
}

static void handleCmd_pwd()
{
	void *tempStack;
	char *tempStr;

	//There is two kinds of objectID, one provide full path,
	//the other just provide raletive path
	if(strncmp("0", ILibPeekStack(&currBrowseDirectoryStack), 1) == 0)
	{//full path
		tempStr = ILibPeekStack(&currBrowseDirectoryStack);
		ignoreRoot(&tempStr);
		changeDelimiter(tempStr, DEFAULTDELIMITER);
		printf("%sweb%s%c", CurrMServerURI, tempStr, DEFAULTDELIMITER);	
	}
	else{//relative path
		ILibCreateStack(&tempStack);
		while(ILibPeekStack(&currBrowseDirectoryStack))
		{
			ILibPushStack(&tempStack, ILibPopStack(&currBrowseDirectoryStack));
		}


		printf("%sweb", CurrMServerURI);
		while(ILibPeekStack(&tempStack))
		{
			tempStr = ILibPopStack(&tempStack);
			ILibPushStack(&currBrowseDirectoryStack, tempStr);
			ignoreRoot(&tempStr);
			changeDelimiter(tempStr, DEFAULTDELIMITER);
			printf("%s%c", tempStr, DEFAULTDELIMITER);				
		}
		ILibClearStack(&tempStack);
	}
	printf("\n");	
}



static void handleCmd_help()
{
	printf("\n\nNone of the commands require arguments.  The commands with a * \n");
	printf("signify ones that will prompt the user for more information once\n");
	printf("the command is called\n\n");
	printf("The available commands are:\n\n");
	printf(" quit    -   shutdown the control point\n");
	printf(" exit    -   same as quit\n");
	printf(" setms   - * select a media server to become the active media server\n");
	printf(" getms   -   print the friendly name of the active media server\n");
	printf(" ls      -   list the contents of the current directory on the active \n");
	printf("             media server\n");
	printf(" cd      - * traverse down one level in the content tree on the active\n");
	printf("             media server\n");
	printf(" cdup    -   traverse up one level in the content tree on the active\n");
	printf("             media server\n");
	printf(" pwd     -   print the http path from the root to your current position in the \n");
	printf("             content tree on the active media server\n");
	printf(" get     - * get an item (HTTP only)\n");
	printf(" play    - * play a link (needs mono)\n");
	printf(" help    -   print this help message\n\n");
}



int CLShell_Run(void)
{
	int status = -1;

	tmp_buffer[0] = '\0';
#ifdef WIN32
	//_CrtCheckMemory();
#endif

	fflush(stdout);
	fgets(tmp_buffer, TMP_BUFFER_SIZE, stdin);
	StrChomp(tmp_buffer);

	if(0 == strcmp(tmp_buffer, "quit") || 0 == strcmp(tmp_buffer, "exit"))
	{
		/* TODO do something to quit pollAgain = 0; */
		return -1;
	}
	else if(0 == strcmp(tmp_buffer, "setms"))
	{
		handleCmd_setms();
	}
	else if(0 == strcmp(tmp_buffer, "getms"))
	{
		handleCmd_getms();
	}
	else if(0 == strcmp(tmp_buffer, "ls"))
	{
		status = DoBrowse(handleCmd_ls);
	}
	else if(0 == strcmp(tmp_buffer, "cd"))
	{
		status = DoBrowse(handleCmd_cd);
	}
	else if(0 == strcmp(tmp_buffer, "cdup"))
	{
		handleCmd_cdup();
	}
	else if(0 == strcmp(tmp_buffer, "pwd"))
	{
		handleCmd_pwd();
	}
	else if(0 == strcmp(tmp_buffer, "help"))
	{
		handleCmd_help();
	}
	else if(0 == strcmp(tmp_buffer, "get"))
	{
		status = DoBrowse(handleCmd_get);
	}
	else if(0 == strcmp(tmp_buffer, "play"))
	{
		status = DoBrowse(handleCmd_play);
	}
	else if(0 == strcmp(tmp_buffer, ""))
	{
		// just prompt again
	}
	else
	{
		printf("Unrecognized command: %s\n", tmp_buffer);
		handleCmd_help();
	}

	if (status < 0){
		/* This means it was not a command requiring I/O and callback
		 * or that a browse request failed, in any case, there won't
		 * be any callback to print the prompt for us, so print it */
		prompt();
	}

	return 0;
}

void CLShell_Init(void *chain)
{
	// create the stack that will be the directory where the
	// user is currently browsing. 
	ILibCreateStack(&currBrowseDirectoryStack);

	// push the root directory onto the directory stack.
	ILibPushStack(&currBrowseDirectoryStack, MakeStrcpy("0"));

	// Initially there is no server and no renderer selected
	currMServerUDN = MakeStrcpy(NULL);

	// the most recent results from a browse.  This has to be a global
	// variable since the results come back through a callback function
	// instead of being returned to the requesting function.
	mostRecentBrowseResults = NULL;

	// the table of known devices on the network.  These are updated via the
	// device added and device removed callbacks.  Note that you should lock them
	// whenever accessing them with the ILibHashTree_Lock and ILibHashTree_UnLock methods
	MS_KnownDevicesTable = ILibInitHashTree();


	// Add the control points to the main chain, and keep pointers to them for use 
	// in the API calls.
	MS_ControlPoint = MMSCP_Init(chain, MSCP_BrowseResponded, MSCP_DeviceAddedOrRemoved);
	prompt();
}


void CLShell_Uninit(void)
{
	// destroy the response wait semaphore.

	// clear off the directory stack by first popping down so that just 
	// the root directory is on it.
	popDirectoryStackToRoot();
	// then free the root directory
	free(ILibPopStack(&currBrowseDirectoryStack));
	// finally, destroy the stack itself
	ILibClearStack(&currBrowseDirectoryStack);

	free(currMServerUDN);

	if(mostRecentBrowseResults != NULL) 
	{
		free(mostRecentBrowseResults);
		mostRecentBrowseResults = NULL;
	}

	// destroy the lists of known nodes and null out their pointers.
	// note tat the ILibDestroyHashTree function calls free on both the 
	// key and the data values stored in the table.
	ILibDestroyHashTree(MS_KnownDevicesTable);
	MS_KnownDevicesTable = NULL;
}
