/*
 *
 * Copyright (c) Sigma Designs, Inc. 2005. All rights reserved.
 *
 */
#ifndef __MS_CARDEA_H__
#define __MS_CARDEA_H__

#include "rmdef/rmdef.h"

/**
 * Extended analog video protection
 */
struct cardea_extended_analog_video_s {
	RMuint8 guid[16];
	RMuint32 size;
	RMuint8 *data;
};

/**
 * Ouput protection level
 */
struct cardea_opl_s {
	RMuint32 min_digital_uncompressed_video;
	RMuint32 min_digital_uncompressed_audio;
	RMuint32 min_analog_video;
	RMuint32 extended_analog_video_size; /* How many extensions */
	struct cardea_extended_analog_video_s *extended_analog_video_opl;
};


/**
 * Initialize CARDEA for a given URL
 *
 * This function tries to get the CARDEA session associated with the URL. If no
 * session is opened for this URL, returns NULL.
 *
 * @param url - url for which a session header is required
 * @return NULL if no session header is available
 */
void * init_cardea(const RMascii *url);


/**
 * Terminate CARDEA
 *
 * This function handles the termination of the CARDEA session, if any.
 * 
 * @return RM_OK on success
 */
RMstatus term_cardea(void);


/**
 * Update the sample ID for a particuler stream and media object
 *
 * @param cardea_context - the context to update
 * @param stream_index - current stream index
 * @param media_object_number - current media object number
 * @param sample_id - pointer to the sample ID data
 * @param sample_id_size - size of the sample id (should be 8)
 */
void update_cardea_sample_id(
		void * cardea_context, 
		RMuint16 stream_index, 
		RMuint32 media_object_number, 
		RMuint8 *sample_id, 
		RMuint32 sample_id_size);

/**
 * Decrypt cardea encrypted data (in place)
 *
 * @param cardea_context - the current cardea context
 * @param stream_index - current stream index
 * @param media_object_number - current media object number
 * @param buf - pointer to the data to decrypt
 * @param size - size of the data to decrypt
 * @return RM_OK on success
 *
 */
RMstatus decrypt_cardea_sample(
		void * cardea_context,
		RMuint16 stream_index,
		RMuint32 media_object_number,
		RMuint8 *buf,
		RMuint32 size);

/**
 * Get the session ID header for a given URL
 *
 * This function tries to get the session header associated with a URL. If no
 * MS DRM session header is required, or if a license has not been obtained,
 * this function will return NUL.
 *
 * @param url - url for which a session header is required
 * @return NULL if no session header is available
 */
RMascii * get_ms_session_header(const RMascii *url);


/**
 * Get the output protection level for a given cardea context
 *
 * @param cardea_context - the current cardea context
 * @param opl - pointer to the structure that should be
 *    filled with the output protection level data
 * @return RM_OK on success
 */
RMstatus get_cardea_opl(
		void * cardea_context,
		struct cardea_opl_s *opl);


#endif /* __MS_CARDEA_H__ */
