/*
 *
 * Copyright (c) Sigma Designs, Inc. 2005. All rights reserved.
 *
 */
#ifndef __MS_UPNP_H__
#define __MS_UPNP_H__

#include "UPnPControlPointStructs.h"
#include "rmdef/rmdef.h"

/* Flags for device status */
#define MS_ISAUTHORIZED 	1	/* The server has authorized us */
#define MS_WAITREGISTRATION	2	/* We are waiting for the server registration response */
#define MS_ISVALIDATED	 	4	/* We are registered to the server */
#define MS_WAITPROXYMITY	8	/* We are waiting for proximity detection to finish */
#define MS_PROXYMITYOK		16	/* We are have been detected as near by the server */
#define MS_FORCEREGISTER	32	/* Force registration, even if already registered */

/* Access MS UPnP extension */ 
#define MS_UPNP(device)		((ms_upnp_extension*)(device->ms_upnp))

/* WMDRM errors */
/* The DMR's certificate is corrupted or its signature cannot be verified. */
#define DRM_INVALID_CERTIFICATE  		100
/* A hash of the DMR's certificate is listed in the transmitter's Certificate
 * Revocation List. */
#define	DRM_CERTIFICATE_REVOKED  		101
/* The DMR must be approved before it can execute the intended operation. */
#define	DRM_MUST_APPROVE  			102
/* The transmitter does not hold a license for the requested action. */
#define	DRM_LICENSE_UNAVAILABLE  		103
/* The transmitter encountered a generic failure. */
#define	DRM_TRANSMITTER_FAILURE  		104
/* The transmitter cannot execute the requested action because the number of
 * devices in use has already reached the maximum permitted. */
#define	DRM_DEVICE_LIMIT_REACHED  		105
/* The proximity detection procedure could not determine that the DMR is near
 * the transmitter. */
#define	DRM_UNABLE_TO_VERIFY_PROXIMITY  	106
/* The DMR must be registered before it can execute the intended operation. */
#define	DRM_MUST_REGISTER  			107
/* The DMR must be revalidated before it can execute the intended operation. */
#define	DRM_MUST_REVALIDATE  			108
/* The DMR has responded to the proximity challenge message incorrectly. */
#define	DRM_INVALID_PROXIMITY_RESPONSE  	109
/* The session is no longer available. */
#define	DRM_INVALID_SESSION  			110
/* The transmitter cannot open the requested file. */
#define	DRM_UNABLE_TO_OPEN_FILE  		111
/* The request received by the transmitter is based on an unsupported protocol
 * version. */
#define	DRM_UNSUPPORTED_PROTOCOL_VERSION  	112
/* The format of the request is not valid. */
#define	DRM_BAD_REQUEST  			113

struct ms_upnp_extension_s;
typedef struct ms_upnp_extension_s ms_upnp_extension;

/**
 * Initialize internals
 */
void init_ms_upnp(void);

/**
 * Initialize a MS UPnP server
 * @param device
 */
void init_ms_device(struct UPnPDevice *device);


/**
 * Destroy a MS UPnP server
 * @param device
 */
void destroy_ms_device(struct UPnPDevice *device);


/**
 * Get the status of the device with a MS server
 * @param device
 */
RMuint32 ms_device_status(struct UPnPDevice *device);

/*
 * Try to register the client to the given device
 * @param device - device to register to
 * @return < 0 on error
 */
RMint32 register_ms_client(struct UPnPDevice *device);

/**
 * Get a license for a given URL, on the given device
 * @param device
 * @param uri - URL to get
 * @param session_header - returns the session header for this URL, should not
 *                         be freed or modified.
 * @return 0 on success, < 0 on error
 */
RMint32 get_ms_license(
		struct UPnPDevice *device, 
		RMascii *url, 
		RMascii **session_header);

/**
 * Cleanup internals
 */
void destroy_ms_upnp(void);

#endif /* __MS_UPNP_H__ */
