/*
 *
 * Copyright (c) Sigma Designs, Inc. 2005. All rights reserved.
 *
 */
#define ALLOW_OS_CODE 1
#include "rmdef/rmdef.h"
#include "rmcore/include/rmcore.h"
#include "../include/ms_upnp.h"
#include "ms_upnp_types.h"
#include "ms_upnp_event.h"
#include "wmdrmndcoreapi.h"
#include "../include/ms_cardea.h"

/** If this is set to 1, it will force the device to register once, at least */
static RMuint32 force_registration = 0;

/** All the knwon ms_device, keep track of them, also used by ms_cardea.h, so not static */
extern struct ms_upnp_extension_s *ms_device[MAX_MS_DEVICE];

#ifdef WITH_CARDEA_DIRECT_API
/* Testing purpose */
extern void init_ms_client(void);
#endif

/**
 * Initialize internals
 */
void init_ms_upnp(void)
{
	RMuint32 i;
	
	/* Init events internals */
	init_ms_event();
	
#ifdef WITH_CARDEA_DIRECT_API
	/* Init our client internals, done by the daemon */
	init_ms_client();
#endif

	for (i=0;i<MAX_MS_DEVICE; i++)
		ms_device[i] = NULL;
}

/*
 * Initialize a MS UPnP server
 * @param device
 */
void init_ms_device(struct UPnPDevice *device)
{
	RMuint32 i;
	
	device->ms_upnp = RMMalloc(sizeof(struct ms_upnp_extension_s));
	MS_UPNP(device)->flags = 0;
	MS_UPNP(device)->drm_context = init_wmdrmnet_context();
	if (MS_UPNP(device)->drm_context == NULL){
		RMDBGLOG((ENABLE,"Error creating DRM context for %s\n", device->FriendlyName));
	} else {
		if (force_registration){
			RMDBGLOG((ENABLE,"Force registration on %s\n", device->FriendlyName));
			MS_UPNP(device)->flags |= MS_FORCEREGISTER;
		}
	}
	MS_UPNP(device)->session_header = NULL;
	MS_UPNP(device)->url = NULL;

	/* Register this MS device */
	for (i=0;i<MAX_MS_DEVICE;i++){
		if (ms_device[i] == NULL){
			ms_device[i] = MS_UPNP(device);
			break;
		}
	}
}

/*
 * Destroy a MS UPnP server
 * @param device
 */
void destroy_ms_device(struct UPnPDevice *device)
{
	RMuint32 i;

	if ((device != NULL) && (MS_UPNP(device) != NULL)){
		if (MS_UPNP(device)->drm_context != NULL)
			destroy_wmdrmnet_context(MS_UPNP(device)->drm_context);
		if (MS_UPNP(device)->session_header != NULL)
			RMFree(MS_UPNP(device)->session_header);
		if (MS_UPNP(device)->url != NULL)
			RMFree(MS_UPNP(device)->url);
	
		/* Unregister this MS device */
		for (i=0;i<MAX_MS_DEVICE;i++){
			if (ms_device[i] == MS_UPNP(device)){
				ms_device[i] = NULL;
				break;
			}
		}
		
		RMFree(MS_UPNP(device));
	}
}

/*
 * Get the status of the device with a MS server
 * @param device
 */
RMuint32 ms_device_status(struct UPnPDevice *device)
{
	if ((device != NULL) && (MS_UPNP(device) != NULL))
		return MS_UPNP(device)->flags;
	else
		return 0;
}


