/*
 *
 * Copyright (c) Sigma Designs, Inc. 2005. All rights reserved.
 *
 */
#define ALLOW_OS_CODE 1
#include "rmdef/rmdef.h"
#include "../include/ms_upnp.h"
#include "ms_upnp_event.h"
#include "MSCP_ControlPoint.h"
#include "UPnPControlPointStructs.h"
#include <stdio.h>
#include "ms_upnp_types.h"

#if 0
#define MSEVENTDBG ENABLE
#else
#define MSEVENTDBG DISABLE
#endif

static void callback_registrar_is_validated(struct UPnPService* service, int error, void *user, int result)
{
	struct UPnPDevice *device;
	RMuint32 is_validated;
	
	if (error != 0){
		RMDBGLOG((ENABLE,"Error in is_validated : %d\n", error));
	} else {
		device = service->Parent;
		if (device != NULL && MS_UPNP(device) != NULL){
			is_validated = (MS_UPNP(device)->flags & MS_ISVALIDATED) ? 1 : 0;
			if ((unsigned)result != is_validated){
				/* Our status changed */
				if (result){
					MS_UPNP(device)->flags = MS_UPNP(device)->flags | MS_ISAUTHORIZED;
					fprintf(stderr,"We have been validated by %s\n",device->FriendlyName);
					MS_UPNP(device)->flags |= MS_ISVALIDATED;
					MS_UPNP(device)->flags &= ~MS_WAITREGISTRATION;
				} else {
					MS_UPNP(device)->flags = MS_UPNP(device)->flags & (~MS_ISAUTHORIZED);
					fprintf(stderr,"We have been unvalidated by %s\n",device->FriendlyName);
					MS_UPNP(device)->flags &= ~MS_ISVALIDATED;
				}
			}
			/* RMDBGLOG((ENABLE,"We are%svalidated by %s\n", result ? " " : " not ", device->FriendlyName)); */
			if (result == 0 || (MS_UPNP(device)->flags & MS_FORCEREGISTER) ){
				/* Not registered to this server, check if
				 * registration is already in progress */
				if ((MS_UPNP(device)->flags & MS_WAITREGISTRATION) == 0){
					RMint32 register_status;
					fprintf(stderr,"Try to register with %s\n", device->FriendlyName);
					register_status = register_ms_client(device);
					if (register_status < 0)
						fprintf(stderr,"Error registering at %s : %ld\n", device->FriendlyName, register_status);
					else
						fprintf(stderr,"Registration at %s started.\n", device->FriendlyName);
				} else {
					RMDBGLOG((ENABLE,"Registration pending with %s\n", device->FriendlyName));
				}
			}
		} else {
			fprintf(stderr,"Error, received a registrar callback for an unknown MS device\n");
		}
	}
}

static void callback_registrar_is_authorized(struct UPnPService* service, int error, void *user, int result)
{
	struct UPnPDevice *device;
	RMuint32 is_authorized;

	if (error != 0){
		fprintf(stderr,"Error in is_authorized : %d\n", error);
	} else {
		device = service->Parent;
		if (device != NULL && MS_UPNP(device) != NULL){
			is_authorized = ms_device_status(device) & MS_ISAUTHORIZED;
			if ((unsigned)result != is_authorized){
				/* Our status changed */
				if (result){
					MS_UPNP(device)->flags = MS_UPNP(device)->flags | MS_ISAUTHORIZED;
					fprintf(stderr,"We have been granted authorization by %s\n",device->FriendlyName);
					/* Check if we need to validate again */
					MSCP_Invoke_Registrar_IsValidated(service, callback_registrar_is_validated, NULL, "");
				} else {
					MS_UPNP(device)->flags = MS_UPNP(device)->flags & (~MS_ISAUTHORIZED);
					fprintf(stderr,"We have been denied authorization by %s\n",device->FriendlyName);
				}
			}
		} else {
			fprintf(stderr,"Error, received a registrar callback for an unknown MS device\n");
		}
	}
}

static void callback_registrar_authorization(struct UPnPService* service, unsigned int value)
{
	RMDBGLOG((MSEVENTDBG,"Authorization Event received (%d)\n", value));
	MSCP_Invoke_Registrar_IsAuthorized(service, callback_registrar_is_authorized, NULL, "");
}

static void callback_registrar_validation(struct UPnPService* service, unsigned int value)
{
	RMDBGLOG((MSEVENTDBG,"Validation Event received (%d)\n", value));
	MSCP_Invoke_Registrar_IsValidated(service, callback_registrar_is_validated, NULL, "");
}

void init_ms_event(void)
{
	RMDBGLOG((MSEVENTDBG,"Initialize MS registrar events\n"));
	MSCP_EventCallback_Registrar_AuthorizationDeniedUpdateID = callback_registrar_authorization;
	MSCP_EventCallback_Registrar_AuthorizationGrantedUpdateID = callback_registrar_authorization;
	MSCP_EventCallback_Registrar_ValidationSucceededUpdateID = callback_registrar_validation;
	MSCP_EventCallback_Registrar_ValidationRevokedUpdateID = callback_registrar_validation;
}

