/*
 * INTEL CONFIDENTIAL
 * Copyright (c) 2002, 2003 Intel Corporation.  All rights reserved.
 * 
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel
 * Corporation or its suppliers or licensors.  Title to the
 * Material remains with Intel Corporation or its suppliers and
 * licensors.  The Material contains trade secrets and proprietary
 * and confidential information of Intel or its suppliers and
 * licensors. The Material is protected by worldwide copyright and
 * trade secret laws and treaty provisions.  No part of the Material
 * may be used, copied, reproduced, modified, published, uploaded,
 * posted, transmitted, distributed, or disclosed in any way without
 * Intel's prior express written permission.
 
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you
 * by disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise. Any license
 * under such intellectual property rights must be express and
 * approved by Intel in writing.
 * 
 * $Workfile: ILibAsyncServerSocket.c
 * $Revision: #1.0.1799.42459
 * $Author:   Intel Corporation, Intel Device Builder
 * $Date:     Wednesday, January 19, 2005
 *
 *
 *
 */

#if defined(_WIN32_WCE)
	#define _CRTDBG_MAP_ALLOC
	#include <math.h>
	#include <winerror.h>
	#include <stdlib.h>
	#include <stdio.h>
	#include <stddef.h>
	#include <string.h>
	#include <winsock.h>
	#include <wininet.h>
	#include <windows.h>
	#include <winioctl.h>
	#include <winbase.h>
#elif defined(WIN32)
	#define _CRTDBG_MAP_ALLOC
	#include <math.h>
	#include <winerror.h>
	#include <stdlib.h>
	#include <stdio.h>
	#include <stddef.h>
	#include <string.h>
	#ifdef WINSOCK2
		#include <winsock2.h>
		#include <ws2tcpip.h>
	#else
		#include <winsock.h>
		#include <wininet.h>
	#endif
	#include <windows.h>
	#include <winioctl.h>
	#include <winbase.h>
	#include <crtdbg.h>
#elif defined(_POSIX)
	#include <stdio.h>
	#include <stdlib.h>
	#include <sys/types.h>
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <arpa/inet.h>
	#include <sys/time.h>
	#include <netdb.h>
	#include <string.h>
	#include <sys/ioctl.h>
	#include <net/if.h>
	#include <sys/utsname.h>
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <unistd.h>
	#include <fcntl.h>
	#include <errno.h>
#endif

#include "ILibAsyncServerSocket.h"
#include "ILibAsyncSocket.h"
#include "ILibParsers.h"

#if defined(WIN32) && !defined(_WIN32_WCE)
	#include <errno.h>
#elif defined(_POSIX)
	#include <errno.h>
	#include <semaphore.h>
#endif


#define DEBUGSTATEMENT(x)


struct AsyncServerSocketModule
{
	void (*PreSelect)(void* object,fd_set *readset, fd_set *writeset, fd_set *errorset, int* blocktime);
	void (*PostSelect)(void* object,int slct, fd_set *readset, fd_set *writeset, fd_set *errorset);
	void (*Destroy)(void* object);
	void *Chain;

	int MaxConnection;
	void **AsyncSockets;
	ILibServerScope scope;
	
	#ifdef _WIN32_WCE
		SOCKET ListenSocket;
	#elif defined(WIN32)
		SOCKET ListenSocket;
	#elif defined(_POSIX)
		int ListenSocket;
	#endif
	unsigned short portNumber;
	int listening;

	void (*OnReceive)(void *AsyncServerSocketModule, void *ConnectionToken,char* buffer,int *p_beginPointer, int endPointer, void (**OnInterrupt)(void *AsyncServerSocketMoudle, void *ConnectionToken, void *user),void **user, int *PAUSE);
	void (*OnConnect)(void *AsyncServerSocketModule, void *ConnectionToken,void **user);
	void (*OnDisconnect)(void *AsyncServerSocketModule, void *ConnectionToken, void *user);
	void (*OnInterrupt)(void *AsyncServerSocketModule, void *ConnectionToken, void *user);
	void (*OnSendOK)(void *AsyncServerSocketModule, void *ConnectionToken, void *user);

	void *Tag;
};
struct AsyncServerSocket_Data
{
	struct AsyncServerSocketModule *module;
	ILibAsyncServerSocket_BufferReAllocated Callback;
	void *user;
};

/// <summary>
/// Returns the user Tag associated with the AsyncServer
/// </summary>
/// <param name="AsyncSocketModule">The ILibAsyncServerSocket to query</param>
/// <returns>The user Tag</returns>
void *ILibAsyncServerSocket_GetTag(void *AsyncSocketModule)
{
	struct AsyncServerSocketModule *module = (struct AsyncServerSocketModule*)AsyncSocketModule;
	return(module->Tag);
}
/// <summary>
/// Sets the user Tag associated with the AsyncServer
/// </summary>
/// <param name="AsyncSocketModule">The ILibAsyncServerSocket to save the tag to</param>
void ILibAsyncServerSocket_SetTag(void *AsyncSocketModule, void *tag)
{
	struct AsyncServerSocketModule *module = (struct AsyncServerSocketModule*)AsyncSocketModule;
	module->Tag = tag;
}

/// <summary>
/// Internal method called by ILibAsyncSocket, to signal an interrupt condition
/// </summary>
/// <param name="socketModule">The ILibAsyncServerSocket that was interrupted</param>
/// <param name="user">The associated user tag</param>
static void ILibAsyncServerSocket_OnInterrupt(void *socketModule, void *user)
{
	struct AsyncServerSocket_Data *data = (struct AsyncServerSocket_Data*)user;
	if(data->module->OnInterrupt!=NULL)
	{
		data->module->OnInterrupt(data->module,socketModule,data->user);
	}
	free(user);
}
/// <summary>
/// Chain PreSelect handler
/// </summary>
/// <param name="socketModule"></param>
/// <param name="readset"></param>
/// <param name="writeset"></param>
/// <param name="errorset"></param>
/// <param name="blocktime"></param>
static void ILibAsyncServerSocket_PreSelect(void* socketModule,fd_set *readset, fd_set *writeset, fd_set *errorset, int* blocktime)
{
	struct AsyncServerSocketModule *module = (struct AsyncServerSocketModule*)socketModule;
	int flags,i;

	//
	// The socket isn't put in listening mode, until the chain is started.
	// If this variable==0, that means we need to do that.
	//
	if(module->listening==0)
	{
		//
		// Set the socket to non-block mode, so we can play nice and share the thread
		//
		#ifdef _WIN32_WCE
			flags = 1;
			ioctlsocket(module->ListenSocket,FIONBIO,&flags);
		#elif defined(WIN32)
			flags = 1;
			ioctlsocket(module->ListenSocket,FIONBIO,&flags);
		#elif defined(_POSIX)
			flags = fcntl(module->ListenSocket,F_GETFL,0);
			fcntl(module->ListenSocket,F_SETFL,O_NONBLOCK|flags);
		#endif
		
		//
		// Put the socket in Listen, and add it to the fdset for the Select loop
		//
		module->listening=1;
		listen(module->ListenSocket,4);
		FD_SET(module->ListenSocket,readset);
	}
	else
	{
		// Only put the ListenSocket in the readset, if we are able to handle a new socket
		for(i=0;i<module->MaxConnection;++i)
		{
			if(ILibAsyncSocket_IsFree(module->AsyncSockets[i])!=0)
			{
				FD_SET(module->ListenSocket,readset);
				break;
			}
		}
	}
}
void ILibAsyncServerSocket_SetReAllocateNotificationCallback(void *AsyncServerSocketToken, void *ConnectionToken, ILibAsyncServerSocket_BufferReAllocated Callback)
{
	struct AsyncServerSocket_Data *data = (struct AsyncServerSocket_Data*)ILibAsyncSocket_GetUser(ConnectionToken);
	data->Callback = Callback;
}
/// <summary>
/// Chain PostSelect handler
/// </summary>
/// <param name="socketModule"></param>
/// <param name="slct"></param>
/// <param name="readset"></param>
/// <param name="writeset"></param>
/// <param name="errorset"></param>
static void ILibAsyncServerSocket_PostSelect(void* socketModule,int slct, fd_set *readset, fd_set *writeset, fd_set *errorset)
{
	struct AsyncServerSocket_Data *data;
	struct sockaddr_in addr;
	int addrlen;
	struct AsyncServerSocketModule *module = (struct AsyncServerSocketModule*)socketModule;
	int i,flags;
	struct sockaddr_in receivingAddress;
	int receivingAddressLength = sizeof(struct sockaddr_in);
	#ifdef _WIN32_WCE
		SOCKET NewSocket;
	#elif defined(WIN32)
		SOCKET NewSocket;
	#elif defined(_POSIX)
		int NewSocket;
	#endif

	if(FD_ISSET(module->ListenSocket,readset)!=0)
	{
		//
		// There are pending TCP connection requests
		//
		for(i=0;i<module->MaxConnection;++i)
		{
			//
			// Check to see if we have available resources to handle this connection request
			//
			if(ILibAsyncSocket_IsFree(module->AsyncSockets[i])!=0)
			{
				addrlen = sizeof(addr);
				NewSocket = accept(module->ListenSocket,(struct sockaddr*)&addr,&addrlen);
				if(NewSocket!= ~0)
				{
				switch(module->scope)
				{
					case ILibServerScope_LocalLoopback:
						// Check that the caller ip address is the same as the receive IP address
						getsockname(NewSocket,(struct sockaddr*)&receivingAddress,&receivingAddressLength);
						if(receivingAddress.sin_addr.s_addr!=addr.sin_addr.s_addr)
						{
							#if defined(WIN32) || defined(_WIN32_WCE)
								closesocket(NewSocket);
							#else
								close(NewSocket);
							#endif
								NewSocket=~0;
						}
						break;
					case ILibServerScope_LocalSegment:
						getsockname(NewSocket,(struct sockaddr*)&receivingAddress,&receivingAddressLength);
						break;
					default:
						break;
					}
				}
				if (NewSocket != ~0)
				{
					//
					// Set this new socket to non-blocking mode, so we can play nice and share thread
					//
					#ifdef _WIN32_WCE
						flags = 1;
						ioctlsocket(NewSocket,FIONBIO,&flags);
					#elif defined(WIN32)
						flags = 1;
						ioctlsocket(NewSocket,FIONBIO,&flags);
					#elif defined(_POSIX)
						flags = fcntl(NewSocket,F_GETFL,0);
						fcntl(NewSocket,F_SETFL,O_NONBLOCK|flags);
					#endif
					//
					// Instantiate a module to contain all the data about this connection
					//
					data = (struct AsyncServerSocket_Data*)malloc(sizeof(struct AsyncServerSocket_Data));
					memset(data,0,sizeof(struct AsyncServerSocket_Data));
					data->module = socketModule;

					ILibAsyncSocket_UseThisSocket(module->AsyncSockets[i],&NewSocket,&ILibAsyncServerSocket_OnInterrupt,data);
					ILibAsyncSocket_SetRemoteAddress(module->AsyncSockets[i],addr.sin_addr.s_addr);

					//
					// Notify the user about this new connection
					//
					if(module->OnConnect!=NULL)
					{
						module->OnConnect(module,module->AsyncSockets[i],&(data->user));
					}
				}
				else {break;}
			}
		}
	}
}
/// <summary>
/// Chain Destroy handler
/// </summary>
/// <param name="socketModule"></param>
static void ILibAsyncServerSocket_Destroy(void *socketModule)
{
	struct AsyncServerSocketModule *module =(struct AsyncServerSocketModule*)socketModule;

	free(module->AsyncSockets);
	#ifdef _WIN32_WCE
		closesocket(module->ListenSocket);
	#elif defined(WIN32)
		closesocket(module->ListenSocket);
	#elif defined(_POSIX)
		close(module->ListenSocket);
	#endif	
}
/// <summary>
/// Internal method dispatched by the OnData event of the underlying ILibAsyncSocket
/// </summary>
/// <param name="socketModule"></param>
/// <param name="buffer"></param>
/// <param name="p_beginPointer"></param>
/// <param name="endPointer"></param>
/// <param name="OnInterrupt"></param>
/// <param name="user"></param>
/// <param name="PAUSE"></param>
static void ILibAsyncServerSocket_OnData(void* socketModule,char* buffer,int *p_beginPointer, int endPointer,void (**OnInterrupt)(void *AsyncSocketMoudle, void *user),void **user, int *PAUSE)
{
	struct AsyncServerSocket_Data *data = (struct AsyncServerSocket_Data*)(*user);
	int bpointer = *p_beginPointer;

	//
	// Pass the received data up
	//
	if(data->module->OnReceive!=NULL)
	{
		data->module->OnReceive(data->module,socketModule,buffer,&bpointer,endPointer,&(data->module->OnInterrupt),&(data->user),PAUSE);
		*p_beginPointer = bpointer;
	}
}
/// <summary>
/// Internal method dispatched by the OnDisconnect event of the underlying ILibAsyncSocket
/// </summary>
/// <param name="socketModule"></param>
/// <param name="user"></param>
static void ILibAsyncServerSocket_OnDisconnect(void* socketModule, void *user)
{
	struct AsyncServerSocket_Data *data = (struct AsyncServerSocket_Data*)user;

	//
	// Pass this Disconnect event up
	//
	if(data->module->OnDisconnect!=NULL)
	{
		data->module->OnDisconnect(data->module,socketModule,data->user);
	}

	//
	// If the chain is shutting down, we need to free some resources
	//
	if(ILibIsChainBeingDestroyed(data->module->Chain)==0)
	{
		free(data);
	}
}
/// <summary>
/// Internal method dispatched by the OnSendOK event of the underlying ILibAsyncSocket
/// </summary>
/// <param name="socketModule"></param>
/// <param name="user"></param>
static void ILibAsyncServerSocket_OnSendOK(void* socketModule, void *user)
{
	struct AsyncServerSocket_Data *data = (struct AsyncServerSocket_Data*)user;

	//
	// Pass the OnSendOK event up
	//
	if(data->module->OnSendOK!=NULL)
	{
		data->module->OnSendOK(data->module,socketModule,data->user);
	}
}
/// <summary>
/// Internal method dispatched by ILibAsyncSocket, to signal that the buffers have been reallocated
/// </summary>
/// <param name="ConnectionToken">The ILibAsyncSocket sender</param>
/// <param name="user">The AsyncServerSocket_Data object</param>
/// <param name="offSet">The offset to the new buffer location</param>
static void ILibAsyncServerSocket_OnBufferReAllocated(void *ConnectionToken, void *user, ptrdiff_t offSet)
{
	struct AsyncServerSocket_Data *data = (struct AsyncServerSocket_Data*)user;
	if(data->Callback!=NULL)
	{
		//
		// If someone above us, has registered for this callback, we need to fire it,
		// with the correct user object
		//
		data->Callback(data->module,ConnectionToken,data->user,offSet);
	}
}

/// <summary>
/// Internal method dispatched by the OnSendOK event of the underlying ILibAsyncSocket
/// </summary>
/// <param name="Chain">The chain to add this module to</param>
/// <param name="MaxConnections">The max number of simultaneous connections that will be allowed</param>
/// <param name="PortNumber">The port number to bind to. 0 will select a random port</param>
/// <param name="initialBufferSize">The initial size of the receive buffer</param>
/// <param name="OnConnect">Function Pointer that triggers when a connection is established</param>
/// <param name="OnDisconnect">Function Pointer that triggers when a connection is closed</param>
/// <param name="OnReceive">Function Pointer that triggers when data is received</param>
/// <param name="OnInterrupt">Function Pointer that triggers when connection interrupted</param>
/// <param name="OnSendOK">Function Pointer that triggers when pending sends are complete</param>
/// <returns>An ILibAsyncServerSocket module</returns>
void *ILibCreateAsyncServerSocketModule(void *Chain, int MaxConnections, int PortNumber, int initialBufferSize, void (*OnConnect)(void *AsyncServerSocketModule, void *ConnectionToken,void **user),void (*OnDisconnect)(void *AsyncServerSocketModule, void *ConnectionToken, void *user),void (*OnReceive)(void *AsyncServerSocketModule, void *ConnectionToken,char* buffer,int *p_beginPointer, int endPointer,void (**OnInterrupt)(void *AsyncServerSocketMoudle, void *ConnectionToken, void *user), void **user, int *PAUSE),void (*OnInterrupt)(void *AsyncServerSocketModule, void *ConnectionToken, void *user), void (*OnSendOK)(void *AsyncServerSocketModule, void *ConnectionToken, void *user))
{
	struct AsyncServerSocketModule *RetVal;
	int i;

	//
	// Instantiate a new AsyncServer module
	//
	RetVal = (struct AsyncServerSocketModule*)malloc(sizeof(struct AsyncServerSocketModule));
	memset(RetVal,0,sizeof(struct AsyncServerSocketModule));
	RetVal->PreSelect = &ILibAsyncServerSocket_PreSelect;
	RetVal->PostSelect = &ILibAsyncServerSocket_PostSelect;
	RetVal->Destroy = &ILibAsyncServerSocket_Destroy;
	RetVal->Chain = Chain;
	RetVal->OnConnect = OnConnect;
	RetVal->OnDisconnect = OnDisconnect;
	RetVal->OnInterrupt = OnInterrupt;
	RetVal->OnSendOK = OnSendOK;
	RetVal->OnReceive = OnReceive;
	RetVal->MaxConnection = MaxConnections;
	RetVal->AsyncSockets = (void**)malloc(MaxConnections*sizeof(void*));
	RetVal->portNumber = PortNumber;
	
	//
	// Create our socket pool
	//
	for(i=0;i<MaxConnections;++i)
	{
		RetVal->AsyncSockets[i] = ILibCreateAsyncSocketModule(Chain,initialBufferSize,&ILibAsyncServerSocket_OnData,NULL,&ILibAsyncServerSocket_OnDisconnect, &ILibAsyncServerSocket_OnSendOK);
		//
		// We want to know about any buffer reallocations, because anything above us may want to know
		//
		ILibAsyncSocket_SetReAllocateNotificationCallback(RetVal->AsyncSockets[i],&ILibAsyncServerSocket_OnBufferReAllocated);
	}
	ILibAddToChain(Chain,RetVal);

	//
	// Get our listening socket
	//
	#ifdef WINSOCK2
		RetVal->portNumber = ILibGetStreamSocket(htonl(INADDR_ANY),RetVal->portNumber,(HANDLE*)&(RetVal->ListenSocket));
	#else
		RetVal->portNumber = ILibGetStreamSocket(htonl(INADDR_ANY),RetVal->portNumber,&(RetVal->ListenSocket));
	#endif

	return(RetVal);
}
/// <summary>
/// Returns the port number the server is bound to
/// </summary>
/// <param name="ServerSocketModule">The ILibAsyncServer to query</param>
/// <returns>The listening port number</returns>
unsigned short ILibAsyncServerSocket_GetPortNumber(void *ServerSocketModule)
{
	struct AsyncServerSocketModule *ASSM = (struct AsyncServerSocketModule*)ServerSocketModule;
	return(ASSM->portNumber);
}
