/*
* INTEL CONFIDENTIAL
* Copyright (c) 2002, 2003 Intel Corporation.  All rights reserved.
* 
* The source code contained or described herein and all documents
* related to the source code ("Material") are owned by Intel
* Corporation or its suppliers or licensors.  Title to the
* Material remains with Intel Corporation or its suppliers and
* licensors.  The Material contains trade secrets and proprietary
* and confidential information of Intel or its suppliers and
* licensors. The Material is protected by worldwide copyright and
* trade secret laws and treaty provisions.  No part of the Material
* may be used, copied, reproduced, modified, published, uploaded,
* posted, transmitted, distributed, or disclosed in any way without
* Intel's prior express written permission.

* No license under any patent, copyright, trade secret or other
* intellectual property right is granted to or conferred upon you
* by disclosure or delivery of the Materials, either expressly, by
* implication, inducement, estoppel or otherwise. Any license
* under such intellectual property rights must be express and
* approved by Intel in writing.
* 
* $Workfile: MSCP_ControlPoint.c
* $Revision: #1.0.1799.42459
* $Author:   Intel Corporation, Intel Device Builder
* $Date:     Monday, March 21, 2005
*
*
*
*/


#if defined(WIN32) || defined(_WIN32_WCE)
#	ifndef MICROSTACK_NO_STDAFX
#		include "stdafx.h"
#	endif
#	define _CRTDBG_MAP_ALLOC
#	include <winerror.h>
char* MSCP_PLATFORM = "WINDOWS, UPnP/1.0, Intel MicroStack/1.0.1799";
#else
char* MSCP_PLATFORM = "POSIX, UPnP/1.0, Intel MicroStack/1.0.1799";
#endif


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#if defined(WIN32) || defined(_WIN32_WCE)
#	include <stddef.h>
#	if defined(WINSOCK2)
#	include <winsock2.h>
#	include <ws2tcpip.h>
#	elif defined(WINSOCK1)
#	include <winsock.h>
#	include <wininet.h>
#	endif
#	include <windows.h>
#	include <winioctl.h>
#	include <winbase.h>
#	include <crtdbg.h>
#	define strncasecmp(x,y,z) _strnicmp(x,y,z)
#	define gettimeofday(x,y) (x)->tv_sec = GetTickCount()/1000;(x)->tv_usec = 1000*(GetTickCount()%1000)
#else
#	include <sys/types.h>
#	include <sys/socket.h>
#	include <netinet/in.h>
#	include <arpa/inet.h>
#	include <sys/time.h>
#	include <netdb.h>
#	include <sys/ioctl.h>
#	include <net/if.h>
#	include <sys/utsname.h>
#	include <sys/socket.h>
#	include <netinet/in.h>
#	include <unistd.h>
#	include <fcntl.h>
#	include <errno.h>
#	include <semaphore.h>
#endif

#include "ILibParsers.h"
#include "ILibSSDPClient.h"
#include "ILibWebServer.h"
#include "ILibWebClient.h"
#include "ILibAsyncSocket.h"
#include "MSCP_ControlPoint.h"

#define UPNP_PORT 1900
#define UPNP_GROUP "239.255.255.250"
#define MSCP_MIN(a,b) (((a)<(b))?(a):(b))

#include "ILibParsers.h"
#include "ILibSSDPClient.h"
#include "ILibWebServer.h"
#include "ILibWebClient.h"
#include "MSCP_ControlPoint.h"

#define INVALID_DATA 0
#define DEBUGSTATEMENT(x)
#define LVL3DEBUG(x)

static const char *UPNPCP_SOAP_Header = "POST %s HTTP/1.1\r\nHost: %s:%d\r\nUser-Agent: %s, UPnP/1.0, Intel MicroStack/1.0.1799\r\nSOAPACTION: \"%s#%s\"\r\nContent-Type: text/xml; charset=\"utf-8\"\r\nContent-Length: %d\r\n\r\n";
static const char *UPNPCP_SOAP_BodyHead = "<?xml version=\"1.0\" encoding=\"utf-8\"?><s:Envelope s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\" xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\"><s:Body><u:";
static const char *UPNPCP_SOAP_BodyTail = "></s:Body></s:Envelope>";

void MSCP_Renew(void *state);
void MSCP_SSDP_Sink(void *sender, char* UDN, int Alive, char* LocationURL, int Timeout,UPnPSSDP_MESSAGE m, void *cp);

struct CustomUserData
{
	int Timeout;
	char* buffer;
};
struct MSCP_CP
{
	void (*PreSelect)(void* object,fd_set *readset, fd_set *writeset, fd_set *errorset, int* blocktime);
	void (*PostSelect)(void* object,int slct, fd_set *readset, fd_set *writeset, fd_set *errorset);
	void (*Destroy)(void* object);
	void (*DiscoverSink)(struct UPnPDevice *device);
	void (*RemoveSink)(struct UPnPDevice *device);
	
	//ToDo: Check to see if this is necessary
	void (*EventCallback_DimmingService_LoadLevelStatus)(struct UPnPService* Service,unsigned char value);
	void (*EventCallback_SwitchPower_Status)(struct UPnPService* Service,int value);
	
	struct LifeTimeMonitorStruct *LifeTimeMonitor;
	
	void *HTTP;
	void *SSDP;
	void *WebServer;
	
	void* SIDTable;
	void* DeviceTable;
	
	void *Chain;
	int RecheckFlag;
	int AddressListLength;
	int *AddressList;
};

void (*MSCP_EventCallback_ContentDirectory_ContainerUpdateIDs)(struct UPnPService* Service,char* value);
void (*MSCP_EventCallback_ContentDirectory_SystemUpdateID)(struct UPnPService* Service,unsigned int value);
void (*MSCP_EventCallback_ContentDirectory_TransferIDs)(struct UPnPService* Service,char* value);
void (*MSCP_EventCallback_Registrar_AuthorizationDeniedUpdateID)(struct UPnPService* Service,unsigned int value);
void (*MSCP_EventCallback_Registrar_ValidationSucceededUpdateID)(struct UPnPService* Service,unsigned int value);
void (*MSCP_EventCallback_Registrar_ValidationRevokedUpdateID)(struct UPnPService* Service,unsigned int value);
void (*MSCP_EventCallback_Registrar_AuthorizationGrantedUpdateID)(struct UPnPService* Service,unsigned int value);


struct InvokeStruct
{
	struct UPnPService *Service;
	void *CallbackPtr;
	void *User;
};
struct UPnPServiceInfo
{
	char* serviceType;
	char* SCPDURL;
	char* controlURL;
	char* eventSubURL;
	char* serviceId;
	struct UPnPServiceInfo *Next;
};
struct MSCP__Stack
{
	void *data;
	struct MSCP__Stack *next;
};


void MSCP_CP_ProcessDeviceRemoval(struct MSCP_CP* CP, struct UPnPDevice *device);



static void MSCP_DestructUPnPService(struct UPnPService *service)
{
	struct UPnPAction *a1,*a2;
	struct UPnPStateVariable *sv1,*sv2;
	int i;
	
	a1 = service->Actions;
	while(a1!=NULL)
	{
		a2 = a1->Next;
		free(a1->Name);
		free(a1);
		a1 = a2;
	}
	
	sv1 = service->Variables;
	while(sv1!=NULL)
	{
		sv2 = sv1->Next;
		free(sv1->Name);
		if(sv1->Min!=NULL) {free(sv1->Min);}
		if(sv1->Max!=NULL) {free(sv1->Max);}
		if(sv1->Step!=NULL) {free(sv1->Step);}
		for(i=0;i<sv1->NumAllowedValues;++i)
		{
			free(sv1->AllowedValues[i]);
		}
		if(sv1->AllowedValues!=NULL) {free(sv1->AllowedValues);}
		free(sv1);
		sv1 = sv2;
	}
	if(service->ControlURL!=NULL) {free(service->ControlURL);}
	if(service->SCPDURL!=NULL) {free(service->SCPDURL);}
	if(service->ServiceId!=NULL) {free(service->ServiceId);}
	if(service->ServiceType!=NULL) {free(service->ServiceType);}
	if(service->SubscriptionURL!=NULL) {free(service->SubscriptionURL);}
	if(service->SubscriptionID!=NULL)
	{
		ILibLifeTime_Remove(((struct MSCP_CP*)service->Parent->CP)->LifeTimeMonitor,service);
		ILibDeleteEntry(((struct MSCP_CP*)service->Parent->CP)->SIDTable,service->SubscriptionID,(int)strlen(service->SubscriptionID));
		free(service->SubscriptionID);
		service->SubscriptionID = NULL;
	}
	
	free(service);
}
static void MSCP_DestructUPnPDevice(struct UPnPDevice *device)
{
	struct UPnPDevice *d1,*d2;
	struct UPnPService *s1,*s2;
	
	void *en,*en2;
	char *key;
	int keyLen;
	void *val,*ht2;
	
	
	d1 = device->EmbeddedDevices;
	while(d1!=NULL)
	{
		d2 = d1->Next;
		MSCP_DestructUPnPDevice(d1);
		d1 = d2;
	}
	
	s1 = device->Services;
	while(s1!=NULL)
	{
		s2 = s1->Next;
		MSCP_DestructUPnPService(s1);
		s1 = s2;
	}
	
	LVL3DEBUG(printf("\r\n\r\nDevice Destructed\r\n");)
	if(device->PresentationURL!=NULL) {free(device->PresentationURL);}
	if(device->ManufacturerName!=NULL) {free(device->ManufacturerName);}
	if(device->ManufacturerURL!=NULL) {free(device->ManufacturerURL);}
	if(device->ModelName!=NULL) {free(device->ModelName);}
	if(device->ModelNumber!=NULL) {free(device->ModelNumber);}
	if(device->ModelURL!=NULL) {free(device->ModelURL);}
	if(device->ModelDescription!=NULL) {free(device->ModelDescription);}
	if(device->DeviceType!=NULL) {free(device->DeviceType);}
	if(device->FriendlyName!=NULL) {free(device->FriendlyName);}
	if(device->LocationURL!=NULL) {free(device->LocationURL);}
	if(device->UDN!=NULL) {free(device->UDN);}
	if(device->InterfaceToHost!=NULL) {free(device->InterfaceToHost);}
	if(device->Reserved3!=NULL) {free(device->Reserved3);}
	
	
	if(device->CustomTagTable != NULL) 
	{
		en = ILibHashTree_GetEnumerator(device->CustomTagTable);
		while(ILibHashTree_MoveNext(en)==0)
		{
			ILibHashTree_GetValue(en,&key,&keyLen,&ht2);
			en2 = ILibHashTree_GetEnumerator(ht2);
			while(ILibHashTree_MoveNext(en2)==0)
			{
				ILibHashTree_GetValue(en2,&key,&keyLen,&val);
				if(val!=NULL)
				{
					free(val);
				}
			}
			ILibHashTree_DestroyEnumerator(en2);
			ILibDestroyHashTree(ht2);
		}
		ILibHashTree_DestroyEnumerator(en);
		ILibDestroyHashTree(device->CustomTagTable);
	}
	
	
	free(device);
}

void MSCP_AddRef(struct UPnPDevice *device)
{
	struct UPnPDevice *d = device;
	while(d->Parent!=NULL) {d = d->Parent;}
	++d->ReferenceCount;
}
void MSCP_Release(struct UPnPDevice *device)
{
	struct UPnPDevice *d = device;
	
	while(d->Parent!=NULL) {d = d->Parent;}
	--d->ReferenceCount;
	if(d->ReferenceCount<=0)
	{
		MSCP_DestructUPnPDevice(d);
	}
}
//void UPnPDeviceDescriptionInterruptSink(void *sender, void *user1, void *user2)
//{
	//	struct CustomUserData *cd = (struct CustomUserData*)user1;
	//	free(cd->buffer);
	//	free(user1);
	//}
#if 0
static int MSCP_IsLegacyDevice(struct packetheader *header)
{
	struct packetheader_field_node *f;
	struct parser_result_field *prf;
	struct parser_result *r,*r2;
	int Legacy=1;
	// Check version of Device
	f = header->FirstField;
	while(f!=NULL)
	{
		if(f->FieldLength==6 && strncasecmp(f->Field,"SERVER",6)==0)
		{
			// Check MSCP_ version of the Control Point which invoked us
			r = ILibParseString(f->FieldData,0,f->FieldDataLength," ",1);
			prf = r->FirstResult;
			while(prf!=NULL)
			{
				if(prf->datalength>5 && memcmp(prf->data,"UPnP/",5)==0)
				{
					r2 = ILibParseString(prf->data+5,0,prf->datalength-5,".",1);
					r2->FirstResult->data[r2->FirstResult->datalength]=0;
					r2->LastResult->data[r2->LastResult->datalength]=0;
					if(atoi(r2->FirstResult->data)==1 && atoi(r2->LastResult->data)>0)
					{
						Legacy=0;
					}
					ILibDestructParserResults(r2);
				}
				prf = prf->NextResult;
			}
			ILibDestructParserResults(r);
		}
		if(Legacy)
		{
			f = f->NextField;
		}
		else
		{
			break;
		}
	}
	return(Legacy);
}
static void MSCP_Push(struct MSCP__Stack **pp_Top, void *data)
{
	struct MSCP__Stack *frame = (struct MSCP__Stack*)malloc(sizeof(struct MSCP__Stack));
	frame->data = data;
	frame->next = *pp_Top;
	*pp_Top = frame;
}
static void *MSCP_Pop(struct MSCP__Stack **pp_Top)
{
	struct MSCP__Stack *frame = *pp_Top;
	void *RetVal = NULL;
	
	if(frame!=NULL)
	{
		*pp_Top = frame->next;
		RetVal = frame->data;
		free(frame);
	}
	return(RetVal);
}
static void *MSCP_Peek(struct MSCP__Stack **pp_Top)
{
	struct MSCP__Stack *frame = *pp_Top;
	void *RetVal = NULL;
	
	if(frame!=NULL)
	{
		RetVal = (*pp_Top)->data;
	}
	return(RetVal);
}
static void MSCP_Flush(struct MSCP__Stack **pp_Top)
{
	while(MSCP_Pop(pp_Top)!=NULL) {}
	*pp_Top = NULL;
}
#endif
struct UPnPDevice* MSCP_GetDeviceAtUDN(void *v_CP,char* UDN)
{
	struct UPnPDevice *RetVal = NULL;
	struct MSCP_CP* CP = (struct MSCP_CP*)v_CP;
	
	ILibHashTree_Lock(CP->DeviceTable);
	RetVal = (struct UPnPDevice*)ILibGetEntry(CP->DeviceTable,UDN,(int)strlen(UDN));
	ILibHashTree_UnLock(CP->DeviceTable);
	return(RetVal);
}
static struct packetheader *MSCP_BuildPacket(char* IP, int Port, char* Path, char* cmd)
{
	struct packetheader *RetVal = ILibCreateEmptyPacket();
	char* HostLine = (char*)malloc((int)strlen(IP)+7);
	int HostLineLength = sprintf(HostLine,"%s:%d",IP,Port);
	
	ILibSetVersion(RetVal,"1.1",3);
	ILibSetDirective(RetVal,cmd,(int)strlen(cmd),Path,(int)strlen(Path));
	ILibAddHeaderLine(RetVal,"Host",4,HostLine,HostLineLength);
	ILibAddHeaderLine(RetVal,"User-Agent",10,MSCP_PLATFORM,(int)strlen(MSCP_PLATFORM));
	free(HostLine);
	return(RetVal);
}

/*static void MSCP_RemoveServiceFromDevice(struct UPnPDevice *device, struct UPnPService *service)
{
	struct UPnPService *s = device->Services;
	
	if(s==service)
	{
		device->Services = s->Next;
		MSCP_DestructUPnPService(service);
		return;
	}
	while(s->Next!=NULL)
	{
		if(s->Next == service)
		{
			s->Next = s->Next->Next;
			MSCP_DestructUPnPService(service);
			return;
		}
		s = s->Next;
	}
}*/

static void MSCP_ProcessDevice(void *v_cp,struct UPnPDevice *device)
{
	struct MSCP_CP* cp = (struct MSCP_CP*)v_cp;
	struct UPnPDevice *EmbeddedDevice = device->EmbeddedDevices;
	while(EmbeddedDevice!=NULL)
	{
		MSCP_ProcessDevice(v_cp,EmbeddedDevice);
		EmbeddedDevice = EmbeddedDevice->Next;
	}
	
	// Create a table entry for each Unique Device Name, for easy mapping
	// of ssdp:byebye packets. This way any byebye packet from the device
	// heirarchy will result in the device being removed.
	
	ILibHashTree_Lock(cp->DeviceTable);
	ILibAddEntry(cp->DeviceTable,device->UDN,(int)strlen(device->UDN),device);
	ILibHashTree_UnLock(cp->DeviceTable);
}

static void MSCP_PrintUPnPDevice(int indents, struct UPnPDevice *device)
{
	struct UPnPService *s;
	struct UPnPDevice *d;
	struct UPnPAction *a;
	int x=0;
	
	for(x=0;x<indents;++x) {printf(" ");}
	printf("Device: %s\r\n",device->DeviceType);
	
	for(x=0;x<indents;++x) {printf(" ");}
	printf("Friendly: %s\r\n",device->FriendlyName);
	
	s = device->Services;
	while(s!=NULL)
	{
		for(x=0;x<indents;++x) {printf(" ");}
		printf("   Service: %s\r\n",s->ServiceType);
		a = s->Actions;
		while(a!=NULL)
		{
			for(x=0;x<indents;++x) {printf(" ");}
			printf("      Action: %s\r\n",a->Name);
			a = a->Next;
		}
		s = s->Next;
	}
	
	d = device->EmbeddedDevices;
	while(d!=NULL)
	{
		MSCP_PrintUPnPDevice(indents+5,d);
		d = d->Next;
	}
}
struct UPnPService *MSCP_GetService(struct UPnPDevice *device, char* ServiceName, int length)
{
	struct UPnPService *RetService = NULL;
	struct UPnPService *s = device->Services;
	struct parser_result *pr,*pr2;
	
	pr = ILibParseString(ServiceName,0,length,":",1);
	while(s!=NULL)
	{
		pr2 = ILibParseString(s->ServiceType,0,(int)strlen(s->ServiceType),":",1);
		if(length-pr->LastResult->datalength >= (int)strlen(s->ServiceType)-pr2->LastResult->datalength && memcmp(ServiceName,s->ServiceType,length-pr->LastResult->datalength)==0)
		{
			if(atoi(pr->LastResult->data)<=atoi(pr2->LastResult->data))
			{
				RetService = s;
				ILibDestructParserResults(pr2);
				break;
			}
		}
		ILibDestructParserResults(pr2);
		s = s->Next;
	}
	ILibDestructParserResults(pr);
	
	return(RetService);
}

struct UPnPService *MSCP_GetService_ContentDirectory(struct UPnPDevice *device)
{
	return(MSCP_GetService(device,"urn:schemas-upnp-org:service:ContentDirectory:1",47));
}
struct UPnPService *MSCP_GetService_Registrar(struct UPnPDevice *device)
{
	return(MSCP_GetService(device,"urn:microsoft.com:service:X_MS_MediaReceiverRegistrar:1",55));
}


static struct UPnPDevice *MSCP_GetDevice2(struct UPnPDevice *device, int index, int *c_Index)
{
	struct UPnPDevice *RetVal = NULL;
	struct UPnPDevice *e_Device = NULL;
	int currentIndex = *c_Index;
	
	if(strncmp(device->DeviceType,"urn:schemas-upnp-org:device:MediaServer:",40)==0 && atoi(device->DeviceType+40)>=1)
	
	{
		++currentIndex;
		if(currentIndex==index)
		{
			*c_Index = currentIndex;
			return(device);
		}
	}
	
	e_Device = device->EmbeddedDevices;
	while(e_Device!=NULL)
	{
		RetVal = MSCP_GetDevice2(e_Device,index,&currentIndex);
		if(RetVal!=NULL)
		{
			break;
		}
		e_Device = e_Device->Next;
	}
	
	*c_Index = currentIndex;
	return(RetVal);
}
struct UPnPDevice* MSCP_GetDevice1(struct UPnPDevice *device,int index)
{
	int c_Index = -1;
	return(MSCP_GetDevice2(device,index,&c_Index));
}
int MSCP_GetDeviceCount(struct UPnPDevice *device)
{
	int RetVal = 0;
	struct UPnPDevice *e_Device = device->EmbeddedDevices;
	
	while(e_Device!=NULL)
	{
		RetVal += MSCP_GetDeviceCount(e_Device);
		e_Device = e_Device->Next;
	}
	if(strncmp(device->DeviceType,"urn:schemas-upnp-org:device:MediaServer:1",41)==0)
	{
		++RetVal;
	}
	
	return(RetVal);
}

static int MSCP_GetErrorCode(char *buffer, int length)
{
	int RetVal = 500;
	struct ILibXMLNode *xml,*rootXML;
	
	char *temp;
	int tempLength;
	
	rootXML = xml = ILibParseXML(buffer,0,length);
	ILibProcessXMLNodeList(xml);
	
	while(xml!=NULL)
	{
		if(xml->NameLength==8 && memcmp(xml->Name,"Envelope",8)==0)
		{
			xml = xml->Next;
			while(xml!=NULL)
			{
				if(xml->NameLength==4 && memcmp(xml->Name,"Body",4)==0)
				{
					xml = xml->Next;
					while(xml!=NULL)
					{
						if(xml->NameLength==5 && memcmp(xml->Name,"Fault",5)==0)
						{
							xml = xml->Next;
							while(xml!=NULL)
							{
								if(xml->NameLength==6 && memcmp(xml->Name,"detail",6)==0)
								{
									xml = xml->Next;
									while(xml!=NULL)
									{
										if(xml->NameLength==9 && memcmp(xml->Name,"MSCP_Error",9)==0)
										{
											xml = xml->Next;
											while(xml!=NULL)
											{
												if(xml->NameLength==9 && memcmp(xml->Name,"errorCode",9)==0)
												{
													tempLength = ILibReadInnerXML(xml,&temp);
													temp[tempLength] = 0;
													RetVal =atoi(temp);
													xml = NULL;
												}
												if(xml!=NULL) {xml = xml->Peer;}
											}
										}
										if(xml!=NULL) {xml = xml->Peer;}
									}
								}
								if(xml!=NULL) {xml = xml->Peer;}
							}
						}
						if(xml!=NULL) {xml = xml->Peer;}
					}
				}
				if(xml!=NULL) {xml = xml->Peer;}
			}
		}
		if(xml!=NULL) {xml = xml->Peer;}
	}
	ILibDestructXMLNodeList(rootXML);
	return(RetVal);
}
static void MSCP_ProcessSCPD(char* buffer, int length, struct UPnPService *service)
{
	struct UPnPAction *action;
	struct UPnPStateVariable *sv = NULL;
	struct UPnPAllowedValue *av = NULL;
	struct UPnPAllowedValue *avs = NULL;
	
	struct ILibXMLNode *xml,*rootXML;
	int flg2,flg3,flg4;
	
	char* tempString;
	int tempStringLength;
	
	struct UPnPDevice *root = service->Parent;
	while(root->Parent!=NULL) {root = root->Parent;}
	
	rootXML = xml = ILibParseXML(buffer,0,length);
	if(ILibProcessXMLNodeList(xml)!=0)
	{
		root->SCPDError=UPNP_ERROR_SCPD_NOT_WELL_FORMED;
		ILibDestructXMLNodeList(rootXML);
		return;
	}
	
	while(xml!=NULL && strncmp(xml->Name,"!",1)==0)
	{
		xml = xml->Next;
	}
	xml = xml->Next;
	while(xml!=NULL)
	{
		if(xml->NameLength==10 && memcmp(xml->Name,"actionList",10)==0)
		{
			xml = xml->Next;
			flg2 = 0;
			while(flg2==0)
			{
				if(xml->NameLength==6 && memcmp(xml->Name,"action",6)==0)
				{
					action = (struct UPnPAction*)malloc(sizeof(struct UPnPAction));
					action->Name = NULL;
					action->Next = service->Actions;
					service->Actions = action;
					
					xml = xml->Next;
					flg3 = 0;
					while(flg3==0)
					{
						if(xml->NameLength==4 && memcmp(xml->Name,"name",4)==0)
						{
							tempStringLength = ILibReadInnerXML(xml,&tempString);
							action->Name = (char*)malloc(1+tempStringLength);
							memcpy(action->Name,tempString,tempStringLength);
							action->Name[tempStringLength] = '\0';
						}
						if(xml->Peer==NULL)
						{
							flg3 = -1;
							xml = xml->Parent;
						}
						else
						{
							xml = xml->Peer;
						}
					}
				}
				if(xml->Peer==NULL)
				{
					flg2 = -1;
					xml = xml->Parent;
				}
				else
				{
					xml = xml->Peer;
				}
			}
		}
		if(xml->NameLength==17 && memcmp(xml->Name,"serviceStateTable",17)==0)
		{
			if(xml->Next->StartTag!=0)
			{
				xml = xml->Next;
				flg2 = 0;
				while(flg2==0)
				{
					if(xml->NameLength==13 && memcmp(xml->Name,"stateVariable",13)==0)
					{
						sv = (struct UPnPStateVariable*)malloc(sizeof(struct UPnPStateVariable));
						sv->AllowedValues = NULL;
						sv->NumAllowedValues = 0;
						sv->Max = NULL;
						sv->Min = NULL;
						sv->Step = NULL;
						sv->Name = NULL;
						sv->Next = service->Variables;
						service->Variables = sv;
						sv->Parent = service;
						
						xml = xml->Next;
						flg3 = 0;
						while(flg3==0)
						{
							if(xml->NameLength==4 && memcmp(xml->Name,"name",4)==0)
							{
								tempStringLength = ILibReadInnerXML(xml,&tempString);
								sv->Name = (char*)malloc(1+tempStringLength);
								memcpy(sv->Name,tempString,tempStringLength);
								sv->Name[tempStringLength] = '\0';
							}
							if(xml->NameLength==16 && memcmp(xml->Name,"allowedValueList",16)==0)
							{
								if(xml->Next->StartTag!=0)
								{
									avs = NULL;
									xml = xml->Next;
									flg4 = 0;
									while(flg4==0)
									{
										if(xml->NameLength==12 && memcmp(xml->Name,"allowedValue",12)==0)
										{
											av = (struct UPnPAllowedValue*)malloc(sizeof(struct UPnPAllowedValue));
											av->Next = avs;
											avs = av;
											
											tempStringLength = ILibReadInnerXML(xml,&tempString);
											av->Value = (char*)malloc(1+tempStringLength);
											memcpy(av->Value,tempString,tempStringLength);
											av->Value[tempStringLength] = '\0';
										}
										if(xml->Peer!=NULL)
										{
											xml = xml->Peer;
										}
										else
										{
											xml = xml->Parent;
											flg4 = -1;
										}
									}
									av = avs;
									while(av!=NULL)
									{
										++sv->NumAllowedValues;
										av = av->Next;
									}
									av = avs;
									sv->AllowedValues = (char**)malloc(sv->NumAllowedValues*sizeof(char*));
									for(flg4=0;flg4<sv->NumAllowedValues;++flg4)
									{
										sv->AllowedValues[flg4] = av->Value;
										av = av->Next;
									}
									av = avs;
									while(av!=NULL)
									{
										avs = av->Next;
										free(av);
										av = avs;
									}
								}
							}
							if(xml->NameLength==17 && memcmp(xml->Name,"allowedValueRange",17)==0)
							{
								if(xml->Next->StartTag!=0)
								{
									xml = xml->Next;
									flg4 = 0;
									while(flg4==0)
									{
										if(xml->NameLength==7)
										{
											if(memcmp(xml->Name,"minimum",7)==0)
											{
												tempStringLength = ILibReadInnerXML(xml,&tempString);
												sv->Min = (char*)malloc(1+tempStringLength);
												memcpy(sv->Min,tempString,tempStringLength);
												sv->Min[tempStringLength] = '\0';
											}
											else if(memcmp(xml->Name,"maximum",7)==0)
											{
												tempStringLength = ILibReadInnerXML(xml,&tempString);
												sv->Max = (char*)malloc(1+tempStringLength);
												memcpy(sv->Max,tempString,tempStringLength);
												sv->Max[tempStringLength] = '\0';
											}
										}
										if(xml->NameLength==4 && memcmp(xml->Name,"step",4)==0)
										{
											tempStringLength = ILibReadInnerXML(xml,&tempString);
											sv->Step = (char*)malloc(1+tempStringLength);
											memcpy(sv->Step,tempString,tempStringLength);
											sv->Step[tempStringLength] = '\0';
										}
										if(xml->Peer!=NULL)
										{
											xml = xml->Peer;
										}
										else
										{
											xml = xml->Parent;
											flg4 = -1;
										}
									}
								}
							}
							if(xml->Peer!=NULL)
							{
								xml = xml->Peer;
							}
							else
							{
								flg3 = -1;
								xml = xml->Parent;
							}
						}
					}
					if(xml->Peer!=NULL)
					{
						xml = xml->Peer;
					}
					else
					{
						xml = xml->Parent;
						flg2 = -1;
					}
				}
			}
		}
		xml = xml->Peer;
	}
	
	ILibDestructXMLNodeList(rootXML);
}
static void MSCP_InterfaceChanged(struct UPnPDevice *device)
{
	void *cp = device->CP;
	char *UDN;
	char *LocationURL;
	int Timeout;
	
	Timeout = device->CacheTime;
	UDN = (char*)malloc((int)strlen(device->UDN)+1);
	strcpy(UDN,device->UDN);
	LocationURL = device->Reserved3;
	device->Reserved3 = NULL;
	
	MSCP_SSDP_Sink(NULL, device->UDN, 0, NULL, 0,UPnPSSDP_NOTIFY,device->CP);
	MSCP_SSDP_Sink(NULL, UDN, -1, LocationURL, Timeout, UPnPSSDP_NOTIFY,cp);
	
	free(UDN);
	free(LocationURL);
}
static void MSCP_ExpiredDevice(struct UPnPDevice *device)
{
	LVL3DEBUG(printf("Device[%s] failed to re-advertise in a timely manner\r\n",device->FriendlyName);)
	while(device->Parent!=NULL) {device = device->Parent;}
	MSCP_SSDP_Sink(NULL, device->UDN, 0, NULL, 0,UPnPSSDP_NOTIFY,device->CP);
}
static void MSCP_FinishProcessingDevice(struct MSCP_CP* CP, struct UPnPDevice *RootDevice)
{
	int Timeout = RootDevice->CacheTime;
	struct UPnPDevice *RetDevice;
	int i=0;
	
	RootDevice->ReferenceCount = 1;
	do
	{
		RetDevice = MSCP_GetDevice1(RootDevice,i++);
		if(RetDevice!=NULL)
		{
			RetDevice->Reserved=1;
			MSCP_AddRef(RetDevice);
			if(CP->DiscoverSink!=NULL)
			{
				CP->DiscoverSink(RetDevice);
			}
		}
	}while(RetDevice!=NULL);
	ILibLifeTime_Add(CP->LifeTimeMonitor,RootDevice,Timeout,&MSCP_ExpiredDevice,NULL);
}
static void MSCP_SCPD_Sink(
void *WebReaderToken,
int IsInterrupt,
struct packetheader *header,
char *buffer,
int *p_BeginPointer,
int EndPointer,
int done,
void *dv,
void *sv,
int *PAUSE)
{
	struct UPnPDevice *device;
	struct UPnPService *service = (struct UPnPService*)sv;
	struct MSCP_CP *CP = service->Parent->CP;
	
	if(!(header==NULL || header->StatusCode!=200) && done!=0)
	{
		MSCP_ProcessSCPD(buffer,EndPointer, service);
		
		device = service->Parent;
		while(device->Parent!=NULL)
		{
			device = device->Parent;
		}
		--device->SCPDLeft;
		if(device->SCPDLeft==0)
		{
			if(device->SCPDError==0)
			{
				MSCP_FinishProcessingDevice(CP,device);
			}
			else if(IsInterrupt==0)
			{
				MSCP_CP_ProcessDeviceRemoval(CP,device);
				MSCP_DestructUPnPDevice(device);
			}
		}
	}
	else
	{
		if(done!=0 && (header==NULL || header->StatusCode!=200))
		{
			device = service->Parent;
			while(device->Parent!=NULL)
			{
				device = device->Parent;
			}
			--device->SCPDLeft;
			device->SCPDError=1;
			if(device->SCPDLeft==0 && IsInterrupt==0)
			{
				MSCP_CP_ProcessDeviceRemoval(CP,device);
				MSCP_DestructUPnPDevice(device);
			}
		}
	}
}
static void MSCP_CalculateSCPD_FetchCount(struct UPnPDevice *device)
{
	int count = 0;
	struct UPnPDevice *root;
	struct UPnPDevice *e_Device = device->EmbeddedDevices;
	struct UPnPService *s;
	
	while(e_Device!=NULL)
	{
		MSCP_CalculateSCPD_FetchCount(e_Device);
		e_Device = e_Device->Next;
	}
	
	s = device->Services;
	while(s!=NULL)
	{
		++count;
		s = s->Next;
	}
	
	root = device;
	while(root->Parent!=NULL)
	{
		root = root->Parent;
	}
	root->SCPDLeft += count;
}
static void MSCP_SCPD_Fetch(struct UPnPDevice *device)
{
	struct UPnPDevice *e_Device = device->EmbeddedDevices;
	struct UPnPService *s;
	char *IP,*Path;
	unsigned short Port;
	struct packetheader *p;
	struct sockaddr_in addr;
	
	while(e_Device!=NULL)
	{
		MSCP_SCPD_Fetch(e_Device);
		e_Device = e_Device->Next;
	}
	
	s = device->Services;
	while(s!=NULL)
	{
		ILibParseUri(s->SCPDURL,&IP,&Port,&Path);
		DEBUGSTATEMENT(printf("SCPD: %s Port: %d Path: %s\r\n",IP,Port,Path));
		p = MSCP_BuildPacket(IP,Port,Path,"GET");
		
		memset((char *)&addr, 0,sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = inet_addr(IP);
		addr.sin_port = htons(Port);
		
		ILibWebClient_PipelineRequest(
		((struct MSCP_CP*)device->CP)->HTTP,
		&addr,
		p,
		&MSCP_SCPD_Sink,
		device,
		s);
		
		free(IP);
		free(Path);
		s = s->Next;
	}
}

char *MSCP_GetCustomTagFromDevice(struct UPnPDevice *d, char* FullNameSpace, char* Name)
{
	if(ILibHasEntry(d->CustomTagTable,FullNameSpace,(int)strlen(FullNameSpace))!=0)
	{
		return(ILibGetEntry(ILibGetEntry(d->CustomTagTable,FullNameSpace,(int)strlen(FullNameSpace)),Name,(int)strlen(Name)));
	}
	else
	{
		return(NULL);
	}
}

static struct UPnPDevice* MSCP_ProcessDeviceXML_device(struct ILibXMLNode *xml, void *v_CP,const char *BaseURL, int Timeout, int RecvAddr)
{
	struct ILibXMLNode *tempNode;
	int flg,flg2;
	char *tempString;
	int tempStringLength;
	
	char *tempString2;
	
	struct parser_result *tpr;
	
	char* ServiceType = NULL;
	int ServiceTypeLength = 0;
	char* SCPDURL = NULL;
	int SCPDURLLength = 0;
	char* EventSubURL = NULL;
	int EventSubURLLength = 0;
	char* ControlURL = NULL;
	int ControlURLLength = 0;
	
	struct UPnPDevice *tempDevice;
	struct UPnPService *TempService;
	struct UPnPDevice *device = (struct UPnPDevice*)malloc(sizeof(struct UPnPDevice));
	memset(device,0,sizeof(struct UPnPDevice));
	
	device->CustomTagTable = ILibInitHashTree();
	
	
	device->CP = v_CP;
	device->CacheTime = Timeout;
	device->InterfaceToHost = (char*)malloc(16);
	sprintf(device->InterfaceToHost,"%d.%d.%d.%d",(RecvAddr&0xFF),((RecvAddr>>8)&0xFF),((RecvAddr>>16)&0xFF),((RecvAddr>>24)&0xFF));
	
	xml = xml->Next;
	while(xml!=NULL)
	{
		if(xml->NameLength==10 && memcmp(xml->Name,"deviceList",10)==0)
		{
			if(xml->Next->StartTag!=0)
			{
				xml = xml->Next;
				flg2 = 0;
				while(flg2==0)
				{
					if(xml->NameLength==6 && memcmp(xml->Name,"device",6)==0)
					{
						tempDevice = MSCP_ProcessDeviceXML_device(xml,v_CP,BaseURL,Timeout, RecvAddr);
						tempDevice->Parent = device;
						tempDevice->Next = device->EmbeddedDevices;
						device->EmbeddedDevices = tempDevice;
					}
					if(xml->Peer==NULL)
					{
						flg2 = 1;
						xml = xml->Parent;
					}
					else
					{
						xml = xml->Peer;
					}
				}
			}
		} else
		if(xml->NameLength==3 && memcmp(xml->Name,"UDN",3)==0)
		{
			tempStringLength = ILibReadInnerXML(xml,&tempString);
			if(tempStringLength>5)
			{
				if(memcmp(tempString,"uuid:",5)==0)
				{
					tempString += 5;
					tempStringLength -= 5;
				}
				device->UDN = (char*)malloc(tempStringLength+1);
				memcpy(device->UDN,tempString,tempStringLength);
				device->UDN[tempStringLength] = '\0';
			}
		} else
		if(xml->NameLength==10 && memcmp(xml->Name,"deviceType",10) == 0)
		{
			tempStringLength = ILibReadInnerXML(xml,&tempString);
			
			device->DeviceType = (char*)malloc(tempStringLength+1);
			memcpy(device->DeviceType,tempString,tempStringLength);
			device->DeviceType[tempStringLength] = '\0';
		} else
		if(xml->NameLength==12 && memcmp(xml->Name,"friendlyName",12) == 0)
		{
			tempStringLength = ILibReadInnerXML(xml,&tempString);
			device->FriendlyName = (char*)malloc(1+tempStringLength);
			memcpy(device->FriendlyName,tempString,tempStringLength);
			device->FriendlyName[tempStringLength] = '\0';
		} else
		if(xml->NameLength==12 && memcmp(xml->Name,"manufacturer",12) == 0)
		{
			tempStringLength = ILibReadInnerXML(xml,&tempString);
			device->ManufacturerName = (char*)malloc(1+tempStringLength);
			memcpy(device->ManufacturerName,tempString,tempStringLength);
			device->ManufacturerName[tempStringLength] = '\0';
		} else
		if(xml->NameLength==15 && memcmp(xml->Name,"manufacturerURL",15) == 0)
		{
			tempStringLength = ILibReadInnerXML(xml,&tempString);
			device->ManufacturerURL = (char*)malloc(1+tempStringLength);
			memcpy(device->ManufacturerURL,tempString,tempStringLength);
			device->ManufacturerURL[tempStringLength] = '\0';
		} else
		if(xml->NameLength==16 && memcmp(xml->Name,"modelDescription",16) == 0)
		{
			tempStringLength = ILibReadInnerXML(xml,&tempString);
			device->ModelDescription = (char*)malloc(1+tempStringLength);
			memcpy(device->ModelDescription,tempString,tempStringLength);
			device->ModelDescription[tempStringLength] = '\0';
		} else
		if(xml->NameLength==9 && memcmp(xml->Name,"modelName",9) == 0)
		{
			tempStringLength = ILibReadInnerXML(xml,&tempString);
			device->ModelName = (char*)malloc(1+tempStringLength);
			memcpy(device->ModelName,tempString,tempStringLength);
			device->ModelName[tempStringLength] = '\0';
		} else
		if(xml->NameLength==11 && memcmp(xml->Name,"modelNumber",11) == 0)
		{
			tempStringLength = ILibReadInnerXML(xml,&tempString);
			device->ModelNumber = (char*)malloc(1+tempStringLength);
			memcpy(device->ModelNumber,tempString,tempStringLength);
			device->ModelNumber[tempStringLength] = '\0';
		} else
		if(xml->NameLength==8 && memcmp(xml->Name,"modelURL",8) == 0)
		{
			tempStringLength = ILibReadInnerXML(xml,&tempString);
			device->ModelURL = (char*)malloc(1+tempStringLength);
			memcpy(device->ModelURL,tempString,tempStringLength);
			device->ModelURL[tempStringLength] = '\0';
		} else
		if(xml->NameLength==15 && memcmp(xml->Name,"presentationURL",15) == 0)
		{
			tempStringLength = ILibReadInnerXML(xml,&tempString);
			tempString[tempStringLength] = 0;
			tpr = ILibParseString(tempString,0,tempStringLength,"://",3);
			if(tpr->NumResults==1)
			{
				/* RelativeURL */
				if(tempString[0]=='/')
				{
					device->PresentationURL = (char*)malloc(1+strlen(BaseURL)+tempStringLength);
					memcpy(device->PresentationURL,BaseURL,strlen(BaseURL));
					strcpy(device->PresentationURL+strlen(BaseURL),tempString+1);
				}
				else
				{
					device->PresentationURL = (char*)malloc(2+strlen(BaseURL)+tempStringLength);
					memcpy(device->PresentationURL,BaseURL,strlen(BaseURL));
					strcpy(device->PresentationURL+strlen(BaseURL),tempString);
				}
			}
			else
			{
				/* AbsoluteURL */
				device->PresentationURL = (char*)malloc(1+tempStringLength);
				memcpy(device->PresentationURL,tempString,tempStringLength);
				device->PresentationURL[tempStringLength] = '\0';
			}
			ILibDestructParserResults(tpr);
		} else
		if(xml->NameLength==9 && memcmp(xml->Name,"X_DLNADOC",9)==0)
		{
			tempString = ILibXML_LookupNamespace(xml,xml->NSTag,xml->NSLength);
			if(strcmp(tempString,"urn:schemas-dlna-org:device-1-0")==0)
			{
				tempStringLength = ILibReadInnerXML(xml,&tempString);
				tempString2 = (char*)malloc(tempStringLength+1);
				memcpy(tempString2,tempString,tempStringLength);
				tempString2[tempStringLength]=0;
				if(ILibGetEntry(device->CustomTagTable,"urn:schemas-dlna-org:device-1-0",31)==NULL)
				{
					ILibAddEntry(device->CustomTagTable,"urn:schemas-dlna-org:device-1-0",31,ILibInitHashTree());
				}
				ILibAddEntry(ILibGetEntry(device->CustomTagTable,"urn:schemas-dlna-org:device-1-0",31),xml->Name,xml->NameLength,tempString2);
			}
		} 
		else
		
		if(xml->NameLength==11 && memcmp(xml->Name,"serviceList",11)==0)
		{
			tempNode = xml;
			xml = xml->Next;
			while(xml!=NULL)
			{
				if(xml->NameLength==7 && memcmp(xml->Name,"service",7)==0)
				{
					ServiceType = NULL;
					ServiceTypeLength = 0;
					SCPDURL = NULL;
					SCPDURLLength = 0;
					EventSubURL = NULL;
					EventSubURLLength = 0;
					ControlURL = NULL;
					ControlURLLength = 0;
					
					xml = xml->Next;
					flg = 0;
					while(flg==0)
					{
						if(xml->NameLength==11 && memcmp(xml->Name,"serviceType",11)==0)
						{
							ServiceTypeLength = ILibReadInnerXML(xml,&ServiceType);
						} else
						if(xml->NameLength==7 && memcmp(xml->Name,"SCPDURL",7) == 0)
						{
							SCPDURLLength = ILibReadInnerXML(xml,&SCPDURL);
						} else
						if(xml->NameLength==10 && memcmp(xml->Name,"controlURL",10) == 0)
						{
							ControlURLLength = ILibReadInnerXML(xml,&ControlURL);
						} else
						if(xml->NameLength==11 && memcmp(xml->Name,"eventSubURL",11) == 0)
						{
							EventSubURLLength = ILibReadInnerXML(xml,&EventSubURL);
						}
						
						if(xml->Peer!=NULL)
						{
							xml = xml->Peer;
						}
						else
						{
							flg = 1;
							xml = xml->Parent;
						}
					}
					
					/* Finished Parsing the ServiceSection, build the Service */
					ServiceType[ServiceTypeLength] = '\0';
					SCPDURL[SCPDURLLength] = '\0';
					EventSubURL[EventSubURLLength] = '\0';
					ControlURL[ControlURLLength] = '\0';
					
					TempService = (struct UPnPService*)malloc(sizeof(struct UPnPService));
					TempService->SubscriptionID = NULL;
					TempService->ServiceId = NULL;
					TempService->Actions = NULL;
					TempService->Variables = NULL;
					TempService->Next = NULL;
					TempService->Parent = device;
					if(EventSubURLLength>=7 && memcmp(EventSubURL,"http://",6)==0)
					{
						/* Explicit */
						TempService->SubscriptionURL = (char*)malloc(EventSubURLLength+1);
						memcpy(TempService->SubscriptionURL,EventSubURL,EventSubURLLength);
						TempService->SubscriptionURL[EventSubURLLength] = '\0';
					}
					else
					{
						/* Relative */
						if(memcmp(EventSubURL,"/",1)!=0)
						{
							TempService->SubscriptionURL = (char*)malloc(EventSubURLLength+(int)strlen(BaseURL)+1);
							memcpy(TempService->SubscriptionURL,BaseURL,(int)strlen(BaseURL));
							memcpy(TempService->SubscriptionURL+(int)strlen(BaseURL),EventSubURL,EventSubURLLength);
							TempService->SubscriptionURL[EventSubURLLength+(int)strlen(BaseURL)] = '\0';
						}
						else
						{
							TempService->SubscriptionURL = (char*)malloc(EventSubURLLength+(int)strlen(BaseURL)+1);
							memcpy(TempService->SubscriptionURL,BaseURL,(int)strlen(BaseURL));
							memcpy(TempService->SubscriptionURL+(int)strlen(BaseURL),EventSubURL+1,EventSubURLLength-1);
							TempService->SubscriptionURL[EventSubURLLength+(int)strlen(BaseURL)-1] = '\0';
						}
					}
					if(ControlURLLength>=7 && memcmp(ControlURL,"http://",6)==0)
					{
						/* Explicit */
						TempService->ControlURL = (char*)malloc(ControlURLLength+1);
						memcpy(TempService->ControlURL,ControlURL,ControlURLLength);
						TempService->ControlURL[ControlURLLength] = '\0';
					}
					else
					{
						/* Relative */
						if(memcmp(ControlURL,"/",1)!=0)
						{
							TempService->ControlURL = (char*)malloc(ControlURLLength+(int)strlen(BaseURL)+1);
							memcpy(TempService->ControlURL,BaseURL,(int)strlen(BaseURL));
							memcpy(TempService->ControlURL+(int)strlen(BaseURL),ControlURL,ControlURLLength);
							TempService->ControlURL[ControlURLLength+(int)strlen(BaseURL)] = '\0';
						}
						else
						{
							TempService->ControlURL = (char*)malloc(ControlURLLength+(int)strlen(BaseURL)+1);
							memcpy(TempService->ControlURL,BaseURL,(int)strlen(BaseURL));
							memcpy(TempService->ControlURL+(int)strlen(BaseURL),ControlURL+1,ControlURLLength-1);
							TempService->ControlURL[ControlURLLength+(int)strlen(BaseURL)-1] = '\0';
						}
					}
					if(SCPDURLLength>=7 && memcmp(SCPDURL,"http://",6)==0)
					{
						/* Explicit */
						TempService->SCPDURL = (char*)malloc(SCPDURLLength+1);
						memcpy(TempService->SCPDURL,SCPDURL,SCPDURLLength);
						TempService->SCPDURL[SCPDURLLength] = '\0';
					}
					else
					{
						/* Relative */
						if(memcmp(SCPDURL,"/",1)!=0)
						{
							TempService->SCPDURL = (char*)malloc(SCPDURLLength+(int)strlen(BaseURL)+1);
							memcpy(TempService->SCPDURL,BaseURL,(int)strlen(BaseURL));
							memcpy(TempService->SCPDURL+(int)strlen(BaseURL),SCPDURL,SCPDURLLength);
							TempService->SCPDURL[SCPDURLLength+(int)strlen(BaseURL)] = '\0';
						}
						else
						{
							TempService->SCPDURL = (char*)malloc(SCPDURLLength+(int)strlen(BaseURL)+1);
							memcpy(TempService->SCPDURL,BaseURL,(int)strlen(BaseURL));
							memcpy(TempService->SCPDURL+(int)strlen(BaseURL),SCPDURL+1,SCPDURLLength-1);
							TempService->SCPDURL[SCPDURLLength+(int)strlen(BaseURL)-1] = '\0';
						}
					}
					
					TempService->ServiceType = (char*)malloc(ServiceTypeLength+1);
					sprintf(TempService->ServiceType,ServiceType,ServiceTypeLength);
					TempService->Next = device->Services;
					device->Services = TempService;
					
					DEBUGSTATEMENT(printf("ServiceType: %s\r\nSCPDURL: %s\r\nEventSubURL: %s\r\nControl URL: %s\r\n",ServiceType,SCPDURL,EventSubURL,ControlURL));
				}
				xml = xml->Peer;
			}
			xml = tempNode;
		} // End of ServiceList
		xml = xml->Peer;
	} // End of While
	
	return(device);
}

static void MSCP_ProcessDeviceXML(void *v_CP,char* buffer, int BufferSize, char* LocationURL, int RecvAddr, int Timeout)
{
	struct UPnPDevice *RootDevice = NULL;
	
	char* IP;
	unsigned short Port;
	char* Path;
	
	char* BaseURL = NULL;
	
	struct ILibXMLNode *rootXML;
	struct ILibXMLNode *xml;
	char* tempString;
	int tempStringLength;

	if (BufferSize == 0)
		return;
	
	rootXML = ILibParseXML(buffer,0,BufferSize);
	ILibProcessXMLNodeList(rootXML);
	ILibXML_BuildNamespaceLookupTable(rootXML);
	
	xml = rootXML;
	xml = xml->Next;
	while(xml!=NULL)
	{
		if(xml->NameLength==7 && memcmp(xml->Name,"URLBase",7)==0)
		{
			tempStringLength = ILibReadInnerXML(xml,&tempString);
			if(tempString[tempStringLength-1]!='/')
			{
				BaseURL = (char*)malloc(2+tempStringLength);
				memcpy(BaseURL,tempString,tempStringLength);
				BaseURL[tempStringLength] = '/';
				BaseURL[tempStringLength+1] = '\0';
			}
			else
			{
				BaseURL = (char*)malloc(1+tempStringLength);
				memcpy(BaseURL,tempString,tempStringLength);
				BaseURL[tempStringLength] = '\0';
			}
			break;
		}
		xml = xml->Peer;
	}
	
	if(BaseURL==NULL)
	{
		ILibParseUri(LocationURL,&IP,&Port,&Path);
		BaseURL = (char*)malloc(18+(int)strlen(IP));
		sprintf(BaseURL,"http://%s:%d/",IP,Port);
		
		free(IP);
		free(Path);
	}
	
	DEBUGSTATEMENT(printf("BaseURL: %s\r\n",BaseURL));
	
	xml = rootXML;
	xml = xml->Next;
	while(xml->NameLength!=6 && memcmp(xml->Name,"device",6)!=0 && xml!=NULL)
	{
		xml = xml->Peer;
	}
	if(xml==NULL)
	{
		/* Error */
		ILibDestructXMLNodeList(rootXML);
		return;
	}
	
	RootDevice = MSCP_ProcessDeviceXML_device(xml,v_CP,BaseURL,Timeout,RecvAddr);
	free(BaseURL);
	ILibDestructXMLNodeList(rootXML);
	
	
	/* Save reference to LocationURL in the RootDevice */
	RootDevice->LocationURL = (char*)malloc(strlen(LocationURL)+1);
	sprintf(RootDevice->LocationURL,"%s",LocationURL);
	
	MSCP_ProcessDevice(v_CP,RootDevice);
	RootDevice->SCPDLeft = 0;
	MSCP_CalculateSCPD_FetchCount(RootDevice);
	if(RootDevice->SCPDLeft==0)
	{
		MSCP_FinishProcessingDevice(v_CP,RootDevice);
	}
	else
	{
		MSCP_SCPD_Fetch(RootDevice);
	}
}

static void MSCP_HTTP_Sink_DeviceDescription(
void *WebReaderToken,
int IsInterrupt,
struct packetheader *header,
char *buffer,
int *p_BeginPointer,
int EndPointer,
int done,
void *user,
void *cp,
int *PAUSE)
{
	struct CustomUserData *customData = (struct CustomUserData*)user;
	if(header!=NULL && done!=0)
	{
		MSCP_ProcessDeviceXML(cp,buffer,EndPointer-(*p_BeginPointer),customData->buffer,header->ReceivingAddress,customData->Timeout);
	}
	if(done!=0)
	{
		free(customData->buffer);
		free(user);
	}
}
static void MSCP__FlushRequest(struct UPnPDevice *device)
{
	struct UPnPDevice *ed = device->EmbeddedDevices;
	struct UPnPService *s = device->Services;
	
	while(ed!=NULL)
	{
		MSCP__FlushRequest(ed);
		ed = ed->Next;
	}
	while(s!=NULL)
	{
		s = s->Next;
	}
}
static void MSCP_CP_RecursiveReleaseAndEventDevice(struct MSCP_CP* CP, struct UPnPDevice *device)
{
	struct UPnPDevice *temp = device->EmbeddedDevices;
	
	while(temp!=NULL)
	{
		MSCP_CP_RecursiveReleaseAndEventDevice(CP,temp);
		temp = temp->Next;
	}
	
	if(device->Reserved!=0)
	{
		if(CP->RemoveSink!=NULL)
		{
			CP->RemoveSink(device);
		}
		MSCP_Release(device);
	}
}
void MSCP_CP_ProcessDeviceRemoval(struct MSCP_CP* CP, struct UPnPDevice *device)
{
	struct UPnPDevice *temp = device->EmbeddedDevices;
	struct UPnPService *s;
	char *IP;
	unsigned short port;
	char *PATH;
	struct UPnPDevice *dTemp = device;
	
	if(dTemp->Parent!=NULL)
	{
		dTemp = dTemp->Parent;
	}
	ILibLifeTime_Remove(CP->LifeTimeMonitor,dTemp);
	
	
	ILibLifeTime_Remove(CP->LifeTimeMonitor,dTemp);
	
	while(temp!=NULL)
	{
		MSCP_CP_ProcessDeviceRemoval(CP,temp);
		temp = temp->Next;
	}
	
	s = device->Services;
	while(s!=NULL)
	{
		//
		// Remove all the pending requests
		//
		ILibParseUri(s->ControlURL,&IP,&port,&PATH);
		ILibWebClient_DeleteRequests(((struct MSCP_CP*)device->CP)->HTTP,IP,(int)port);
		free(IP);
		free(PATH);
		
		ILibLifeTime_Remove(CP->LifeTimeMonitor,s);
		s = s->Next;
	}
	
	if(device->Reserved!=0)
	{
		// device was flagged, and given to the user
		if(CP->RemoveSink!=NULL)
		{
			CP->RemoveSink(device);
		}
		MSCP_Release(device);
	}
	
	ILibHashTree_Lock(CP->DeviceTable);
	ILibDeleteEntry(CP->DeviceTable,device->UDN,(int)strlen(device->UDN));
	if(device->LocationURL!=NULL)
	{
		ILibDeleteEntry(CP->DeviceTable,device->LocationURL,(int)strlen(device->LocationURL));
	}
	ILibHashTree_UnLock(CP->DeviceTable);
}
void MSCP_SSDP_Sink(void *sender, char* UDN, int Alive, char* LocationURL, int Timeout,UPnPSSDP_MESSAGE m,void *cp)
{
	struct CustomUserData *customData;
	char* buffer;
	char* IP;
	unsigned short Port;
	char* Path;
	struct packetheader *p;
	struct sockaddr_in addr;
	
	struct UPnPDevice *device;
	int i=0;
	struct MSCP_CP *CP = (struct MSCP_CP*)cp;
	struct timeval t;
	
	if(Alive!=0)
	{
		if(Timeout==0) {return;}	// ToDo: Solve this the correct way
		/* Hello */
		DEBUGSTATEMENT(printf("BinaryLight Hello\r\n"));
		DEBUGSTATEMENT(printf("LocationURL: %s\r\n",LocationURL));
		ILibHashTree_Lock(CP->DeviceTable);
		if(ILibHasEntry(CP->DeviceTable,LocationURL,(int)strlen(LocationURL))==0 && ILibHasEntry(CP->DeviceTable,UDN,(int)strlen(UDN))==0)
		{
			ILibAddEntry(CP->DeviceTable,LocationURL,(int)strlen(LocationURL),NULL);
			ILibAddEntry(CP->DeviceTable,UDN,(int)strlen(UDN),NULL);
			ILibParseUri(LocationURL,&IP,&Port,&Path);
			DEBUGSTATEMENT(printf("IP: %s Port: %d Path: %s\r\n",IP,Port,Path));
			p = MSCP_BuildPacket(IP,Port,Path,"GET");
			
			memset((char *)&addr, 0,sizeof(addr));
			addr.sin_family = AF_INET;
			addr.sin_addr.s_addr = inet_addr(IP);
			addr.sin_port = htons(Port);
			
			buffer = (char*)malloc((int)strlen(LocationURL)+1);
			strcpy(buffer,LocationURL);
			
			customData = (struct CustomUserData*)malloc(sizeof(struct CustomUserData));
			customData->Timeout = Timeout;
			customData->buffer = buffer;
			
			ILibWebClient_PipelineRequest(
			((struct MSCP_CP*)cp)->HTTP,
			&addr,
			p,
			&MSCP_HTTP_Sink_DeviceDescription,
			customData, 
			cp);
			
			free(IP);
			free(Path);
		}
		else
		{
			// Periodic Notify Packets
			device = (struct UPnPDevice*)ILibGetEntry(CP->DeviceTable,UDN,(int)strlen(UDN));
			if(device!=NULL && m==UPnPSSDP_NOTIFY)
			{
				while(device->Parent!=NULL)
				{
					device = device->Parent;
				}
				if(strcmp(device->LocationURL,LocationURL)==0)
				{
					//Extend LifetimeMonitor duration
					gettimeofday(&t,NULL);
					device->Reserved2 = t.tv_sec;
					ILibLifeTime_Remove(((struct MSCP_CP*)cp)->LifeTimeMonitor,device);
					ILibLifeTime_Add(((struct MSCP_CP*)cp)->LifeTimeMonitor,device,Timeout,&MSCP_ExpiredDevice,NULL);
				}
				else
				{
					//
					// Same device, different Interface
					// Wait up to 7 seconds to see if the old interface is still
					// valid.
					//
					gettimeofday(&t,NULL);
					if(t.tv_sec-device->Reserved2>10)
					{
						device->Reserved2 = t.tv_sec;
						if(device->Reserved3!=NULL) {free(device->Reserved3);}
						device->Reserved3 = (char*)malloc(strlen(LocationURL)+1);
						strcpy(device->Reserved3,LocationURL);
						ILibLifeTime_Remove(((struct MSCP_CP*)cp)->LifeTimeMonitor,device);
						ILibLifeTime_Add(((struct MSCP_CP*)cp)->LifeTimeMonitor,device,7,&MSCP_InterfaceChanged,NULL);
					}
				}
			}
		}
		ILibHashTree_UnLock(CP->DeviceTable);
	}
	else
	{
		/* Bye Bye */
		DEBUGSTATEMENT(printf("BinaryLight ByeBye\r\n"));
		ILibHashTree_Lock(CP->DeviceTable);
		device = (struct UPnPDevice*)ILibGetEntry(CP->DeviceTable,UDN,(int)strlen(UDN));
		ILibHashTree_UnLock(CP->DeviceTable);
		if(device!=NULL)
		{
			while(device->Parent!=NULL)
			{
				device = device->Parent;
			}
			ILibLifeTime_Remove(((struct MSCP_CP*)cp)->LifeTimeMonitor,device);
			MSCP_CP_ProcessDeviceRemoval(CP,device);
			i = device->ReferenceTiedToEvents;
			while(i!=0)
			{
				MSCP_Release(device);
				--i;
			}
			MSCP_Release(device);
		}
	}
}

static void MSCP_ContentDirectory_EventSink(char* buffer, int bufferlength, struct UPnPService *service)
{
	struct ILibXMLNode *xml,*rootXML;
	char *tempString;
	int tempStringLength;
	int flg,flg2;
	
	char* ContainerUpdateIDs = 0;
	unsigned int SystemUpdateID = 0;
	char* TransferIDs = 0;
	unsigned long TempULong;
	
	/* Parse SOAP */
	rootXML = xml = ILibParseXML(buffer,0,bufferlength);
	ILibProcessXMLNodeList(xml);
	
	while(xml!=NULL)
	{
		if(xml->NameLength==11 && memcmp(xml->Name,"propertyset",11)==0)
		{
			if(xml->Next->StartTag!=0)
			{
				flg = 0;
				xml = xml->Next;
				while(flg==0)
				{
					if(xml->NameLength==8 && memcmp(xml->Name,"property",8)==0)
					{
						xml = xml->Next;
						flg2 = 0;
						while(flg2==0)
						{
							if(xml->NameLength==18 && memcmp(xml->Name,"ContainerUpdateIDs",18) == 0)
							{
								tempStringLength = ILibReadInnerXML(xml,&tempString);
								tempString[tempStringLength] = '\0';
								ContainerUpdateIDs = tempString;
								if(MSCP_EventCallback_ContentDirectory_ContainerUpdateIDs != NULL)
								{
									MSCP_EventCallback_ContentDirectory_ContainerUpdateIDs(service,ContainerUpdateIDs);
								}
							}
							if(xml->NameLength==14 && memcmp(xml->Name,"SystemUpdateID",14) == 0)
							{
								tempStringLength = ILibReadInnerXML(xml,&tempString);
								if(ILibGetULong(tempString,tempStringLength,&TempULong)==0)
								{
									SystemUpdateID = (unsigned int) TempULong;
								}
								if(MSCP_EventCallback_ContentDirectory_SystemUpdateID != NULL)
								{
									MSCP_EventCallback_ContentDirectory_SystemUpdateID(service,SystemUpdateID);
								}
							}
							if(xml->NameLength==11 && memcmp(xml->Name,"TransferIDs",11) == 0)
							{
								tempStringLength = ILibReadInnerXML(xml,&tempString);
								tempString[tempStringLength] = '\0';
								TransferIDs = tempString;
								if(MSCP_EventCallback_ContentDirectory_TransferIDs != NULL)
								{
									MSCP_EventCallback_ContentDirectory_TransferIDs(service,TransferIDs);
								}
							}
							if(xml->Peer!=NULL)
							{
								xml = xml->Peer;
							}
							else
							{
								flg2 = -1;
								xml = xml->Parent;
							}
						}
					}
					if(xml->Peer!=NULL)
					{
						xml = xml->Peer;
					}
					else
					{
						flg = -1;
						xml = xml->Parent;
					}
				}
			}
		}
		xml = xml->Peer;
	}
	
	ILibDestructXMLNodeList(rootXML);
}

static void MSCP_urn_microsoft_com_service_X_MS_MediaReceiverRegistrar_EventSink(char* buffer, int bufferlength, struct UPnPService *service)
{
	struct ILibXMLNode *xml,*rootXML;
	char *tempString;
	int tempStringLength;
	int flg,flg2;
	
	unsigned int AuthorizationDeniedUpdateID = 0;
	unsigned int ValidationSucceededUpdateID = 0;
	unsigned int ValidationRevokedUpdateID = 0;
	unsigned int AuthorizationGrantedUpdateID = 0;
	unsigned long TempULong;
	
	/* Parse SOAP */
	rootXML = xml = ILibParseXML(buffer,0,bufferlength);
	ILibProcessXMLNodeList(xml);
	
	while(xml!=NULL)
	{
		if(xml->NameLength==11 && memcmp(xml->Name,"propertyset",11)==0)
		{
			if(xml->Next->StartTag!=0)
			{
				flg = 0;
				xml = xml->Next;
				while(flg==0)
				{
					if(xml->NameLength==8 && memcmp(xml->Name,"property",8)==0)
					{
						xml = xml->Next;
						flg2 = 0;
						while(flg2==0)
						{
							if(xml->NameLength==27 && memcmp(xml->Name,"AuthorizationDeniedUpdateID",27) == 0)
							{
								tempStringLength = ILibReadInnerXML(xml,&tempString);
								if(ILibGetULong(tempString,tempStringLength,&TempULong)==0)
								{
									AuthorizationDeniedUpdateID = (unsigned int) TempULong;
								}
								if(MSCP_EventCallback_Registrar_AuthorizationDeniedUpdateID != NULL)
								{
									MSCP_EventCallback_Registrar_AuthorizationDeniedUpdateID(service,AuthorizationDeniedUpdateID);
								}
							}
							if(xml->NameLength==27 && memcmp(xml->Name,"ValidationSucceededUpdateID",27) == 0)
							{
								tempStringLength = ILibReadInnerXML(xml,&tempString);
								if(ILibGetULong(tempString,tempStringLength,&TempULong)==0)
								{
									ValidationSucceededUpdateID = (unsigned int) TempULong;
								}
								if(MSCP_EventCallback_Registrar_ValidationSucceededUpdateID != NULL)
								{
									MSCP_EventCallback_Registrar_ValidationSucceededUpdateID(service,ValidationSucceededUpdateID);
								}
							}
							if(xml->NameLength==25 && memcmp(xml->Name,"ValidationRevokedUpdateID",25) == 0)
							{
								tempStringLength = ILibReadInnerXML(xml,&tempString);
								if(ILibGetULong(tempString,tempStringLength,&TempULong)==0)
								{
									ValidationRevokedUpdateID = (unsigned int) TempULong;
								}
								if(MSCP_EventCallback_Registrar_ValidationRevokedUpdateID != NULL)
								{
									MSCP_EventCallback_Registrar_ValidationRevokedUpdateID(service,ValidationRevokedUpdateID);
								}
							}
							if(xml->NameLength==28 && memcmp(xml->Name,"AuthorizationGrantedUpdateID",28) == 0)
							{
								tempStringLength = ILibReadInnerXML(xml,&tempString);
								if(ILibGetULong(tempString,tempStringLength,&TempULong)==0)
								{
									AuthorizationGrantedUpdateID = (unsigned int) TempULong;
								}
								if(MSCP_EventCallback_Registrar_AuthorizationGrantedUpdateID != NULL)
								{
									MSCP_EventCallback_Registrar_AuthorizationGrantedUpdateID(service,AuthorizationGrantedUpdateID);
								}
							}
							if(xml->Peer!=NULL)
							{
								xml = xml->Peer;
							}
							else
							{
								flg2 = -1;
								xml = xml->Parent;
							}
						}
					}
					if(xml->Peer!=NULL)
					{
						xml = xml->Peer;
					}
					else
					{
						flg = -1;
						xml = xml->Parent;
					}
				}
			}
		}
		xml = xml->Peer;
	}
	
	ILibDestructXMLNodeList(rootXML);
}


static void MSCP_OnEventSink(
struct ILibWebServer_Session *sender,
int InterruptFlag,
struct packetheader *header,
char *buffer,
int *BeginPointer,
int BufferSize,
int done)
{
	int type_length;
	char* sid = NULL;
	void* value = NULL;
	struct UPnPService *service = NULL;
	struct packetheader_field_node *field = NULL;
	struct packetheader *resp;
	if(done!=0)
	{
		resp = ILibCreateEmptyPacket();
		ILibSetVersion(resp,"1.1",3);
		ILibAddHeaderLine(resp,"Server",6,MSCP_PLATFORM,(int)strlen(MSCP_PLATFORM));
		ILibAddHeaderLine(resp,"Content-Length",14,"0",1);
		field = header->FirstField;
		while(field!=NULL)
		{
			if(field->FieldLength==3)
			{
				if(strncasecmp(field->Field,"SID",3)==0)
				{
					sid = (char*)malloc(field->FieldDataLength+1);
					sprintf(sid,"%s",field->FieldData);
					value = ILibGetEntry(((struct MSCP_CP*)sender->User)->SIDTable,field->FieldData,field->FieldDataLength);
					break;
				}
			}
			field = field->NextField;
		}
		
		if(value==NULL)
		{
			/* Not a valid SID */
			ILibSetStatusCode(resp,412,"Failed",6);
		}
		else
		{
			ILibSetStatusCode(resp,200,"OK",2);
			service = (struct UPnPService*)value;
			
			type_length = (int)strlen(service->ServiceType);
			if(type_length>46 && strncmp("urn:schemas-upnp-org:service:ContentDirectory:",service->ServiceType,46)==0)
			{
				MSCP_ContentDirectory_EventSink(buffer, BufferSize, service);
			}
			else
			if(type_length>54 && strncmp("urn:microsoft.com:service:X_MS_MediaReceiverRegistrar:",service->ServiceType,54)==0)
			{
				MSCP_urn_microsoft_com_service_X_MS_MediaReceiverRegistrar_EventSink(buffer, BufferSize, service);
			}
			
		}
		ILibWebServer_Send(sender,resp);
		if(sid!=NULL){free(sid);}
	}
}
static void MSCP_OnUnSubscribeSink(
void *WebReaderToken,
int IsInterrupt,
struct packetheader *header,
char *buffer,
int *BeginPointer,
int BufferSize,
int done,
void *user,
void *vcp,
int *PAUSE)
{
	struct UPnPService *s;
	//struct MSCP_CP *cp = (struct MSCP_CP*)vcp;
	
	if(done!=0)
	{
		s = (struct UPnPService*)user;
		if(header!=NULL)
		{
			if(header->StatusCode==200)
			{
				/* Successful */
			}
		}
		MSCP_Release(s->Parent);
	}
}

static void MSCP_OnSubscribeSink(
void *WebReaderToken,
int IsInterrupt,
struct packetheader *header,
char *buffer,
int *BeginPointer,
int BufferSize,
int done,
void *user,
void *vcp,
int *PAUSE)
{
	struct UPnPService *s;
	struct UPnPDevice *d;
	struct packetheader_field_node *field;
	struct parser_result *p;
	struct MSCP_CP *cp = (struct MSCP_CP*)vcp;
	
	if(done!=0)
	{
		s = (struct UPnPService*)user;
		if(header!=NULL)
		{
			if(header->StatusCode==200)
			{
				/* Successful */
				field = header->FirstField;
				while(field!=NULL)
				{
					if(field->FieldLength==3 && strncasecmp(field->Field,"SID",3)==0 && s->SubscriptionID==NULL)
					{
						s->SubscriptionID = (char*)malloc(1+field->FieldDataLength);
						strcpy(s->SubscriptionID,field->FieldData);
						ILibAddEntry(cp->SIDTable,field->FieldData,field->FieldDataLength,s);
					} else
					if(field->FieldLength==7 && strncasecmp(field->Field,"TIMEOUT",7)==0)
					{
						p = ILibParseString(field->FieldData,0,field->FieldDataLength,"-",1);
						p->LastResult->data[p->LastResult->datalength] = '\0';
						MSCP_AddRef(s->Parent);
						d = s->Parent;
						while(d->Parent!=NULL) {d = d->Parent;}
						++d->ReferenceTiedToEvents;
						ILibLifeTime_Add(cp->LifeTimeMonitor,s,atoi(p->LastResult->data)/2,&MSCP_Renew,NULL);
						ILibDestructParserResults(p);
					}
					field = field->NextField;
				}
			}
		}
		MSCP_Release(s->Parent);
	}
}

void MSCP_Renew(void *state)
{
	struct UPnPService *service = (struct UPnPService*)state;
	struct UPnPDevice *d = service->Parent;
	char *IP;
	char *Path;
	unsigned short Port;
	struct packetheader *p;
	char* TempString;
	struct sockaddr_in destaddr;
	
	ILibParseUri(service->SubscriptionURL,&IP,&Port,&Path);
	p = ILibCreateEmptyPacket();
	ILibSetVersion(p,"1.1",3);
	
	ILibSetDirective(p,"SUBSCRIBE",9,Path,(int)strlen(Path));
	
	TempString = (char*)malloc((int)strlen(IP)+7);
	sprintf(TempString,"%s:%d",IP,Port);
	
	ILibAddHeaderLine(p,"HOST",4,TempString,(int)strlen(TempString));
	free(TempString);
	
	ILibAddHeaderLine(p,"SID",3,service->SubscriptionID,(int)strlen(service->SubscriptionID));
	ILibAddHeaderLine(p,"TIMEOUT",7,"Second-180",10);
	ILibAddHeaderLine(p,"User-Agent",10,MSCP_PLATFORM,(int)strlen(MSCP_PLATFORM));
	
	memset((char *)&destaddr, 0,sizeof(destaddr));
	destaddr.sin_family = AF_INET;
	destaddr.sin_addr.s_addr = inet_addr(IP);
	destaddr.sin_port = htons(Port);
	
	ILibWebClient_PipelineRequest(
	((struct MSCP_CP*)service->Parent->CP)->HTTP,
	&destaddr,
	p,
	&MSCP_OnSubscribeSink,
	(void*)service, service->Parent->CP);
	
	while(d->Parent!=NULL) {d = d->Parent;}
	--d->ReferenceTiedToEvents;
	free(IP);
	free(Path);
}

struct UPnPDevice* MSCP_GetDeviceEx(struct UPnPDevice *device, char* DeviceType, int counter, int number)
{
	struct UPnPDevice *RetVal = NULL;
	struct UPnPDevice *d = device->EmbeddedDevices;
	struct parser_result *pr,*pr2;
	int DeviceTypeLength = (int)strlen(DeviceType);
	int TempLength = (int)strlen(device->DeviceType);
	
	while(d != NULL && RetVal==NULL)
	{
		RetVal = MSCP_GetDeviceEx(d,DeviceType,counter,number);
		d = d->Next;
	}
	
	if(RetVal==NULL)
	{
		pr = ILibParseString(DeviceType,0,DeviceTypeLength,":",1);
		pr2 = ILibParseString(device->DeviceType,0,TempLength,":",1);
		
		if(DeviceTypeLength-pr->LastResult->datalength == TempLength - pr2->LastResult->datalength && atoi(pr->LastResult->data) >= atoi(pr2->LastResult->data) && memcmp(DeviceType,device->DeviceType,DeviceTypeLength-pr->LastResult->datalength)==0)
		{
			ILibDestructParserResults(pr);
			ILibDestructParserResults(pr2);
			if(number == (++counter)) return(device);
		}
		ILibDestructParserResults(pr);
		ILibDestructParserResults(pr2);
		return(NULL);
	}
	else
	{
		return(RetVal);
	}
}
int MSCP_HasAction(struct UPnPService *s, char* action)
{
	struct UPnPAction *a = s->Actions;
	
	while(a!=NULL)
	{
		if(strcmp(action,a->Name)==0) return(-1);
		a = a->Next;
	}
	return(0);
}
static void MSCP_StopCP(void *v_CP)
{
	int i;
	struct UPnPDevice *Device;
	struct MSCP_CP *CP= (struct MSCP_CP*)v_CP;
	void *en;
	char *key;
	int keyLength;
	void *data;
	
	
	en = ILibHashTree_GetEnumerator(CP->DeviceTable);
	while(ILibHashTree_MoveNext(en)==0)
	{
		ILibHashTree_GetValue(en,&key,&keyLength,&data);
		if(data!=NULL)
		{
			// This is a UPnPDevice
			Device = (struct UPnPDevice*)data;
			if(Device->Parent==NULL)
			{
				// This is the Root Device (Which is only in the table once)
				MSCP_CP_RecursiveReleaseAndEventDevice(CP,Device);
				i = Device->ReferenceTiedToEvents;
				while(i!=0)
				{
					MSCP_Release(Device);
					--i;
				}
				MSCP_Release(Device);
			}
		}
	}
	ILibHashTree_DestroyEnumerator(en);
	ILibDestroyHashTree(CP->SIDTable);
	ILibDestroyHashTree(CP->DeviceTable);
	free(CP->AddressList);
}
void MSCP__CP_IPAddressListChanged(void *CPToken)
{
	if(ILibIsChainBeingDestroyed(((struct MSCP_CP*)CPToken)->Chain)==0)
	{
		((struct MSCP_CP*)CPToken)->RecheckFlag = 1;
		ILibForceUnBlockChain(((struct MSCP_CP*)CPToken)->Chain);
	}
}

static void MSCP_CP_PreSelect(void *CPToken,fd_set *readset, fd_set *writeset, fd_set *errorset, int *blocktime)
{
	struct MSCP_CP *CP= (struct MSCP_CP*)CPToken;
	void *en;
	
	struct UPnPDevice *device;
	char *key;
	int keyLength;
	void *data;
	void *q;
	
	int *IPAddressList;
	int NumAddressList;
	int i;
	int found;
	
	if(CP->RecheckFlag!=0)
	{
		CP->RecheckFlag = 0;
		
		NumAddressList = ILibGetLocalIPAddressList(&IPAddressList);
		q = ILibQueue_Create();
		
		ILibHashTree_Lock(CP->DeviceTable);
		en = ILibHashTree_GetEnumerator(CP->DeviceTable);
		while(ILibHashTree_MoveNext(en)==0)
		{
			ILibHashTree_GetValue(en,&key,&keyLength,&data);
			if(data!=NULL)
			{
				// This is a MSCP_ Device
				device = (struct UPnPDevice*)data;
				if(device->Parent==NULL)
				{
					// This is the root device, which is in the table exactly once
					found = 0;
					for(i=0;i<NumAddressList;++i)
					{
						if((unsigned int)IPAddressList[i]==inet_addr(device->InterfaceToHost))
						{
							found = 1;
							break;
						}
					}
					if(found==0)
					{
						// Queue Device to be removed, so we can process it
						// outside of the lock
						
						ILibQueue_EnQueue(q,device);
					}
				}
			}
		}
		
		ILibHashTree_DestroyEnumerator(en);
		ILibHashTree_UnLock(CP->DeviceTable);
		
		while(ILibQueue_PeekQueue(q)!=NULL)
		{
			device = ILibQueue_DeQueue(q);
			MSCP_CP_ProcessDeviceRemoval(CP,device);
			MSCP_Release(device);
		}
		ILibQueue_Destroy(q);
		ILibSSDP_IPAddressListChanged(CP->SSDP);
		free(CP->AddressList);
		CP->AddressListLength = NumAddressList;
		CP->AddressList = IPAddressList;
	}
}
static void MSCP_OnSessionSink(struct ILibWebServer_Session *session, void *User)
{
	session->OnReceive = &MSCP_OnEventSink;
	session->User = User;
}
void *MSCP_CreateControlPoint(void *Chain, void(*A)(struct UPnPDevice*),void(*R)(struct UPnPDevice*))
{
	struct MSCP_CP *cp = (struct MSCP_CP*)malloc(sizeof(struct MSCP_CP));
	
	cp->Destroy = &MSCP_StopCP;
	cp->PostSelect = NULL;
	cp->PreSelect = &MSCP_CP_PreSelect;
	cp->DiscoverSink = A;
	cp->RemoveSink = R;
	
	cp->WebServer = ILibWebServer_Create(Chain,5,0,&MSCP_OnSessionSink,cp);
	cp->SIDTable = ILibInitHashTree();
	cp->DeviceTable = ILibInitHashTree();
	
	cp->SSDP = ILibCreateSSDPClientModule(Chain,"urn:schemas-upnp-org:device:MediaServer:1", 41, &MSCP_SSDP_Sink,cp);
	
	cp->HTTP = ILibCreateWebClient(5,Chain);
	ILibAddToChain(Chain,cp);
	cp->LifeTimeMonitor = ILibCreateLifeTime(Chain);
	
	cp->Chain = Chain;
	cp->RecheckFlag = 0;
	cp->AddressListLength = ILibGetLocalIPAddressList(&(cp->AddressList));
	return((void*)cp);
}

static void MSCP_Invoke_ContentDirectory_Browse_Sink(
void *WebReaderToken,
int IsInterrupt,
struct packetheader *header,
char *buffer,
int *p_BeginPointer,
int EndPointer,
int done,
void *_service,
void *state,
int *PAUSE)
{
	struct UPnPService *Service = (struct UPnPService*)_service;
	struct InvokeStruct *_InvokeData = (struct InvokeStruct*)state;
	int ArgLeft = 4;
	struct ILibXMLNode *xml;
	struct ILibXMLNode *__xml;
	char *tempBuffer;
	int tempBufferLength;
	unsigned long TempULong;
	char* Result = NULL;
	unsigned int NumberReturned = 0;
	unsigned int TotalMatches = 0;
	unsigned int UpdateID = 0;
	LVL3DEBUG(char *DebugBuffer;)
	
	if(done==0){return;}
	LVL3DEBUG(DebugBuffer = (char*)malloc(EndPointer+1);)
	LVL3DEBUG(memcpy(DebugBuffer,buffer,EndPointer);)
	LVL3DEBUG(DebugBuffer[EndPointer]=0;)
	LVL3DEBUG(printf("\r\n SOAP Recieved:\r\n%s\r\n",DebugBuffer);)
	LVL3DEBUG(free(DebugBuffer);)
	if(_InvokeData->CallbackPtr==NULL)
	{
		MSCP_Release(Service->Parent);
		free(_InvokeData);
		return;
	}
	else
	{
		if(header==NULL)
		{
			/* Connection Failed */
			((void (*)(struct UPnPService*,int,void*,char*,unsigned int,unsigned int,unsigned int))_InvokeData->CallbackPtr)(Service,IsInterrupt==0?-1:IsInterrupt,_InvokeData->User,INVALID_DATA,INVALID_DATA,INVALID_DATA,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
		else if(header->StatusCode!=200)
		{
			/* SOAP Fault */
			((void (*)(struct UPnPService*,int,void*,char*,unsigned int,unsigned int,unsigned int))_InvokeData->CallbackPtr)(Service,MSCP_GetErrorCode(buffer,EndPointer-(*p_BeginPointer)),_InvokeData->User,INVALID_DATA,INVALID_DATA,INVALID_DATA,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
	}
	
	__xml = xml = ILibParseXML(buffer,0,EndPointer-(*p_BeginPointer));
	ILibProcessXMLNodeList(xml);
	while(xml!=NULL)
	{
		if(xml->NameLength==14 && memcmp(xml->Name,"BrowseResponse",14)==0)
		{
			xml = xml->Next;
			while(xml!=NULL)
			{
				if(xml->NameLength==6 && memcmp(xml->Name,"Result",6) == 0)
				{
					--ArgLeft;
					tempBufferLength = ILibReadInnerXML(xml,&tempBuffer);
					if(tempBufferLength!=0)
					{
						tempBuffer[tempBufferLength] = '\0';
						Result = tempBuffer;
						ILibInPlaceXmlUnEscape(Result);
					}
				}
				else 
				if(xml->NameLength==14 && memcmp(xml->Name,"NumberReturned",14) == 0)
				{
					--ArgLeft;
					tempBufferLength = ILibReadInnerXML(xml,&tempBuffer);
					if(ILibGetULong(tempBuffer,tempBufferLength,&TempULong)==0)
					{
						NumberReturned = (unsigned int) TempULong;
					}
				}
				else 
				if(xml->NameLength==12 && memcmp(xml->Name,"TotalMatches",12) == 0)
				{
					--ArgLeft;
					tempBufferLength = ILibReadInnerXML(xml,&tempBuffer);
					if(ILibGetULong(tempBuffer,tempBufferLength,&TempULong)==0)
					{
						TotalMatches = (unsigned int) TempULong;
					}
				}
				else 
				if(xml->NameLength==8 && memcmp(xml->Name,"UpdateID",8) == 0)
				{
					--ArgLeft;
					tempBufferLength = ILibReadInnerXML(xml,&tempBuffer);
					if(ILibGetULong(tempBuffer,tempBufferLength,&TempULong)==0)
					{
						UpdateID = (unsigned int) TempULong;
					}
				}
				xml = xml->Peer;
			}
		}
		if(xml!=NULL) {xml = xml->Next;}
	}
	ILibDestructXMLNodeList(__xml);
	
	if(ArgLeft!=0)
	{
		((void (*)(struct UPnPService*,int,void*,char*,unsigned int,unsigned int,unsigned int))_InvokeData->CallbackPtr)(Service,-2,_InvokeData->User,INVALID_DATA,INVALID_DATA,INVALID_DATA,INVALID_DATA);
	}
	else
	{
		((void (*)(struct UPnPService*,int,void*,char*,unsigned int,unsigned int,unsigned int))_InvokeData->CallbackPtr)(Service,0,_InvokeData->User,Result,NumberReturned,TotalMatches,UpdateID);
	}
	MSCP_Release(Service->Parent);
	free(_InvokeData);
}
static void MSCP_Invoke_ContentDirectory_GetSearchCapabilities_Sink(
void *WebReaderToken,
int IsInterrupt,
struct packetheader *header,
char *buffer,
int *p_BeginPointer,
int EndPointer,
int done,
void *_service,
void *state,
int *PAUSE)
{
	struct UPnPService *Service = (struct UPnPService*)_service;
	struct InvokeStruct *_InvokeData = (struct InvokeStruct*)state;
	int ArgLeft = 1;
	struct ILibXMLNode *xml;
	struct ILibXMLNode *__xml;
	char *tempBuffer;
	int tempBufferLength;
	char* SearchCaps = NULL;
	LVL3DEBUG(char *DebugBuffer;)
	
	if(done==0){return;}
	LVL3DEBUG(DebugBuffer = (char*)malloc(EndPointer+1);)
	LVL3DEBUG(memcpy(DebugBuffer,buffer,EndPointer);)
	LVL3DEBUG(DebugBuffer[EndPointer]=0;)
	LVL3DEBUG(printf("\r\n SOAP Recieved:\r\n%s\r\n",DebugBuffer);)
	LVL3DEBUG(free(DebugBuffer);)
	if(_InvokeData->CallbackPtr==NULL)
	{
		MSCP_Release(Service->Parent);
		free(_InvokeData);
		return;
	}
	else
	{
		if(header==NULL)
		{
			/* Connection Failed */
			((void (*)(struct UPnPService*,int,void*,char*))_InvokeData->CallbackPtr)(Service,IsInterrupt==0?-1:IsInterrupt,_InvokeData->User,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
		else if(header->StatusCode!=200)
		{
			/* SOAP Fault */
			((void (*)(struct UPnPService*,int,void*,char*))_InvokeData->CallbackPtr)(Service,MSCP_GetErrorCode(buffer,EndPointer-(*p_BeginPointer)),_InvokeData->User,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
	}
	
	__xml = xml = ILibParseXML(buffer,0,EndPointer-(*p_BeginPointer));
	ILibProcessXMLNodeList(xml);
	while(xml!=NULL)
	{
		if(xml->NameLength==29 && memcmp(xml->Name,"GetSearchCapabilitiesResponse",29)==0)
		{
			xml = xml->Next;
			while(xml!=NULL)
			{
				if(xml->NameLength==10 && memcmp(xml->Name,"SearchCaps",10) == 0)
				{
					--ArgLeft;
					tempBufferLength = ILibReadInnerXML(xml,&tempBuffer);
					if(tempBufferLength!=0)
					{
						tempBuffer[tempBufferLength] = '\0';
						SearchCaps = tempBuffer;
						ILibInPlaceXmlUnEscape(SearchCaps);
					}
				}
				xml = xml->Peer;
			}
		}
		if(xml!=NULL) {xml = xml->Next;}
	}
	ILibDestructXMLNodeList(__xml);
	
	if(ArgLeft!=0)
	{
		((void (*)(struct UPnPService*,int,void*,char*))_InvokeData->CallbackPtr)(Service,-2,_InvokeData->User,INVALID_DATA);
	}
	else
	{
		((void (*)(struct UPnPService*,int,void*,char*))_InvokeData->CallbackPtr)(Service,0,_InvokeData->User,SearchCaps);
	}
	MSCP_Release(Service->Parent);
	free(_InvokeData);
}
static void MSCP_Invoke_ContentDirectory_GetSortCapabilities_Sink(
void *WebReaderToken,
int IsInterrupt,
struct packetheader *header,
char *buffer,
int *p_BeginPointer,
int EndPointer,
int done,
void *_service,
void *state,
int *PAUSE)
{
	struct UPnPService *Service = (struct UPnPService*)_service;
	struct InvokeStruct *_InvokeData = (struct InvokeStruct*)state;
	int ArgLeft = 1;
	struct ILibXMLNode *xml;
	struct ILibXMLNode *__xml;
	char *tempBuffer;
	int tempBufferLength;
	char* SortCaps = NULL;
	LVL3DEBUG(char *DebugBuffer;)
	
	if(done==0){return;}
	LVL3DEBUG(DebugBuffer = (char*)malloc(EndPointer+1);)
	LVL3DEBUG(memcpy(DebugBuffer,buffer,EndPointer);)
	LVL3DEBUG(DebugBuffer[EndPointer]=0;)
	LVL3DEBUG(printf("\r\n SOAP Recieved:\r\n%s\r\n",DebugBuffer);)
	LVL3DEBUG(free(DebugBuffer);)
	if(_InvokeData->CallbackPtr==NULL)
	{
		MSCP_Release(Service->Parent);
		free(_InvokeData);
		return;
	}
	else
	{
		if(header==NULL)
		{
			/* Connection Failed */
			((void (*)(struct UPnPService*,int,void*,char*))_InvokeData->CallbackPtr)(Service,IsInterrupt==0?-1:IsInterrupt,_InvokeData->User,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
		else if(header->StatusCode!=200)
		{
			/* SOAP Fault */
			((void (*)(struct UPnPService*,int,void*,char*))_InvokeData->CallbackPtr)(Service,MSCP_GetErrorCode(buffer,EndPointer-(*p_BeginPointer)),_InvokeData->User,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
	}
	
	__xml = xml = ILibParseXML(buffer,0,EndPointer-(*p_BeginPointer));
	ILibProcessXMLNodeList(xml);
	while(xml!=NULL)
	{
		if(xml->NameLength==27 && memcmp(xml->Name,"GetSortCapabilitiesResponse",27)==0)
		{
			xml = xml->Next;
			while(xml!=NULL)
			{
				if(xml->NameLength==8 && memcmp(xml->Name,"SortCaps",8) == 0)
				{
					--ArgLeft;
					tempBufferLength = ILibReadInnerXML(xml,&tempBuffer);
					if(tempBufferLength!=0)
					{
						tempBuffer[tempBufferLength] = '\0';
						SortCaps = tempBuffer;
						ILibInPlaceXmlUnEscape(SortCaps);
					}
				}
				xml = xml->Peer;
			}
		}
		if(xml!=NULL) {xml = xml->Next;}
	}
	ILibDestructXMLNodeList(__xml);
	
	if(ArgLeft!=0)
	{
		((void (*)(struct UPnPService*,int,void*,char*))_InvokeData->CallbackPtr)(Service,-2,_InvokeData->User,INVALID_DATA);
	}
	else
	{
		((void (*)(struct UPnPService*,int,void*,char*))_InvokeData->CallbackPtr)(Service,0,_InvokeData->User,SortCaps);
	}
	MSCP_Release(Service->Parent);
	free(_InvokeData);
}
static void MSCP_Invoke_ContentDirectory_GetSystemUpdateID_Sink(
void *WebReaderToken,
int IsInterrupt,
struct packetheader *header,
char *buffer,
int *p_BeginPointer,
int EndPointer,
int done,
void *_service,
void *state,
int *PAUSE)
{
	struct UPnPService *Service = (struct UPnPService*)_service;
	struct InvokeStruct *_InvokeData = (struct InvokeStruct*)state;
	int ArgLeft = 1;
	struct ILibXMLNode *xml;
	struct ILibXMLNode *__xml;
	char *tempBuffer;
	int tempBufferLength;
	unsigned long TempULong;
	unsigned int Id = 0;
	LVL3DEBUG(char *DebugBuffer;)
	
	if(done==0){return;}
	LVL3DEBUG(DebugBuffer = (char*)malloc(EndPointer+1);)
	LVL3DEBUG(memcpy(DebugBuffer,buffer,EndPointer);)
	LVL3DEBUG(DebugBuffer[EndPointer]=0;)
	LVL3DEBUG(printf("\r\n SOAP Recieved:\r\n%s\r\n",DebugBuffer);)
	LVL3DEBUG(free(DebugBuffer);)
	if(_InvokeData->CallbackPtr==NULL)
	{
		MSCP_Release(Service->Parent);
		free(_InvokeData);
		return;
	}
	else
	{
		if(header==NULL)
		{
			/* Connection Failed */
			((void (*)(struct UPnPService*,int,void*,unsigned int))_InvokeData->CallbackPtr)(Service,IsInterrupt==0?-1:IsInterrupt,_InvokeData->User,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
		else if(header->StatusCode!=200)
		{
			/* SOAP Fault */
			((void (*)(struct UPnPService*,int,void*,unsigned int))_InvokeData->CallbackPtr)(Service,MSCP_GetErrorCode(buffer,EndPointer-(*p_BeginPointer)),_InvokeData->User,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
	}
	
	__xml = xml = ILibParseXML(buffer,0,EndPointer-(*p_BeginPointer));
	ILibProcessXMLNodeList(xml);
	while(xml!=NULL)
	{
		if(xml->NameLength==25 && memcmp(xml->Name,"GetSystemUpdateIDResponse",25)==0)
		{
			xml = xml->Next;
			while(xml!=NULL)
			{
				if(xml->NameLength==2 && memcmp(xml->Name,"Id",2) == 0)
				{
					--ArgLeft;
					tempBufferLength = ILibReadInnerXML(xml,&tempBuffer);
					if(ILibGetULong(tempBuffer,tempBufferLength,&TempULong)==0)
					{
						Id = (unsigned int) TempULong;
					}
				}
				xml = xml->Peer;
			}
		}
		if(xml!=NULL) {xml = xml->Next;}
	}
	ILibDestructXMLNodeList(__xml);
	
	if(ArgLeft!=0)
	{
		((void (*)(struct UPnPService*,int,void*,unsigned int))_InvokeData->CallbackPtr)(Service,-2,_InvokeData->User,INVALID_DATA);
	}
	else
	{
		((void (*)(struct UPnPService*,int,void*,unsigned int))_InvokeData->CallbackPtr)(Service,0,_InvokeData->User,Id);
	}
	MSCP_Release(Service->Parent);
	free(_InvokeData);
}

static void MSCP_Invoke_Registrar_IsAuthorized_Sink(
void *WebReaderToken,
int IsInterrupt,
struct packetheader *header,
char *buffer,
int *p_BeginPointer,
int EndPointer,
int done,
void *_service,
void *state,
int *PAUSE)
{
	struct UPnPService *Service = (struct UPnPService*)_service;
	struct InvokeStruct *_InvokeData = (struct InvokeStruct*)state;
	int ArgLeft = 1;
	struct ILibXMLNode *xml;
	struct ILibXMLNode *__xml;
	char *tempBuffer;
	int tempBufferLength;
	long TempLong;
	int Result = 0;
	LVL3DEBUG(char *DebugBuffer;)
	
	if(done==0){return;}
	LVL3DEBUG(DebugBuffer = (char*)malloc(EndPointer+1);)
	LVL3DEBUG(memcpy(DebugBuffer,buffer,EndPointer);)
	LVL3DEBUG(DebugBuffer[EndPointer]=0;)
	LVL3DEBUG(printf("\r\n SOAP Recieved:\r\n%s\r\n",DebugBuffer);)
	LVL3DEBUG(free(DebugBuffer);)
	if(_InvokeData->CallbackPtr==NULL)
	{
		MSCP_Release(Service->Parent);
		free(_InvokeData);
		return;
	}
	else
	{
		if(header==NULL)
		{
			/* Connection Failed */
			((void (*)(struct UPnPService*,int,void*,int))_InvokeData->CallbackPtr)(Service,IsInterrupt==0?-1:IsInterrupt,_InvokeData->User,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
		else if(header->StatusCode!=200)
		{
			/* SOAP Fault */
			((void (*)(struct UPnPService*,int,void*,int))_InvokeData->CallbackPtr)(Service,MSCP_GetErrorCode(buffer,EndPointer-(*p_BeginPointer)),_InvokeData->User,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
	}
	
	__xml = xml = ILibParseXML(buffer,0,EndPointer-(*p_BeginPointer));
	ILibProcessXMLNodeList(xml);
	while(xml!=NULL)
	{
		if(xml->NameLength==20 && memcmp(xml->Name,"IsAuthorizedResponse",20)==0)
		{
			xml = xml->Next;
			while(xml!=NULL)
			{
				if(xml->NameLength==6 && memcmp(xml->Name,"Result",6) == 0)
				{
					--ArgLeft;
					tempBufferLength = ILibReadInnerXML(xml,&tempBuffer);
					if(ILibGetLong(tempBuffer,tempBufferLength,&TempLong)==0)
					{
						Result = (int) TempLong;
					}
				}
				xml = xml->Peer;
			}
		}
		if(xml!=NULL) {xml = xml->Next;}
	}
	ILibDestructXMLNodeList(__xml);
	
	if(ArgLeft!=0)
	{
		((void (*)(struct UPnPService*,int,void*,int))_InvokeData->CallbackPtr)(Service,-2,_InvokeData->User,INVALID_DATA);
	}
	else
	{
		((void (*)(struct UPnPService*,int,void*,int))_InvokeData->CallbackPtr)(Service,0,_InvokeData->User,Result);
	}
	MSCP_Release(Service->Parent);
	free(_InvokeData);
}

static void MSCP_Invoke_Registrar_IsValidated_Sink(
void *WebReaderToken,
int IsInterrupt,
struct packetheader *header,
char *buffer,
int *p_BeginPointer,
int EndPointer,
int done,
void *_service,
void *state,
int *PAUSE)
{
	struct UPnPService *Service = (struct UPnPService*)_service;
	struct InvokeStruct *_InvokeData = (struct InvokeStruct*)state;
	int ArgLeft = 1;
	struct ILibXMLNode *xml;
	struct ILibXMLNode *__xml;
	char *tempBuffer;
	int tempBufferLength;
	long TempLong;
	int Result = 0;
	LVL3DEBUG(char *DebugBuffer;)
	
	if(done==0){return;}
	LVL3DEBUG(DebugBuffer = (char*)malloc(EndPointer+1);)
	LVL3DEBUG(memcpy(DebugBuffer,buffer,EndPointer);)
	LVL3DEBUG(DebugBuffer[EndPointer]=0;)
	LVL3DEBUG(printf("\r\n SOAP Recieved:\r\n%s\r\n",DebugBuffer);)
	LVL3DEBUG(free(DebugBuffer);)
	if(_InvokeData->CallbackPtr==NULL)
	{
		MSCP_Release(Service->Parent);
		free(_InvokeData);
		return;
	}
	else
	{
		if(header==NULL)
		{
			/* Connection Failed */
			((void (*)(struct UPnPService*,int,void*,int))_InvokeData->CallbackPtr)(Service,IsInterrupt==0?-1:IsInterrupt,_InvokeData->User,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
		else if(header->StatusCode!=200)
		{
			/* SOAP Fault */
			((void (*)(struct UPnPService*,int,void*,int))_InvokeData->CallbackPtr)(Service,MSCP_GetErrorCode(buffer,EndPointer-(*p_BeginPointer)),_InvokeData->User,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
	}
	
	__xml = xml = ILibParseXML(buffer,0,EndPointer-(*p_BeginPointer));
	ILibProcessXMLNodeList(xml);
	while(xml!=NULL)
	{
		if(xml->NameLength==19 && memcmp(xml->Name,"IsValidatedResponse",19)==0)
		{
			xml = xml->Next;
			while(xml!=NULL)
			{
				if(xml->NameLength==6 && memcmp(xml->Name,"Result",6) == 0)
				{
					--ArgLeft;
					tempBufferLength = ILibReadInnerXML(xml,&tempBuffer);
					if(ILibGetLong(tempBuffer,tempBufferLength,&TempLong)==0)
					{
						Result = (int) TempLong;
					}
				}
				xml = xml->Peer;
			}
		}
		if(xml!=NULL) {xml = xml->Next;}
	}
	ILibDestructXMLNodeList(__xml);
	
	if(ArgLeft!=0)
	{
		((void (*)(struct UPnPService*,int,void*,int))_InvokeData->CallbackPtr)(Service,-2,_InvokeData->User,INVALID_DATA);
	}
	else
	{
		((void (*)(struct UPnPService*,int,void*,int))_InvokeData->CallbackPtr)(Service,0,_InvokeData->User,Result);
	}
	MSCP_Release(Service->Parent);
	free(_InvokeData);
}

static void MSCP_Invoke_Registrar_RegisterDevice_Sink(
void *WebReaderToken,
int IsInterrupt,
struct packetheader *header,
char *buffer,
int *p_BeginPointer,
int EndPointer,
int done,
void *_service,
void *state,
int *PAUSE)
{
	struct UPnPService *Service = (struct UPnPService*)_service;
	struct InvokeStruct *_InvokeData = (struct InvokeStruct*)state;
	int ArgLeft = 1;
	struct ILibXMLNode *xml;
	struct ILibXMLNode *__xml;
	char *tempBuffer;
	int tempBufferLength;
	unsigned char* __RegistrationRespMsg = NULL;
	int __RegistrationRespMsgLength=0;
	LVL3DEBUG(char *DebugBuffer;)
	
	if(done==0){return;}
	LVL3DEBUG(DebugBuffer = (char*)malloc(EndPointer+1);)
	LVL3DEBUG(memcpy(DebugBuffer,buffer,EndPointer);)
	LVL3DEBUG(DebugBuffer[EndPointer]=0;)
	LVL3DEBUG(printf("\r\n SOAP Recieved:\r\n%s\r\n",DebugBuffer);)
	LVL3DEBUG(free(DebugBuffer);)
	if(_InvokeData->CallbackPtr==NULL)
	{
		MSCP_Release(Service->Parent);
		free(_InvokeData);
		return;
	}
	else
	{
		if(header==NULL)
		{
			/* Connection Failed */
			((void (*)(struct UPnPService*,int,void*,unsigned char*))_InvokeData->CallbackPtr)(Service,IsInterrupt==0?-1:IsInterrupt,_InvokeData->User,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
		else if(header->StatusCode!=200)
		{
			/* SOAP Fault */
			((void (*)(struct UPnPService*,int,void*,unsigned char*))_InvokeData->CallbackPtr)(Service,MSCP_GetErrorCode(buffer,EndPointer-(*p_BeginPointer)),_InvokeData->User,INVALID_DATA);
			MSCP_Release(Service->Parent);
			free(_InvokeData);
			return;
		}
	}
	
	__xml = xml = ILibParseXML(buffer,0,EndPointer-(*p_BeginPointer));
	ILibProcessXMLNodeList(xml);
	while(xml!=NULL)
	{
		if(xml->NameLength==22 && memcmp(xml->Name,"RegisterDeviceResponse",22)==0)
		{
			xml = xml->Next;
			while(xml!=NULL)
			{
				if(xml->NameLength==19 && memcmp(xml->Name,"RegistrationRespMsg",19) == 0)
				{
					--ArgLeft;
					tempBufferLength = ILibReadInnerXML(xml,&tempBuffer);
					__RegistrationRespMsgLength=ILibBase64Decode(tempBuffer,tempBufferLength,&__RegistrationRespMsg);
				}
				xml = xml->Peer;
			}
		}
		if(xml!=NULL) {xml = xml->Next;}
	}
	ILibDestructXMLNodeList(__xml);
	
	if(ArgLeft!=0)
	{
		((void (*)(struct UPnPService*,int,void*,unsigned char*,int))_InvokeData->CallbackPtr)(Service,-2,_InvokeData->User,INVALID_DATA,INVALID_DATA);
	}
	else
	{
		((void (*)(struct UPnPService*,int,void*,unsigned char*,int))_InvokeData->CallbackPtr)(Service,0,_InvokeData->User,__RegistrationRespMsg,__RegistrationRespMsgLength);
	}
	MSCP_Release(Service->Parent);
	free(_InvokeData);
	if(__RegistrationRespMsg!=NULL)
	{
		free(__RegistrationRespMsg);
	}
}

void MSCP_Invoke_ContentDirectory_Browse(struct UPnPService *service,void (*CallbackPtr)(struct UPnPService*,int,void*,char*,unsigned int,unsigned int,unsigned int), void* user, char* unescaped_ObjectID, char* unescaped_BrowseFlag, char* unescaped_Filter, unsigned int StartingIndex, unsigned int RequestedCount, char* unescaped_SortCriteria)
{
	int headerLength;
	char *headerBuffer;
	char *SoapBodyTemplate;
	char* buffer;
	int bufferLength;
	char* IP;
	unsigned short Port;
	char* Path;
	struct sockaddr_in addr;
	struct InvokeStruct *invoke_data = (struct InvokeStruct*)malloc(sizeof(struct InvokeStruct));
	char* ObjectID;
	char* BrowseFlag;
	char* Filter;
	char* SortCriteria;
	
	if(service==NULL)
	{
		free(invoke_data);
		return;
	}
	
	
	ObjectID = (char*)malloc(1+ILibXmlEscapeLength(unescaped_ObjectID));
	ILibXmlEscape(ObjectID,unescaped_ObjectID);
	BrowseFlag = (char*)malloc(1+ILibXmlEscapeLength(unescaped_BrowseFlag));
	ILibXmlEscape(BrowseFlag,unescaped_BrowseFlag);
	Filter = (char*)malloc(1+ILibXmlEscapeLength(unescaped_Filter));
	ILibXmlEscape(Filter,unescaped_Filter);
	SortCriteria = (char*)malloc(1+ILibXmlEscapeLength(unescaped_SortCriteria));
	ILibXmlEscape(SortCriteria,unescaped_SortCriteria);
	buffer = (char*)malloc((int)strlen(service->ServiceType)+(int)strlen(ObjectID)+(int)strlen(BrowseFlag)+(int)strlen(Filter)+(int)strlen(SortCriteria)+420);
	SoapBodyTemplate = "%sBrowse xmlns:u=\"%s\"><ObjectID>%s</ObjectID><BrowseFlag>%s</BrowseFlag><Filter>%s</Filter><StartingIndex>%u</StartingIndex><RequestedCount>%u</RequestedCount><SortCriteria>%s</SortCriteria></u:Browse%s";
	bufferLength = sprintf(buffer,SoapBodyTemplate,UPNPCP_SOAP_BodyHead,service->ServiceType, ObjectID, BrowseFlag, Filter, StartingIndex, RequestedCount, SortCriteria,UPNPCP_SOAP_BodyTail);
	LVL3DEBUG(printf("\r\n SOAP Sent: \r\n%s\r\n",buffer);)
	free(ObjectID);
	free(BrowseFlag);
	free(Filter);
	free(SortCriteria);
	
	MSCP_AddRef(service->Parent);
	ILibParseUri(service->ControlURL,&IP,&Port,&Path);
	
	headerBuffer = (char*)malloc(169 + (int)strlen(MSCP_PLATFORM) + (int)strlen(Path) + (int)strlen(IP) + (int)strlen(service->ServiceType));
	headerLength = sprintf(headerBuffer,UPNPCP_SOAP_Header,Path,IP,Port,MSCP_PLATFORM,service->ServiceType,"Browse",bufferLength);
	
	memset((char *)&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = inet_addr(IP);
	addr.sin_port = htons(Port);
	
	invoke_data->CallbackPtr = CallbackPtr;
	invoke_data->User = user;
	ILibWebClient_PipelineRequestEx(
	((struct MSCP_CP*)service->Parent->CP)->HTTP,
	&addr,
	headerBuffer,
	headerLength,
	0,
	buffer,
	bufferLength,
	0,
	&MSCP_Invoke_ContentDirectory_Browse_Sink,
	service,
	invoke_data);
	
	free(IP);
	free(Path);
}
void MSCP_Invoke_ContentDirectory_GetSearchCapabilities(struct UPnPService *service,void (*CallbackPtr)(struct UPnPService*,int,void*,char*), void* user)
{
	int headerLength;
	char *headerBuffer;
	char *SoapBodyTemplate;
	char* buffer;
	int bufferLength;
	char* IP;
	unsigned short Port;
	char* Path;
	struct sockaddr_in addr;
	struct InvokeStruct *invoke_data = (struct InvokeStruct*)malloc(sizeof(struct InvokeStruct));
	
	if(service==NULL)
	{
		free(invoke_data);
		return;
	}
	
	
	buffer = (char*)malloc((int)strlen(service->ServiceType)+274);
	SoapBodyTemplate = "%sGetSearchCapabilities xmlns:u=\"%s\"></u:GetSearchCapabilities%s";
	bufferLength = sprintf(buffer,SoapBodyTemplate,UPNPCP_SOAP_BodyHead,service->ServiceType,UPNPCP_SOAP_BodyTail);
	LVL3DEBUG(printf("\r\n SOAP Sent: \r\n%s\r\n",buffer);)
	
	MSCP_AddRef(service->Parent);
	ILibParseUri(service->ControlURL,&IP,&Port,&Path);
	
	headerBuffer = (char*)malloc(184 + (int)strlen(MSCP_PLATFORM) + (int)strlen(Path) + (int)strlen(IP) + (int)strlen(service->ServiceType));
	headerLength = sprintf(headerBuffer,UPNPCP_SOAP_Header,Path,IP,Port,MSCP_PLATFORM,service->ServiceType,"GetSearchCapabilities",bufferLength);
	
	memset((char *)&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = inet_addr(IP);
	addr.sin_port = htons(Port);
	
	invoke_data->CallbackPtr = CallbackPtr;
	invoke_data->User = user;
	ILibWebClient_PipelineRequestEx(
	((struct MSCP_CP*)service->Parent->CP)->HTTP,
	&addr,
	headerBuffer,
	headerLength,
	0,
	buffer,
	bufferLength,
	0,
	&MSCP_Invoke_ContentDirectory_GetSearchCapabilities_Sink,
	service,
	invoke_data);
	
	free(IP);
	free(Path);
}
void MSCP_Invoke_ContentDirectory_GetSortCapabilities(struct UPnPService *service,void (*CallbackPtr)(struct UPnPService*,int,void*,char*), void* user)
{
	int headerLength;
	char *headerBuffer;
	char *SoapBodyTemplate;
	char* buffer;
	int bufferLength;
	char* IP;
	unsigned short Port;
	char* Path;
	struct sockaddr_in addr;
	struct InvokeStruct *invoke_data = (struct InvokeStruct*)malloc(sizeof(struct InvokeStruct));
	
	if(service==NULL)
	{
		free(invoke_data);
		return;
	}
	
	
	buffer = (char*)malloc((int)strlen(service->ServiceType)+270);
	SoapBodyTemplate = "%sGetSortCapabilities xmlns:u=\"%s\"></u:GetSortCapabilities%s";
	bufferLength = sprintf(buffer,SoapBodyTemplate,UPNPCP_SOAP_BodyHead,service->ServiceType,UPNPCP_SOAP_BodyTail);
	LVL3DEBUG(printf("\r\n SOAP Sent: \r\n%s\r\n",buffer);)
	
	MSCP_AddRef(service->Parent);
	ILibParseUri(service->ControlURL,&IP,&Port,&Path);
	
	headerBuffer = (char*)malloc(182 + (int)strlen(MSCP_PLATFORM) + (int)strlen(Path) + (int)strlen(IP) + (int)strlen(service->ServiceType));
	headerLength = sprintf(headerBuffer,UPNPCP_SOAP_Header,Path,IP,Port,MSCP_PLATFORM,service->ServiceType,"GetSortCapabilities",bufferLength);
	
	memset((char *)&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = inet_addr(IP);
	addr.sin_port = htons(Port);
	
	invoke_data->CallbackPtr = CallbackPtr;
	invoke_data->User = user;
	ILibWebClient_PipelineRequestEx(
	((struct MSCP_CP*)service->Parent->CP)->HTTP,
	&addr,
	headerBuffer,
	headerLength,
	0,
	buffer,
	bufferLength,
	0,
	&MSCP_Invoke_ContentDirectory_GetSortCapabilities_Sink,
	service,
	invoke_data);
	
	free(IP);
	free(Path);
}
void MSCP_Invoke_ContentDirectory_GetSystemUpdateID(struct UPnPService *service,void (*CallbackPtr)(struct UPnPService*,int,void*,unsigned int), void* user)
{
	int headerLength;
	char *headerBuffer;
	char *SoapBodyTemplate;
	char* buffer;
	int bufferLength;
	char* IP;
	unsigned short Port;
	char* Path;
	struct sockaddr_in addr;
	struct InvokeStruct *invoke_data = (struct InvokeStruct*)malloc(sizeof(struct InvokeStruct));
	
	if(service==NULL)
	{
		free(invoke_data);
		return;
	}
	
	
	buffer = (char*)malloc((int)strlen(service->ServiceType)+266);
	SoapBodyTemplate = "%sGetSystemUpdateID xmlns:u=\"%s\"></u:GetSystemUpdateID%s";
	bufferLength = sprintf(buffer,SoapBodyTemplate,UPNPCP_SOAP_BodyHead,service->ServiceType,UPNPCP_SOAP_BodyTail);
	LVL3DEBUG(printf("\r\n SOAP Sent: \r\n%s\r\n",buffer);)
	
	MSCP_AddRef(service->Parent);
	ILibParseUri(service->ControlURL,&IP,&Port,&Path);
	
	headerBuffer = (char*)malloc(180 + (int)strlen(MSCP_PLATFORM) + (int)strlen(Path) + (int)strlen(IP) + (int)strlen(service->ServiceType));
	headerLength = sprintf(headerBuffer,UPNPCP_SOAP_Header,Path,IP,Port,MSCP_PLATFORM,service->ServiceType,"GetSystemUpdateID",bufferLength);
	
	memset((char *)&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = inet_addr(IP);
	addr.sin_port = htons(Port);
	
	invoke_data->CallbackPtr = CallbackPtr;
	invoke_data->User = user;
	ILibWebClient_PipelineRequestEx(
	((struct MSCP_CP*)service->Parent->CP)->HTTP,
	&addr,
	headerBuffer,
	headerLength,
	0,
	buffer,
	bufferLength,
	0,
	&MSCP_Invoke_ContentDirectory_GetSystemUpdateID_Sink,
	service,
	invoke_data);
	
	free(IP);
	free(Path);
}
void MSCP_Invoke_Registrar_IsAuthorized(struct UPnPService *service,void (*CallbackPtr)(struct UPnPService*,int,void*,int), void* user, char* unescaped_DeviceID)
{
	int headerLength;
	char *headerBuffer;
	char *SoapBodyTemplate;
	char* buffer;
	int bufferLength;
	char* IP;
	unsigned short Port;
	char* Path;
	struct sockaddr_in addr;
	struct InvokeStruct *invoke_data = (struct InvokeStruct*)malloc(sizeof(struct InvokeStruct));
	char* DeviceID;
	
	if(service==NULL)
	{
		free(invoke_data);
		return;
	}
	
	
	DeviceID = (char*)malloc(1+ILibXmlEscapeLength(unescaped_DeviceID));
	ILibXmlEscape(DeviceID,unescaped_DeviceID);
	buffer = (char*)malloc((int)strlen(service->ServiceType)+(int)strlen(DeviceID)+277);
	SoapBodyTemplate = "%sIsAuthorized xmlns:u=\"%s\"><DeviceID>%s</DeviceID></u:IsAuthorized%s";
	bufferLength = sprintf(buffer,SoapBodyTemplate,UPNPCP_SOAP_BodyHead,service->ServiceType, DeviceID,UPNPCP_SOAP_BodyTail);
	LVL3DEBUG(printf("\r\n SOAP Sent: \r\n%s\r\n",buffer);)
	free(DeviceID);
	
	MSCP_AddRef(service->Parent);
	ILibParseUri(service->ControlURL,&IP,&Port,&Path);
	
	headerBuffer = (char*)malloc(175 + (int)strlen(MSCP_PLATFORM) + (int)strlen(Path) + (int)strlen(IP) + (int)strlen(service->ServiceType));
	headerLength = sprintf(headerBuffer,UPNPCP_SOAP_Header,Path,IP,Port,MSCP_PLATFORM,service->ServiceType,"IsAuthorized",bufferLength);
	
	memset((char *)&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = inet_addr(IP);
	addr.sin_port = htons(Port);
	
	invoke_data->CallbackPtr = CallbackPtr;
	invoke_data->User = user;
	ILibWebClient_PipelineRequestEx(
	((struct MSCP_CP*)service->Parent->CP)->HTTP,
	&addr,
	headerBuffer,
	headerLength,
	0,
	buffer,
	bufferLength,
	0,
	&MSCP_Invoke_Registrar_IsAuthorized_Sink,
	service,
	invoke_data);
	
	free(IP);
	free(Path);
}
void MSCP_Invoke_Registrar_IsValidated(struct UPnPService *service,void (*CallbackPtr)(struct UPnPService*,int,void*,int), void* user, char* unescaped_DeviceID)
{
	int headerLength;
	char *headerBuffer;
	char *SoapBodyTemplate;
	char* buffer;
	int bufferLength;
	char* IP;
	unsigned short Port;
	char* Path;
	struct sockaddr_in addr;
	struct InvokeStruct *invoke_data = (struct InvokeStruct*)malloc(sizeof(struct InvokeStruct));
	char* DeviceID;
	
	if(service==NULL)
	{
		free(invoke_data);
		return;
	}
	
	
	DeviceID = (char*)malloc(1+ILibXmlEscapeLength(unescaped_DeviceID));
	ILibXmlEscape(DeviceID,unescaped_DeviceID);
	buffer = (char*)malloc((int)strlen(service->ServiceType)+(int)strlen(DeviceID)+275);
	SoapBodyTemplate = "%sIsValidated xmlns:u=\"%s\"><DeviceID>%s</DeviceID></u:IsValidated%s";
	bufferLength = sprintf(buffer,SoapBodyTemplate,UPNPCP_SOAP_BodyHead,service->ServiceType, DeviceID,UPNPCP_SOAP_BodyTail);
	LVL3DEBUG(printf("\r\n SOAP Sent: \r\n%s\r\n",buffer);)
	free(DeviceID);
	
	MSCP_AddRef(service->Parent);
	ILibParseUri(service->ControlURL,&IP,&Port,&Path);
	
	headerBuffer = (char*)malloc(174 + (int)strlen(MSCP_PLATFORM) + (int)strlen(Path) + (int)strlen(IP) + (int)strlen(service->ServiceType));
	headerLength = sprintf(headerBuffer,UPNPCP_SOAP_Header,Path,IP,Port,MSCP_PLATFORM,service->ServiceType,"IsValidated",bufferLength);
	
	memset((char *)&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = inet_addr(IP);
	addr.sin_port = htons(Port);
	
	invoke_data->CallbackPtr = CallbackPtr;
	invoke_data->User = user;
	ILibWebClient_PipelineRequestEx(
	((struct MSCP_CP*)service->Parent->CP)->HTTP,
	&addr,
	headerBuffer,
	headerLength,
	0,
	buffer,
	bufferLength,
	0,
	&MSCP_Invoke_Registrar_IsValidated_Sink,
	service,
	invoke_data);
	
	free(IP);
	free(Path);
}
void MSCP_Invoke_Registrar_RegisterDevice(struct UPnPService *service,void (*CallbackPtr)(struct UPnPService*,int,void*,unsigned char*,int), void* user, unsigned char* RegistrationReqMsg, int RegistrationReqMsgLength)
{
	int headerLength;
	char *headerBuffer;
	char *SoapBodyTemplate;
	char* buffer;
	int bufferLength;
	char* IP;
	unsigned short Port;
	char* Path;
	struct sockaddr_in addr;
	struct InvokeStruct *invoke_data = (struct InvokeStruct*)malloc(sizeof(struct InvokeStruct));
	char* __RegistrationReqMsg;
	int buffer_size;
	
	if(service==NULL)
	{
		free(invoke_data);
		return;
	}
	
	
	ILibBase64Encode(RegistrationReqMsg,RegistrationReqMsgLength,(unsigned char**)&__RegistrationReqMsg);
	SoapBodyTemplate = "%sRegisterDevice xmlns:u=\"%s\"><RegistrationReqMsg>%s</RegistrationReqMsg></u:RegisterDevice%s";
	buffer_size = strlen(SoapBodyTemplate) + strlen(UPNPCP_SOAP_BodyHead) +
	              strlen(service->ServiceType) + strlen(__RegistrationReqMsg) + 
		      strlen(UPNPCP_SOAP_BodyTail)+1;
	buffer = (char*)malloc(buffer_size);
	bufferLength = snprintf(buffer,buffer_size,SoapBodyTemplate,UPNPCP_SOAP_BodyHead,service->ServiceType, __RegistrationReqMsg,UPNPCP_SOAP_BodyTail);
	if (bufferLength >= buffer_size)
		fprintf(stderr,"** Error ** Registration message was truncated : length should be %d but is %d\n", bufferLength, buffer_size);
	LVL3DEBUG(printf("\r\n SOAP Sent: \r\n%s\r\n",buffer);)
	free(__RegistrationReqMsg);
	
	MSCP_AddRef(service->Parent);
	ILibParseUri(service->ControlURL,&IP,&Port,&Path);

	buffer_size = 256 + (int)strlen(MSCP_PLATFORM) + (int)strlen(Path) + (int)strlen(IP) + (int)strlen(service->ServiceType);
	headerBuffer = (char*)malloc(buffer_size);
	headerLength = snprintf(headerBuffer,buffer_size,UPNPCP_SOAP_Header,Path,IP,Port,MSCP_PLATFORM,service->ServiceType,"RegisterDevice",bufferLength);
	if (headerLength >= buffer_size)
		fprintf(stderr,"** Error ** Registration header was truncated : length should be %d but is %d\n", headerLength, buffer_size);
	
	memset((char *)&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = inet_addr(IP);
	addr.sin_port = htons(Port);
	
	invoke_data->CallbackPtr = CallbackPtr;
	invoke_data->User = user;
	ILibWebClient_PipelineRequestEx(
	((struct MSCP_CP*)service->Parent->CP)->HTTP,
	&addr,
	headerBuffer,
	headerLength,
	0,
	buffer,
	bufferLength,
	0,
	&MSCP_Invoke_Registrar_RegisterDevice_Sink,
	service,
	invoke_data);
	
	free(IP);
	free(Path);
}


void MSCP_UnSubscribeUPnPEvents(struct UPnPService *service)
{
	char *IP;
	char *Path;
	unsigned short Port;
	struct packetheader *p;
	char* TempString;
	struct sockaddr_in destaddr;
	
	if(service->SubscriptionID==NULL) {return;}
	ILibParseUri(service->SubscriptionURL,&IP,&Port,&Path);
	p = ILibCreateEmptyPacket();
	ILibSetVersion(p,"1.1",3);
	
	ILibSetDirective(p,"UNSUBSCRIBE",11,Path,(int)strlen(Path));
	
	TempString = (char*)malloc((int)strlen(IP)+7);
	sprintf(TempString,"%s:%d",IP,Port);
	
	ILibAddHeaderLine(p,"HOST",4,TempString,(int)strlen(TempString));
	free(TempString);
	
	ILibAddHeaderLine(p,"User-Agent",10,MSCP_PLATFORM,(int)strlen(MSCP_PLATFORM));
	ILibAddHeaderLine(p,"SID",3,service->SubscriptionID,(int)strlen(service->SubscriptionID));
	
	memset((char *)&destaddr, 0,sizeof(destaddr));
	destaddr.sin_family = AF_INET;
	destaddr.sin_addr.s_addr = inet_addr(IP);
	destaddr.sin_port = htons(Port);
	
	MSCP_AddRef(service->Parent);
	ILibWebClient_PipelineRequest(
	((struct MSCP_CP*)service->Parent->CP)->HTTP,
	&destaddr,
	p,
	&MSCP_OnUnSubscribeSink,
	(void*)service,
	service->Parent->CP);
	
	free(IP);
	free(Path);
}
void MSCP_SubscribeForUPnPEvents(struct UPnPService *service, void(*callbackPtr)(struct UPnPService* service,int OK))
{
	char* callback;
	char *IP;
	char *Path;
	unsigned short Port;
	struct packetheader *p;
	char* TempString;
	struct sockaddr_in destaddr;
	
	ILibParseUri(service->SubscriptionURL,&IP,&Port,&Path);
	p = ILibCreateEmptyPacket();
	ILibSetVersion(p,"1.1",3);
	
	ILibSetDirective(p,"SUBSCRIBE",9,Path,(int)strlen(Path));
	
	TempString = (char*)malloc((int)strlen(IP)+7);
	sprintf(TempString,"%s:%d",IP,Port);
	
	ILibAddHeaderLine(p,"HOST",4,TempString,(int)strlen(TempString));
	free(TempString);
	
	ILibAddHeaderLine(p,"NT",2,"upnp:event",10);
	ILibAddHeaderLine(p,"TIMEOUT",7,"Second-180",10);
	ILibAddHeaderLine(p,"User-Agent",10,MSCP_PLATFORM,(int)strlen(MSCP_PLATFORM));
	
	callback = (char*)malloc(10+(int)strlen(service->Parent->InterfaceToHost)+6+(int)strlen(Path));
	sprintf(callback,"<http://%s:%d%s>",service->Parent->InterfaceToHost,ILibWebServer_GetPortNumber(((struct MSCP_CP*)service->Parent->CP)->WebServer),Path);
	
	ILibAddHeaderLine(p,"CALLBACK",8,callback,(int)strlen(callback));
	free(callback);
	
	memset((char *)&destaddr, 0,sizeof(destaddr));
	destaddr.sin_family = AF_INET;
	destaddr.sin_addr.s_addr = inet_addr(IP);
	destaddr.sin_port = htons(Port);
	
	MSCP_AddRef(service->Parent);
	ILibWebClient_PipelineRequest(
	((struct MSCP_CP*)service->Parent->CP)->HTTP,
	&destaddr,
	p,
	&MSCP_OnSubscribeSink,
	(void*)service,
	service->Parent->CP);
	
	free(IP);
	free(Path);
}

char* MSCP_GetCustomXML_X_DLNADOC(struct UPnPDevice *d)
{
	return(MSCP_GetCustomTagFromDevice(d,X_DLNADOC_NAMESPACE, X_DLNADOC));
}









