################################################################
# Copyright (c) 2001 Sigma Designs Inc. All rights reserved.   #
# Author : Julien Soulier    				       #
# Date : 06-19-2001					       #
#							       #
# This is the base makefile for all realmagic projects         #
# 							       #
################################################################

################################
## Project specific variables ## 
################################
ifdef RUA_DIR
ifeq '$(filter -I$(RUA_DIR),$(RMCFLAGS))' ''
RMCFLAGS += -I$(RUA_DIR)
endif
endif

ifndef DEPEND_FILE
DEPEND_FILE = Makefile.d
endif

ifndef SRC_LIST_FILE
SRC_LIST_FILE = .src.lst
endif

ifndef MAKEFILE_NAME
MAKEFILE_NAME = Makefile
endif	

ifndef LOCAL_MAKEFILE
LOCAL_MAKEFILE = $(MAKEFILE_NAME)
endif

ifndef CCOMPILER
CCOMPILER := gcc
endif

ifndef CXXCOMPILER
CXXCOMPILER := g++
endif

CC := $(CCACHE) $(CROSS)$(CCOMPILER)
CXX := $(CCACHE) $(CROSS)$(CXXCOMPILER)
LD := $(CROSS)$(LD)
OBJCOPY := $(CROSS)objcopy
AR := $(CROSS)ar

ifndef LINKER
# must be = and not :=
LINKTOOL = $(CROSS)$(CCOMPILER)
# LINKTOOL used to be defined a $(CC). However, that makes LINKTOOL ccacheable, 
# which is nonsensical, because ccache only caches compilation, not linking.
else
LINKTOOL := $(CROSS)$(LINKER)
endif

LOCALDIR = $(shell pwd)

ifndef RMPREFIX
export RMPREFIX := $(LOCALDIR)/..
endif

ifeq "$(TARGET_TYPE)" "MICROCODE"
ifndef SPASM
ifdef SPASM_VERSION
#if SPASM version 2.x.x, no need to include the libraries
ifeq ($(findstring 2., $(SPASM_VERSION)), 2.)
SPASM := LANG=C perl /utils/em8xxx/spasm/$(SPASM_VERSION)/bin/spasm
else
SPASM := LANG=C perl -I /utils/em8xxx/spasm/$(SPASM_VERSION)/lib /utils/em8xxx/spasm/$(SPASM_VERSION)/bin/spasm
endif
else
$(error Please define SPASM_VERSION (ex: 1.4.18))
endif 
endif

ifndef SPBIN
ifdef SPBIN_VERSION
SPBIN := /utils/em8xxx/spbin/$(SPBIN_VERSION)/bin/spbin
else
$(error Please define SPBIN_VERSION (ex: 1.2.9))
endif
endif

ifndef SPADE
ifdef SPADE_VERSION
SPADE := /utils/em8xxx/spade/$(SPADE_VERSION)/spade
else
$(error Please define SPADE_VERSION (ex: 1.0))
endif	# SPADE_VERSION
endif 	# SPADE


endif

APIDIR := $(RMPREFIX)/externalapi
LIBDIR := $(RMPREFIX)/lib 
DISTDIR := $(RMPREFIX)/dist

MAKEARGS = -f $(MAKEFILE_NAME) ## the -f option is not passed through MAKEFLAGS
PREPROCESSORFLAGS += $(RMCFLAGS) -I. -D_REENTRANT
ifdef base_dir
ifeq '$(filter -I$(base_dir), $(PREPROCESSORFLAGS))' ''
PREPROCESSORFLAGS += -I$(base_dir)
endif
endif

#########################
## Compilation options ##
#########################

SPASMFLAGS += -d __ASSEMBLY__=1
SPBINFLAGS += -l -p -m -sa -od 0x10000

ifeq ($(findstring -DWITH_UCODE_BOOTLOADER=1, $(RMCFLAGS)),-DWITH_UCODE_BOOTLOADER=1)
SPBINFLAGS += -c
endif

# It has to be a = not a :=
ASFLAGS = $(KERNELSTYLE) $(PREPROCESSORFLAGS) -D__ASSEMBLY__=1

ifeq ($(findstring leakchecker, $(COMPILKIND)), leakchecker)
EXEFLAGS += `glib-config --libs glib`
PREPROCESSORFLAGS += -DUSE_LEAK_CHECKER=1
endif

ifeq ($(findstring withthreads, $(COMPILKIND)), withthreads)
PREPROCESSORFLAGS += -DWITH_THREADS=1
LDFLAGS += -lpthread 
endif

ifeq ($(findstring withfeeblemm, $(COMPILKIND)), withfeeblemm)
# when memory is full, RMMalloc returns NULL and does not RMPanic
# when RMFree is called at a bad address, does not RMPanic
# useful to stresstest the memory manager
PREPROCESSORFLAGS += -DWITH_FEEBLEMM=1
endif

ifeq ($(findstring withdl, $(COMPILKIND)), withdl)
EXEFLAGS += -ldl -rdynamic
endif

ifeq ($(findstring kernel, $(COMPILKIND)), kernel)	## kernel compilation

PREPROCESSORFLAGS += -D__KERNEL__ -DMODULE -DEXPORT_SYMTAB

ifeq ($(findstring uclinux, $(COMPILKIND)), uclinux)	## uCLinux compilation

ifndef MUM_KI
  ifneq '$(filter -DEM86XX_MODE=EM86XX_MODEID_STANDALONE, $(RMCFLAGS))' ''
    ifndef UCLINUX_KERNEL
      ifneq '$(filter -DEM86XX_CHIP=EM86XX_CHIPID_TANGO2, $(RMCFLAGS))' ''
UCLINUX_KERNEL := /utils/thirdparty/kernelsrc/linux-mips
      else  # ! EM86XX_CHIPID_TANGO2
UCLINUX_KERNEL := /utils/thirdparty/kernelsrc/linux-arm
      endif # EM86XX_CHIPID_TANGO2
$(warning Notice: neither MUM_KI nor UCLINUX_KERNEL are defined. Defaulting UCLINUX_KERNEL to $(UCLINUX_KERNEL).)
    endif # UCLINUX_KERNEL
MUM_KI := -I$(UCLINUX_KERNEL)/include
  else  # ! EM86XX_MODEID_STANDALONE
    ifneq '$(filter -DEM86XX_MODE=EM86XX_MODEID_WITHHOST, $(RMCFLAGS))' ''
      ifndef LINUX_KERNEL
$(warning Notice: neither MUM_KI nor UCLINUX_KERNEL are defined. MUM_KI will *NOT* be defined.)
      else   # LINUX_KERNEL
MUM_KI := -I$(LINUX_KERNEL)/include
      endif  # LINUX_KERNEL
    endif # EM86XX_MODE=EM86XX_MODEID_WITHHOST
  endif # EM86XX_MODEID_STANDALONE
endif # MUM_KI

ifndef KERNELSTYLE
ifneq "$(filter -DEM86XX_CHIP=EM86XX_CHIPID_TANGO2,$(RMCFLAGS))" ""
KERNELSTYLE= \
	-fomit-frame-pointer \
	-fno-strict-aliasing \
	-fno-common \
	-G 0 \
	-mno-abicalls \
	-fno-pic \
	-finline-limit=100000 \
	-mabi=32 \
	-Wa,-mips32r2 -Wa,-mfix7000 -Wa,--trap \
	-fno-builtin \
	-mlong-calls \
	-pipe 
else
KERNELSTYLE= \
        -fomit-frame-pointer \
        -fno-strict-aliasing \
        -fno-common \
	-pipe \
	-fno-builtin
endif # tango2 alternative

PREPROCESSORFLAGS += -D__linux__
endif

else # no uclinux

ifndef MUM_KI
MUM_KI=-I/usr/src/linux-$(shell uname -r)/include -I/usr/src/linux-2.4/include
endif

ifndef KERNELSTYLE
KERNELSTYLE= \
        -fomit-frame-pointer \
        -fno-strict-aliasing \
        -fno-common \
	-mpreferred-stack-boundary=2 \
	-pipe 
endif

endif # linux/uclinux alternative done

## Switch for fascist warnings.
ifndef DISABLE_WARNINGS
CWARNINGS+= \
	-Wno-import \
        -Wunused \
        -Wimplicit \
        -Wmain \
        -Wreturn-type \
        -Wswitch \
        -Wtrigraphs \
        -Wchar-subscripts \
        -Wparentheses \
        -Wpointer-arith \
        -Wcast-align \
	-Wuninitialized -O
else
CWARNINGS+= -O2
endif

CFLAGS += $(KERNELSTYLE) $(MUM_KI)

else # user compilation

## Switch for fascist warnings.
ifndef DISABLE_WARNINGS
COMMONWARNINGS+= \
	-Wundef \
	-Wall \
	-Wchar-subscripts \
	-Wsign-compare \
	-Wuninitialized -O
#	-Werror 

CWARNINGS:=$(COMMONWARNINGS) -Wnested-externs -Wmissing-declarations -Wmissing-prototypes
CXXWARNINGS:=$(COMMONWARNINGS) -fcheck-new 
endif

### if libc heades are needed
ifndef USE_STD_LIB
	PREPROCESSORFLAGS += -nostdinc
endif

# If standalone, we may need to enable XIP settings
ifeq ($(findstring -DEM86XX_MODE=EM86XX_MODEID_STANDALONE, $(RMCFLAGS)),-DEM86XX_MODE=EM86XX_MODEID_STANDALONE)
ifeq ($(findstring -DWITH_XIP=1, $(RMCFLAGS)), -DWITH_XIP=1)
ifneq ($(findstring -DINIRQHANDLER=1, $(RMCFLAGS)), -DINIRQHANDLER=1)
CFLAGS += -D__PIC__ -fpic -msingle-pic-base
CXXFLAGS += -D__PIC__ -fpic -msingle-pic-base
LDFLAGS += -fpic -msingle-pic-base 
OBJECTFLAGS += -q 
endif
endif
endif

endif # kernel/user alternative done

## Switch to release/debug compilation ##
# The macro to do this is called ``_DEBUG'' to fit how Windows does. Sorry.
# Additionally, the C specification says that assert is defined to a nop
# if and only if NDEBUG is defined. We don't want assertions in release
# code.
ifeq ($(findstring release, $(COMPILKIND)), release)

PREPROCESSORFLAGS += -U_DEBUG -DNDEBUG
CFLAGS += -O2
CXXFLAGS += -O2

else

PREPROCESSORFLAGS += -D_DEBUG=1 -UNDEBUG

ifneq ($(findstring kernel, $(COMPILKIND)), kernel) # -g only in user compilations
ifeq "$(CCOMPILER)" "gcc"
ifeq "$(CROSS)" ""
CFLAGS += -gdwarf-2
CXXFLAGS += -gdwarf-2
else
CFLAGS += -g
CXXFLAGS += -g
endif # end of CROSS switch
endif # end of COMPILER switch
endif # -g only in user compilations
endif # end of release switch

### Platform Option
ifdef RMPLATFORM
PREPROCESSORFLAGS += -DRMPLATFORM=$(RMPLATFORM)
endif

CFLAGS += $(PREPROCESSORFLAGS) $(CWARNINGS) 
## kernel code is always in C
CXXFLAGS += $(PREPROCESSORFLAGS) $(CXXWARNINGS) 

#########################################################
## select compilation according to COMPILKIND variable ##
#########################################################

TARGET_TYPE := $(strip $(TARGET_TYPE))

ifeq "$(TARGET_TYPE)" "LIBRARY"
ifeq ($(findstring static, $(COMPILKIND)), static)
TARGET_TYPE :=  OBJECT_LIBRARY
else
TARGET_TYPE :=  SHARED_LIBRARY
endif
endif

ifeq ($(findstring static, $(COMPILKIND)), static)
STATIC_LINKS_WITH += $(LINKS_WITH)
LDFLAGS += -static
else
ifeq ($(findstring implicit, $(COMPILKIND)), implicit)
IMPLICIT_LINKS_WITH := $(LINKS_WITH)
else
DEPENDS_ON += $(LINKS_WITH)
endif
endif

####################################################
## Describes the rule to make each type of target ##
####################################################
OBJECT_LIBRARY = $(AR) rcs $@ $^
RELOCATABLE_OBJECT = $(LD) -r $(OBJECTFLAGS) $^ -o $@
SHARED_LIBRARY = $(LINKTOOL) -shared $^ $(LDFLAGS) -o $@
EXECUTABLE = $(LINKTOOL) $^ $(LDFLAGS) $(EXEFLAGS) -o $@ && chmod a+x $@
DEPEND_COMMAND = $(CROSS)gcc $(MUM_KI) $(PREPROCESSORFLAGS) -MM -E `cat $(SRC_LIST_FILE)`
# mips compiler generates a large quantity of informative and debug sections. Keep only the required:
BINARY_FILE = $(OBJCOPY) --remove-section=.reginfo -O binary $< $@

ifdef DEBUGUCODEBUILD
define MICROCODE
	$(SPASM) $(SPASMFLAGS) $^ -o $(UCODE_NAME).bin
	$(SPBIN) $(SPBINFLAGS) -h $(UCODE_NAME).bin
	$(SPADE) $(SPADEFLAGS) -bin $(UCODE_NAME).bin -o $(UCODE_NAME)_debug.bin
	$(call BIN_TO_H, $(UCODE_NAME), $@)
endef
else	# DEBUGUCODEBUILD
define MICROCODE
	$(SPASM) $(SPASMFLAGS) $^ -o $(UCODE_NAME).bin
	$(SPBIN) $(SPBINFLAGS) -h $(UCODE_NAME).bin
	$(call BIN_TO_H, $(UCODE_NAME), $@)
endef
endif	# DEBUGUCODEBUILD

###################################
## Describes ugly shell commands ##
###################################

COPY_FILE_LIST = sed 's/^.*://' | sed 's/\\$$//' | tr ' ' '\n' | sed '/^$$/d' | \
		sed '/^\//d' | \
		sed 's/^\([^/]\)/$(subst /,\/,$(LOCALDIR))\/\1/' | \
		sed '{ : rm_dotdot s/[^/]*\/\.\.\/// ; t rm_dotdot }' | sort -u | \
		xargs cp -f --parents --target-directory=$(1)

COPY_FILE_LIST2 = sed 's/^.*://' | sed 's/\\$$//' | tr ' ' '\n' | sed '/^$$/d' | \
        sed '/^\//d' | sort -u | \
        xargs -l1 cp -f --parents --target-directory=$(1)

REDUCE_DIR_TREE = cd $(1) ; x=1 ; while [ $$x -eq 1 ] ; \
		do x=0 ; [ `ls -1 | wc -l` -eq 1 -a -d `ls -1 | head -n 1` ] && \
		y=`ls -1` && mv $$y/* . && rmdir $$y && x=1 ; \
		done ; exit 0

BIN_TO_H = (echo "RMuint8 $(1)_bin[" ; \
	wc -c < $(1)_release.bin ; \
	echo "] = {"; \
	od -v -An -tx1 $(1)_release.bin | perl -pe 's/([0-9a-fA-f]+)/0x\1,/g'; \
	echo "};") > $(2)

#########################################################
## Compute the local object files and the dependencies ## 
#########################################################

SRC := $(strip $(SRC))
MAIN_SRC := $(strip $(MAIN_SRC))
OBJ := $(addsuffix .o, $(basename $(SRC)))
ifeq "$(TARGET_TYPE)" "MICROCODE"
# NOTE: need the space there ".o " to avoid replacing aac.ovlprog.asm to aac_$VERSION.ovlprog_$VERSION.o
OBJ := $(OBJ) # end of line
OBJ := $(subst .o ,_$(subst .,_,$(SPASM_VERSION)).o , $(OBJ))
endif
MAIN_OBJ := $(addsuffix .o, $(basename $(MAIN_SRC)))
EXE := $(basename $(MAIN_SRC))
SUBDIR := $(dir $(shell echo '$(DEPENDS_ON) $(STATIC_LINKS_WITH) $(IMPLICIT_LINKS_WITH)' | sed -e 's/([^(]*)//g'| sed -e 's/)//g' ))
ifdef NO_LIB_SUBDIR
SUBDIR := $(filter-out ../lib/, $(SUBDIR))
endif

ifneq ($(findstring clean, $(MAKECMDGOALS)) ,clean)	# if target is not clean
ifneq "$(SRC) $(MAIN_SRC)" " "			# if SRC or MAIN_SRC are not void
DEPEND := $(shell grep -v ^\# $(DEPEND_FILE) 2>/dev/null)   # take the dependencies in DEPEND_FILE 
DEPEND := $(subst \, ,$(DEPEND))        # remove the back-slash
DEPEND := $(subst :, , $(DEPEND))       # remove the :
DEPEND := $(filter-out %.o, $(DEPEND))  # remove the .o file names
DEPEND := $(SRC) $(MAIN_SRC) $(DEPEND)  # add SRC and MAIN_SRC in case not present before
DEPEND := $(sort $(DEPEND))             # remove doublons 
endif
endif

ifeq "$(TARGET_TYPE)" "OBJFILE"
TARGET := $(OBJ)
endif

ifeq "$(TARGET_TYPE)" "EXECUTABLE"
TARGET := $(EXE)
endif

ifeq "$(TARGET_TYPE)" "MICROCODE"
ifneq ($(findstring _bin.h, $(TARGET)), _bin.h)
$(error target name is not valid (must end with _bin.h))
endif
UCODE_NAME := $(subst _bin.h,, $(TARGET))
TMP_FILES += $(addprefix $(UCODE_NAME), .log .lst .map .prof .bin _release.bin _labels.h _structures.h)
endif

#################################################################
## put here the targets you want to be updated anytime you ask ##
#################################################################

.PHONY: cleanall clean dist localdist reduce_dist api $(SUBDIR) all local prebuild postbuild copy_shared

#############################
## rule to make the target ##
#############################

all: prebuild $(SUBDIR) $(TARGET) postbuild

local: prebuild $(TARGET) postbuild

ifeq "$(TARGET_TYPE)" "SHARED_LIBRARY"
copy_shared: $(TARGET) $(SUBDIR)
	[ -d $(LIBDIR) ] && cp -f $(TARGET) $(LIBDIR)
else
copy_shared: $(SUBDIR)
endif

ifeq "$(TARGET_TYPE)" "OBJFILE"
copy_static: $(TARGET) $(SUBDIR)
	[ -d $(LIBDIR) ] && cp -f $(TARGET) $(LIBDIR)
endif

ifeq "$(TARGET_TYPE)" "OBJECT_LIBRARY"
$(TARGET): $(OBJ) $(STATIC_LINKS_WITH) $(EXTERNAL_STATIC_LINKS_WITH)
	$(OBJECT_LIBRARY)

copy_static: $(TARGET) $(SUBDIR)
	[ -d $(LIBDIR) ] && cp -f $(TARGET) $(LIBDIR)
endif

ifeq "$(TARGET_TYPE)" "RELOCATABLE_OBJECT"
$(TARGET): $(OBJ) $(STATIC_LINKS_WITH) $(EXTERNAL_STATIC_LINKS_WITH)
	$(RELOCATABLE_OBJECT)

copy_static: $(TARGET) $(SUBDIR)
	[ -d $(LIBDIR) ] && cp -f $(TARGET) $(LIBDIR)
endif

ifeq "$(TARGET_TYPE)" "SHARED_LIBRARY"
$(TARGET): $(OBJ) $(STATIC_LINKS_WITH) $(EXTERNAL_STATIC_LINKS_WITH) $(IMPLICIT_LINKS_WITH)
	$(SHARED_LIBRARY)
endif

ifeq "$(TARGET_TYPE)" "EXECUTABLE"
$(TARGET): %: %.o $(OBJ) $(STATIC_LINKS_WITH) $(EXTERNAL_STATIC_LINKS_WITH) $(IMPLICIT_LINKS_WITH)
	$(EXECUTABLE)
endif

ifeq "$(TARGET_TYPE)" "MICROCODE"
ifneq ($(findstring $(SRC), $(wildcard $(SRC))), $(SRC))
ifneq ($(findstring $(TARGET), $(wildcard $(TARGET))), $(TARGET))
$(error $(TARGET) is missing along with the source files!!!)
else
$(TARGET):
	$(MICROCODE)
endif
else
$(TARGET): $(SRC)
	$(MICROCODE)
endif
endif

ifeq "$(TARGET_TYPE)" "BINARY_FILE"
$(TARGET): $(EXE)
	$(BINARY_FILE)

$(EXE): %: $(OBJ) %.o $(STATIC_LINKS_WITH) $(EXTERNAL_STATIC_LINKS_WITH) $(IMPLICIT_LINKS_WITH)
	$(EXECUTABLE)
endif

$(STATIC_LINKS_WITH): $(@D)

$(IMPLICIT_LINKS_WITH): $(@D)

$(DEPENDS_ON): $(@D)

# we can remove duplicates and sort the targets for the
# rule definition as it bears no consequence about linking
# order.
$(sort $(SUBDIR)):
	@ echo ---- Subdir $(SRCDIR)$@
	$(MAKE) $(MAKEARGS) -C $@ $(MAKECMDGOALS) SRCDIR=$(SRCDIR)$@

$(SRC):

$(MAIN_SRC):

.PHONY: echo-subdirs
echo-subdirs:
	@echo $(SUBDIR)

########################################################
## rules to execute a prebuild or a postbuild command ##
########################################################

prebuild:
ifneq "$(PREBUILD)" ""
	@echo "------START PREBUILD STEP-------"
	$(PREBUILD)
	@echo "------FINISH PREBUILD STEP------"
endif

postbuild:
ifneq "$(POSTBUILD)" ""
	@echo "------START POSTBUILD STEP-------"
	$(POSTBUILD)
	@echo "------FINISH POSTBUILD STEP------"
endif

######################################################
## include and compute if necessary the DEPEND_FILE ##
######################################################

$(SRC_LIST_FILE): $(SRC) $(MAIN_SRC)
	echo $^ > $@

$(DEPEND_FILE): override PREPROCESSORFLAGS := $(filter-out -nostdinc, $(PREPROCESSORFLAGS))
$(DEPEND_FILE): $(DEPEND) $(LOCAL_MAKEFILE) $(SRC_LIST_FILE)
ifneq "$(TARGET_TYPE)" "MICROCODE"
ifneq "$(SRC) $(MAIN_SRC)" " " 
	$(DEPEND_COMMAND) | sed 's/:/: $(LOCAL_MAKEFILE)/' > $@
endif
endif

ifneq ($(findstring clean, $(MAKECMDGOALS)) ,clean)
-include $(DEPEND_FILE)
endif

######################################################
## copy all the needed files to compile the project ##
######################################################

dist: localdist $(SUBDIR)

dist2: localdist2 $(SUBDIR)

localdist:
	@[ -d $(DISTDIR) ] && \
	echo $(EXTRA_DIST_FILES) | cat - $(DEPEND_FILE) 2>/dev/null | $(call COPY_FILE_LIST,$(DISTDIR)) 

localdist2:
	@mkdir -p $(DISTDIR)/$(SRCDIR) && \
	echo $(EXTRA_DIST_FILES) | cat - $(DEPEND_FILE) 2>/dev/null | $(call COPY_FILE_LIST2,$(DISTDIR)/$(SRCDIR))

reduce_dist:
	@$(call REDUCE_DIR_TREE,$(DISTDIR))

api: $(SRC_LIST_FILE)
	@[ -d $(APIDIR) ] && \
	$(DEPEND_COMMAND) | $(call COPY_FILE_LIST,$(APIDIR)) 
	@$(call REDUCE_DIR_TREE,$(APIDIR))


##############################
## how to clean the project ##
##############################

cleanall: clean $(SUBDIR)

clean: 
	$(RM) $(OBJ) $(MAIN_OBJ) $(TARGET) $(DEPEND_FILE) $(SRC_LIST_FILE) $(TMP_FILES) $(EXTRA_CLEAN_FILES) *~ *.gdb
